#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

#--------------------------------------------------------------------- Settings
__build_id__ = 320025
__version__ = 3.20


#---------------------------------------------------------------------- Imports
from ._base import (AuthenticationError, ConnectionError, ConcurrencyError,
                    Config, DatabaseError, DataType, Event, EventType,
                    FileType, IntegrityError, Progress, Register,
                    trace as _trace, VersionError)
from . import _workers
from ._objects.attribute import Attribute
from ._objects.user import Authenticate, User
from ._objects.connection import Connection as _Connection
from ._objects.record import Record


#------------------------------------------------------ Top level HTD functions
@_trace
def connect(location, user=None, file_obj=None):
    """ Connect to a table or db """
    return _Connection(location, user, file_obj)


@_trace
def create(location, identifier=None, attributes={}, users={}, login_user=None,
           htd_file=None):
    """
    Create a new HTD File
    To create a table parse the identifier for the table otherwise leave
    the identifier blank to create a db
    """

    # Confirm location is str
    if not isinstance(location, str):
        raise ValueError("location (%s) must be a str object." % location)

    # Confirm attributes is dict
    if not isinstance(attributes, dict):
        raise ValueError("attributes (%s) must be a dict object." % attributes)

    # Confirm users is dict
    if not isinstance(users, dict):
        raise ValueError("users (%s) must be a dict object." % users)

    # Check for identifier
    if identifier is not None:
        if not isinstance(identifier, list):  # Convert single into a list
            identifier = [identifier]

        # Create ids
        for i in identifier:
            if i not in attributes:
                raise ValueError("Identifier must be in attributes parameter.")

        idn = 0

        # Check Attributes are an Attribute object
        for i in attributes:
            if not isinstance(attributes[i], Attribute):
                raise ValueError("Attribute(s) must be an Attribute object.")

            attributes[i]._id = idn  # Force IDs
            idn += 1

            if i is None:
                raise ValueError("Attribute name cannot be NULL.")

            attributes[i].name = i  # Run name check

        table = None  # Set TABLE mode
    else:
        # Create db attributes
        name_att = Attribute(DataType.STRING)
        name_att._id = 0
        loc_att = Attribute(DataType.STRING)
        loc_att._id = 1

        # Load attribute and identifier
        attributes = {"name": name_att, "location": loc_att}
        identifier = ["name"]
        table = False  # Set DB mode master file

    # Load guest account if needed
    guest = False
    if len(users) == 0:
        users = {"guest": User(None)}
        guest = True

    # Confirm users are valid
    acc = False

    for i in users:
        if not isinstance(users[i], User):  # Error if not User object
            raise ValueError("User(s) must be an User object.")

        if users[i].access == "roadx":  # Check for full access
            acc = True

        users[i].username = i  # Add username check

    if not acc:  # At least one user has full access
        raise ValueError("One User must have full access (roadx)")

    if len(users) > 1:
        if login_user is None:
            raise ValueError("A login_user must be specified")
    else:
        login_user = list(users.keys())[0]

    if login_user not in users:
        raise ValueError("login_user must be in users")
    else:
        if guest:
            cred = Authenticate("guest", None)
        else:
            cred = Authenticate(user_obj=users[login_user])

    # Load file type
    if htd_file is None:
        htd_file = Register.default_fhandler

    # Check htd_file exists
    if htd_file not in Register.fhandlers:
        raise ValueError("htd_file has not been registered.")

    # Check location is correct for file type
    if Register.fhandlers[htd_file]["is_loc"] is not None:
        if not Register.fhandlers[htd_file]["is_loc"](location):
            raise ValueError("An invalid location (%s) was given for the "
                             "file format (%s)" % (location, htd_file))

    # Build Header
    header = {"TAB": table, "ATT": attributes, "USR": users, "KEY": identifier,
              "VER": __version__, "R1M": {}, "RM1": {}}

    # Load and set header
    file_obj = htd_file(location, "roadx", header)

    if file_obj.persistent:  # Close the file if persistent
        file_obj.close()
        file_obj = None

    return connect(location, cred, file_obj=file_obj)  # Return con obj
