#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .._base import Event as Evt, EventType as EvtType
from .._base import Progress, ProgressItem
from threading import current_thread


class ConfigMetaClass(type):
    """ Config for HTD with auto attach to Event """

    errors = True
    force_sync = False
    _debug = False
    _progress = False

    @property
    def debug(mcs):
        return mcs._debug

    @debug.setter
    def debug(mcs, state):
        if state:
            Evt.connect(EvtType.BEFORE_RUN, EvtType.ALL_SIGNALS, call)
            Evt.connect(EvtType.AFTER_RUN, EvtType.ALL_SIGNALS, returned)
            Evt.connect(EvtType.ERROR, EvtType.ALL_SIGNALS, error)
        else:
            Evt.disconnect(EvtType.BEFORE_RUN, EvtType.ALL_SIGNALS, call)
            Evt.disconnect(EvtType.AFTER_RUN, EvtType.ALL_SIGNALS, returned)
            Evt.disconnect(EvtType.ERROR, EvtType.ALL_SIGNALS, error)

        mcs._debug = state

    @property
    def progress(mcs):
        return mcs._progress

    @progress.setter
    def progress(mcs, state):
        arr = ((EvtType.BEFORE_LOCK, EvtType.ALL_SIGNALS, prog_create),
               (EvtType.BEFORE_RUN, EvtType.ALL_SIGNALS, prog_brun),
               (EvtType.AFTER_RUN, EvtType.ALL_SIGNALS, prog_arun),
               (EvtType.AFTER_UNLOCK, EvtType.ALL_SIGNALS, prog_remove),
               (EvtType.ERROR, EvtType.ALL_SIGNALS, prog_remove))

        if state:
            for args in arr:
                Evt.connect(*args)
        else:
            for args in arr:
                Evt.disconnect(*args)

        mcs._progress = True


#--------------------------------------------------------- Callbacks for config
def call(idn, event, signal, *args, **kwargs):
    # Build parameters
    tmp = [str(i) for i in args]
    tmp.extend(["%s=%s" % (i, str(kwargs[i])) for i in kwargs])
    param = ", ".join(tmp)

    print("HTD DEBUG: %s Calling %s(%s)" % (idn, signal, param))


def returned(idn, event, signal, result, *args, **kwargs):
    print("HTD DEBUG: %s %s returned %s" % (idn, signal, result))


def error(idn, event, signal, exception, *args, **kwargs):
    txt = "%s %s, %s (%s)" % (idn, type(exception).__name__, exception, signal)

    print("HTD DEBUG: %s" % txt)


def prog_create(idn, event, signal, *args, **kwargs):
    Progress.add(idn, ProgressItem(current_thread(), 3,
                                   "Waiting for lock (%s)" % signal))


def prog_brun(idn, event, signal, *args, **kwargs):
    task = Progress.get(idn)
    task.count()
    task.text = "Calling %s" % signal


def prog_arun(idn, event, signal, *args, **kwargs):
    task = Progress.get(idn)
    task.count()
    task.text = "%s Returned" % signal


def prog_remove(idn, event, signal, *args, **kwargs):
    Progress.remove(idn)


Config = ConfigMetaClass("Config", (object,), {})  # Trace requires config
