#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from re import sub
from .._base.exceptions import ConcurrencyError


INTERNAL_ATTRIBUTES = [".version"]


def check_record_version(version, current_version):
    """ Check that the record is of the correct version """

    if version is None:
        version = current_version

    if version < current_version:
        raise ConcurrencyError("The data has changed.")

    return version


def get_attributes_by_id(atts):
    """ Sort attributes by id """

    attributes = {}

    for att in atts:
        attr = atts[att]
        attributes[attr._id] = {"size": attr.size, "name": att}

    return attributes


def get_key_as_dict(data, key):
    if type(data) not in [list, tuple]:
        data = [data]

    return dict(zip(key, data))


def get_key_as_string(key, atts):
    return b"".join([key[att] for att in atts])


def get_key_from_data(data, key):
    out = {}

    for attribute in key:
        out[attribute] = data[attribute]

    return out


def sanitise(data, attributes):
    """ Sanitise data so that it can be stored in the db """

    if data is None:
        return None

    # Check data attributes exist
    for att in data:  # Ignore internal attributes
        if att not in attributes and att not in INTERNAL_ATTRIBUTES:
            raise AttributeError("Failed to find sanitise settings for "
                                 "attribute %s" % att)

    for att in attributes:
        if att in data:  # Load data
            tmp = data[att]
        else:
            tmp = None

        attr = attributes[att]  # load attribute

        if tmp is None:
            if attr._default is not None:
                tmp = attr._default  # Load default value
                data[att] = tmp
                continue
            elif attr._required:  # Check required
                raise ValueError("Required attribute (%s) was NULL" % att)
            else:
                data[att] = None
                continue

        typed = attr._dtype  # Load dtype

        if typed.type_check(tmp) is False:  # Type check
            name = ('%s, failed to  match preset pattern for %s. (%s)' %
                    (att, typed, tmp))

            raise ValueError(name)

        if typed.can_range and attr._range is not False:  # Range check
            if not (attr._range[1] < tmp < attr._range[0]):
                raise ValueError('"%s" failed to match range for "%s". (%s)' %
                                 (att, typed, tmp))

        if attr._limit_list[0] and tmp not in attr._limit_list[1]:  # List
            raise KeyError('"%s" failed to match an item in list.' % att)

        tmp = typed.sanitise(tmp)  # Convert data

        if len(tmp) > attr._size:  # Length Check
            raise ValueError('String length check failed for "%s". (%s)' %
                             (att, tmp))

        data[att] = tmp  # Update data

    return data


def sanitise_key(key, header):
    """ Get the key from data and sanitise the key """

    key = get_key_as_dict(key, header["KEY"])

    return sanitise(key, header["ATT"])


def unsanitise(data, attributes):
    """ Unsanitise data so that it can be used by the client """

    for att in data:
        if att in INTERNAL_ATTRIBUTES:  # Ignore internal attributes
            continue

        dtype = attributes[att]._dtype
        patt = attributes[att]._re

        if data[att] is not None:
            if not not patt:  # Run reg expr
                data[att] = sub(patt, b"", data[att])

            data[att] = dtype.unsanitise(data[att])

    return data


def update_record(new, old):
    """ Get missing data from old into new """

    for att in old:
        if att not in new:
            new[att] = old[att]

    return new


def where_compile(where, args, atts):
    """ Compile the where criteria into eval code """

    for i in dict(args):  # Force represent
        where = where.replace("{%s}" % i, "{%s!r}" % i)

        # Sanitise data
        v = args[i]
        tmp = sanitise({v[0]: v[1]}, atts)
        args[i] = tmp[v[0]]

    where = where.format(**args)  # Load args into where
    return compile(where, "", "eval")  # Preload eval string
