#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide

Base File for HTD
Provides slicing and file locking support for files
'''

from os import name, unlink
from os.path import exists, isabs, isdir, sep, split

if name == "posix":
    import fcntl
elif name == "nt":
    try:
        import win32con
        import win32file
        import pywintypes
    except ImportError:
        raise ImportError("win32api needs to be installed, FileLock "
                          "unavailable")
else:
    print("HTD Warning: Unknown operating system, FileLock unavailable.")


def convert_location(location, add, extension=None):
    """ Insert *add* into the location (/a/b.htd => /a/.~lock_b.htd) """

    folder = location.split(sep)
    filename = "".join([add or "", folder[-1]])  # create new filename

    if extension is not None:
        filename = filename.split(".")
        filename.insert(-1, extension)
        filename = ".".join(filename)

    folder.pop()  # remove name
    folder.append(filename)

    return sep.join(folder)


def location_is_file(location):
    """ Check location is a HTD file """

    if not isabs(location):
        return False

    directory, filename = split(location)

    if not isdir(directory):
        return False

    if not filename.endswith(".htd"):
        return False

    return True


class FileTools(object):
    """ Helper for writing to files """

    @staticmethod
    def delete_region(writer, start, stop):
        """ Delete the region from start to stop """

        FileTools.insert_region(writer, start, b"", stop - start)

    if name == "posix":
        @staticmethod
        def insert_region(writer, position, data, skip=0):
            """
            Insert data at position overwriting the region from position to
            position + skip
            """

            jump = 10 ** 6
            leng = len(data)

            if jump < leng:
                jump = leng

            if data == b"":
                cache = []
            else:
                cache = [data]  # Buffer always 1 in front

            writer.seek(position)

            with open(writer.name, "rb") as reader:
                reader.seek(position + skip)
                writing = True

                while writing:  # Write 1MB chunks
                    cache.append(reader.read(jump))

                    writing = cache[0] != b""
                    writer.write(cache[0])
                    cache.pop(0)

            writer.truncate()
    else:
        @staticmethod
        def insert_region(fileio, position, data, skip=0):
            """
            Insert data at position overwriting the region from position to
            position + skip
            """

            jump = 10 ** 6
            leng = len(data)

            if jump < leng:
                jump = leng

            if data == b"":
                cache = []
            else:
                cache = [data]  # Buffer always 1 in front

            write_pos = position
            read_pos = position + skip

            writing = True

            while writing:  # Write 1MB chunks
                fileio.seek(read_pos)
                cache.append(fileio.read(jump))
                read_pos += jump

                fileio.seek(write_pos)
                writing = cache[0] != b""
                fileio.write(cache[0])
                cache.pop(0)
                write_pos = fileio.tell()

            fileio.truncate()

    @staticmethod
    def overwrite_region(writer, position, width, data):
        """ Write data over the region from position to position + width """

        FileTools.insert_region(writer, position, data, skip=width)


class LockFile():
    """ Class which allows manages locking a file """
    def __init__(self):
        self.file = None
        self.locked = False

    def close_lock_file(self, delete_files=False):
        """ Close the lock file with the option to delete the files """

        if self.file is not None:
            self.file.close()
            self.file = None

        if delete_files and exists(self.location):
            unlink(self.location)

    if name == "posix":
        def lock_reading(self):
            """ Lock file for reading """

            if self.file is None:
                self.file = open(self.location, "rb")

            fcntl.flock(self.file, fcntl.LOCK_SH | fcntl.LOCK_NB)
            self.locked = None

    elif name == "nt":
        def lock_reading(self):
            """ Lock file for reading """

            self.unlock()  # NT requires unlock first

            if self.file is None:
                self.file = open(self.location, "rb")

            f = win32file._get_osfhandle(self.file.fileno())

            try:
                win32file.LockFileEx(f, 0 | win32con.LOCKFILE_FAIL_IMMEDIATELY,
                                     0, -0x10000, pywintypes.OVERLAPPED())
            except pywintypes.error as exc_value:
                if exc_value.args[0] == 33:  # File locked exception
                    raise IOError("File is locked")
                else:
                    raise IOError("Unknown Error")

            self.locked = None

    if name == "posix":
        def lock_writing(self):
            """ Lock file for writing """

            if self.file is None:
                self.file = open(self.location, "rb+")

            fcntl.flock(self.file, fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.locked = True

    elif name == "nt":
        def lock_writing(self):
            """ Lock file for writing """

            self.unlock()  # NT requires unlock first

            if self.file is None:
                self.file = open(self.location, "rb+")

            f = win32file._get_osfhandle(self.file.fileno())

            try:
                win32file.LockFileEx(f, win32con.LOCKFILE_EXCLUSIVE_LOCK |
                                     win32con.LOCKFILE_FAIL_IMMEDIATELY, 0,
                                     - 0x10000, pywintypes.OVERLAPPED())
            except pywintypes.error as exc_value:
                if exc_value.args[0] == 33:  # File locked exception
                    raise IOError("File is locked")
                else:
                    raise IOError("Unknown Error")

            self.locked = True

    if name == "posix":
        def unlock(self):
            """ Unlock file """

            if self.file is not None:
                fcntl.flock(self.file, fcntl.LOCK_UN)
                self.locked = False

                self.file.close()
                self.file = None

    elif name == "nt":
        def unlock(self):
            """ Unlock file """

            if self.file is not None:
                f = win32file._get_osfhandle(self.file.fileno())

                try:
                    win32file.UnlockFileEx(f, 0, -0x10000,
                                           pywintypes.OVERLAPPED())
                except pywintypes.error as exc_value:
                    if exc_value.args[0] == 158:  # Don't raise already locked
                        pass
                    else:
                        raise

                self.locked = False

                self.file.close()
                self.file = None
