#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .exceptions import ConnectionError
from os.path import exists


#-------------------------------- Create containers for datatypes and filetypes
class RegisterContainer(object):
    pass

DataType = RegisterContainer()
FileType = RegisterContainer()


#--------------------------------------------------------------- Register Class
class RegisterMetaClass(type):
    dtype_codes = {}
    dtypes = {}
    fhandlers = {}
    _default_dtype = None
    _default_fhandler = None

    @property
    def default_dtype(mcs):
        return mcs._default_dtype

    @default_dtype.setter
    def default_dtype(mcs, dtype):
        if dtype.__name__ not in mcs.dtypes:
            raise ValueError("Default dtype must be a registered dtype.")

        mcs._default_dtype = dtype.__name__

    @property
    def default_fhandler(mcs):
        return mcs._default_fhandler

    @default_fhandler.setter
    def default_fhandler(mcs, fhandler):
        if fhandler not in mcs.fhandlers:
            raise ValueError("Default fhandler must be a registered fhandler.")

        mcs._default_fhandler = fhandler

    @classmethod
    def register_dtype(mcs, dtype):
        setattr(DataType, dtype.__name__, dtype())

        mcs.dtype_codes[dtype.idn] = dtype.__name__
        mcs.dtypes[dtype.__name__] = dtype.idn

        return True

    @classmethod
    def register_fhandler(mcs, fhandler, is_location, is_file=None):
        setattr(FileType, fhandler.__name__, fhandler)

        mcs.fhandlers[fhandler] = {"is_loc": is_location, "is_file": is_file}

        return True

    @classmethod
    def resolve_fhandler(mcs, location):
        """ File Handler resolver - determins file type """

        # Filter fhandlers by location
        fhandlers = [f for f in mcs.fhandlers if
                     mcs.fhandlers[f]["is_loc"](location)]

        # Filter fhandlers by file
        output = []

        for fhandler in fhandlers:
            # if file_format exists then uses File and Header standard
            if fhandler.idn is not None:
                if not exists(location):
                    raise ConnectionError("No HTD file found (%s)" % location)

                # Open file and get fhandler code
                with open(location, "rb") as htdfile:
                    htdfile.seek(8)

                    if htdfile.read(1) == fhandler.idn:
                        output.append(fhandler)

            elif mcs.fhandlers[fhandler]["is_file"] is not None:
                if mcs.fhandlers[fhandler]["is_file"](location):
                    output.append(fhandler)

        fhandlers = output

        # Load fhandler
        if len(fhandlers) == 0:
            raise ConnectionError("Unknown File Format")
        elif len(fhandlers) > 1:
            raise ConnectionError("Could not determine File Format either (%s)"
                                  % fhandlers)
        else:
            return fhandlers[0]

Register = RegisterMetaClass("Register", (object,), {})

from . import types
