#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from threading import current_thread, Lock


class ThreadLock(object):
    """ Class that manages thread locking """
    def __init__(self):
        self._threading_lock = Lock()
        self._locked_lock = None
        self._held = False

    def _acquire(self):
        """ Acquire threading lock """

        if self._held and self._locked_lock == current_thread():
            return True

        response = self._threading_lock.acquire()
        self._locked_lock = current_thread()

        return response

    def _acquire_lock(self):
        """ Acquire lock if thread is not locked otherwise return False """

        if self._held and self._locked_lock == current_thread():
            return True

        locked = False

        if not self._thread_locked:
            self._acquire()
            locked = True

        return locked

    def _hold_thread_lock(self, thread):
        """ Lock to thread and hold until release_from_thread """

        self._threading_lock.acquire()
        self._locked_lock = thread
        self._held = True

    def _release(self):
        """ Release threading lock """

        if self._held and self._locked_lock == current_thread():
            return True

        self._locked_lock = None
        return self._threading_lock.release()

    def _release_thread_lock(self, thread):
        """ Release hold from thread """

        self._locked_lock = None
        self._held = False
        self._threading_lock.release()

    @property
    def _thread_locked(self):
        """ Return True if thread has been locked by this thread """

        return (self._threading_lock.locked() and
                current_thread() == self._locked_lock)
