#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide

Tracing for HTD
Captures progress, logging, triggers, errors and permission levels
'''

from .._base.config import Config
from .._base.event import Event
from threading import Lock

# Event ids
IDN_LOCK = Lock()
IDN = 1


def trace(func):
    """ Trace function  - Manages threads and emits events """

    def callf(*args, **kwargs):
        global IDN

        # Generate name and id for event
        func_name = func.__name__

        IDN_LOCK.acquire()
        i = IDN
        IDN += 1
        IDN_LOCK.release()

        # Check that object isn't closed
        try:
            if not args[0].alive:
                return None
        except (IndexError, AttributeError):
            tmp = None
        else:
            tmp = args[0]
            func_name = ".".join([tmp.__class__.__name__, func_name])

        Event.emit(i, 1, func_name, args, kwargs)

        # Acquire lock
        locked = False

        try:
            if not tmp._thread_locked:
                tmp._acquire()
                locked = True
        except AttributeError:
            pass

        Event.emit(i, 2, func_name, args, kwargs)

        try:
            result = func(*args, **kwargs)
        except Exception as exception:
            # Unlock thread and file lock
            if (hasattr(tmp, "_file") and hasattr(tmp._file, "locked")
                    and tmp._file.locked):
                tmp._file.unlock()

            if hasattr(tmp, "_thread_locked") and tmp._thread_locked:
                tmp._release()

            args = list(args)
            args.insert(0, exception)

            Event.emit(i, 3, func_name, args, kwargs)

            if Config.errors:
                raise

            return False
        else:
            Event.emit(i, 4, func_name, result)

        if locked:  # Release lock this locked it
            try:
                if tmp._thread_locked:
                    tmp._release()
            except AttributeError:
                pass

        Event.emit(i, 5, func_name, result)

        return result

    return callf
