#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide

HTD Data Types
Data types and (un)sanisite functions are managed here
'''

from datetime import date, datetime, time, timedelta
from .. import __version__
from .register import DataType, Register
import sys


#-------------------------------------------------------------------- HTD Types
class DType(object):
    """
    Base for Data Types
    - Provides generic __repr__
    """
    def __repr__(self):
        return ("<HTD%s Data Type (%s)>" %
                (__version__, self.__class__.__name__))


if sys.version_info < (3, 0):
    char = chr
else:
    def char(data):
        """ chr for Python3 so code can remain compatible with Python2  """
        return chr(int(data)).encode("ISO-8859-1")


if sys.version_info < (3, 0):
    ordi = ord
else:
    def ordi(raw):
        """ ord for Python3 so code can remain compatible with Python2 """
        return raw


#--------------------------------------------------------------- Raw data types
class BOOL(DType):
    """
    BOOL data type for HTD
    - Allows values of True and False
    - Takes up 1 byte of storage
    - Uses the identifier \x01
    """
    idn = b"\x01"
    can_range = False

    @staticmethod
    def sanitise(real):
        if real:
            return b"1"
        else:
            return b"0"

    size = 1

    @staticmethod
    def type_check(real):
        return type(real) is bool

    @staticmethod
    def unsanitise(raw):
        return raw == b"1"

Register.register_dtype(BOOL)


class BYTES(DType):
    """
    BYTES data type for HTD
    - Allows values of type bytes
    - Uses the identifier \x02
    """
    idn = b"\x02"
    can_range = False

    @staticmethod
    def sanitise(raw):
        return bytes(raw)

    size = 0

    @staticmethod
    def type_check(real):
        return type(real) is bytes

    @staticmethod
    def unsanitise(raw):
        return bytes(raw)

Register.register_dtype(BYTES)


class STRING(DType):
    """
    STRING data type for HTD
    - Allows values of type str
    - Uses the identifier \x04
    """
    idn = b"\x04"
    can_range = False

    if sys.version_info < (3, 0):
        @staticmethod
        def sanitise(raw):
            return str(raw)
    else:
        @staticmethod
        def sanitise(raw):
            return raw.encode("utf")

    size = 0

    @staticmethod
    def type_check(real):
        return type(real) is str

    if sys.version_info < (3, 0):
        @staticmethod
        def unsanitise(raw):
            return str(raw)
    else:
        @staticmethod
        def unsanitise(raw):
            return raw.decode("utf")

Register.register_dtype(STRING)


#----------------------------------------------- Date and time based data types
class DATE(DType):
    """
    DATE data type for HTD
    - Allows values of type date
    - Takes up 10 bytes of storage
    - Uses the identifier \x07
    """
    idn = b"\x07"
    can_range = False

    @staticmethod
    def sanitise(raw):
        return raw.isoformat().encode("ISO-8859-1")

    size = 10

    @staticmethod
    def type_check(real):
        return type(real) is date

    @staticmethod
    def unsanitise(raw):
        data = raw.decode("ISO-8859-1").split("-")
        return date(int(data[0]), int(data[1]), int(data[2]))

Register.register_dtype(DATE)


class TIME(DType):
    """
    TIME data type for HTD
    - Allows values of type time
    - Takes up 8 bytes of storage
    - Uses the identifier \x08
    """
    idn = b"\x08"
    can_range = False

    @staticmethod
    def sanitise(raw):
        return raw.isoformat().encode("ISO-8859-1")

    size = 8

    @staticmethod
    def type_check(real):
        return type(real) is time

    @staticmethod
    def unsanitise(raw):
        data = raw.decode("ISO-8859-1").split(":")
        data[2] = data[2].split(".")
        data[2].append(0)
        return time(int(data[0]), int(data[1]), int(data[2][0]),
                    int(data[2][1]))

Register.register_dtype(TIME)


class DATETIME(DType):
    """
    DATETIME data type for HTD
    - Allows values of type datetime
    - Takes up 26 bytes of storage
    - Uses the identifier \t
    """
    idn = b"\t"
    can_range = False

    @staticmethod
    def sanitise(raw):
        return raw.isoformat().encode("ISO-8859-1")

    size = 26

    @staticmethod
    def type_check(real):
        return type(real) is datetime

    @staticmethod
    def unsanitise(raw):
        data = raw.decode("ISO-8859-1").split(".")
        dt = datetime.strptime(data[0][:19], "%Y-%m-%dT%H:%M:%S")

        if len(data) > 1:
            us = int(data[1].rstrip("Z"), 10)
        else:
            us = 0

        return dt + timedelta(microseconds=us)

Register.register_dtype(DATETIME)


#------------------------------------------------------------ Number data types
class BYTE(DType):
    """
    BYTE data type for HTD
    - Allows values of type int -128 <= x <= 127
    - Takes up 1 byte of storage
    - Uses the identifier \x0f
    """
    can_range = True
    idn = b"\x0f"
    size = 1

    @staticmethod
    def sanitise(real):
        return char(real + 128)

    @staticmethod
    def type_check(real):
        return type(real) is int

    @staticmethod
    def unsanitise(raw):
        return ord(raw) - 128


class UBYTE(DType):
    """
    UBYTE, unsigned byte, data type for HTD
    - Allows values of type int 0 <= x <= 255
    - Takes up 1 byte of storage
    - Uses the identifier \x10
    """
    can_range = True
    idn = b"\x10"
    size = 1

    @staticmethod
    def sanitise(real):
        return char(real)

    @staticmethod
    def type_check(real):
        return type(real) is int

    @staticmethod
    def unsanitise(raw):
        return ord(raw)


class SHORT(DType):
    """
    SHORT data type for HTD
    - Allows values of type int -32768 <= x <= 32767
    - Takes up 2 byte of storage
    - Uses the identifier \x11
    """
    can_range = True
    idn = b"\x11"
    size = 2

    @staticmethod
    def sanitise(real):
        real += 32768

        a, b = divmod(real, 256)

        return b"".join([char(a), char(b)])

    @staticmethod
    def type_check(real):
        return type(real) is int

    @staticmethod
    def unsanitise(raw):
        a, b = raw

        return (ordi(a) * 256 + ordi(b)) - 32768


class USHORT(DType):
    """
    USHORT, unsigned short, data type for HTD
    - Allows values of type int 0 <= x <= 65535
    - Takes up 2 bytes of storage
    - Uses the identifier \x12
    """
    can_range = True
    idn = b"\x12"
    size = 2

    @staticmethod
    def sanitise(real):
        a, b = divmod(real, 256)

        return b"".join([char(a), char(b)])

    @staticmethod
    def type_check(real):
        return type(real) is int

    @staticmethod
    def unsanitise(raw):
        a, b = raw

        return (ordi(a) * 256) + ordi(b)


class INTEGER(DType):
    """
    INTEGER data type for HTD
    - Allows values of type int -2147483648 <= x <= 2147483647
    - Takes up 4 bytes of storage
    - Uses the identifier \x13
    """
    can_range = True
    idn = b"\x13"
    size = 4

    @staticmethod
    def sanitise(real):
        real += 2147483648

        a, real = divmod(real, 256 ** 3)
        b, real = divmod(real, 256 ** 2)
        c, d = divmod(real, 256)

        return b"".join([char(a), char(b), char(c), char(d)])

    @staticmethod
    def type_check(real):
        return type(real) is int

    @staticmethod
    def unsanitise(raw):
        a, b, c, d = raw

        return ((ordi(a) * 256 ** 3) + (ordi(b) * 256 ** 2) +
                (ordi(c) * 256) + ordi(d) - 2147483648)


class UINTEGER(DType):
    """
    UINTEGER, unsigned integer, data type for HTD
    - Allows values of type int 0 <= x <= 4294967295
    - Takes up 4 bytes of storage
    - Uses the identifier \x14
    """
    can_range = True
    idn = b"\x14"
    size = 4

    @staticmethod
    def sanitise(real):
        a, real = divmod(real, 256 ** 3)
        b, real = divmod(real, 256 ** 2)
        c, d = divmod(real, 256)

        return b"".join([char(a), char(b), char(c), char(d)])

    @staticmethod
    def type_check(real):
        return type(real) is int

    @staticmethod
    def unsanitise(raw):
        a, b, c, d = raw

        return ((ordi(a) * 256 ** 3) + (ordi(b) * 256 ** 2) +
                (ordi(c) * 256) + ordi(d))


class LONG(DType):
    """
    ULONG data type for HTD
    - Allows values of type int or long (if using python2)
        from -9223372036854775808 <= x <= 9223372036854775806
    - Takes up 8 bytes of storage
    - Uses the identifier \x15
    """
    can_range = True
    idn = b"\x15"
    size = 8

    @staticmethod
    def sanitise(real):
        real += 9223372036854775808

        a, real = divmod(real, 256 ** 7)
        b, real = divmod(real, 256 ** 6)
        c, real = divmod(real, 256 ** 5)
        d, real = divmod(real, 256 ** 4)
        e, real = divmod(real, 256 ** 3)
        f, real = divmod(real, 256 ** 2)
        g, h = divmod(real, 256)

        return b"".join([char(a), char(b), char(c), char(d), char(e), char(f),
                         char(g), char(h)])

    if sys.version_info < (3, 0):
        @staticmethod
        def type_check(real):
            return type(real) is long
    else:
        @staticmethod
        def type_check(real):
            return type(real) is int

    @staticmethod
    def unsanitise(raw):
        a, b, c, d, e, f, g, h = raw

        return ((ordi(a) * 256 ** 7) + (ordi(b) * 256 ** 6) +
                (ordi(c) * 256 ** 5) + (ordi(d) * 256 ** 4) +
                (ordi(e) * 256 ** 3) + (ordi(f) * 256 ** 2) +
                (ordi(g) * 256) + ordi(h)) - 9223372036854775808


class ULONG(DType):
    """
    ULONG, unsigned long, data type for HTD
    - Allows values of type int or long (if using python2)
        from 0 <= x <= 18446744073709551615
    - Takes up 8 byte of storage
    - Uses the identifier \x16
    """
    can_range = True
    idn = b"\x16"
    size = 8

    @staticmethod
    def sanitise(real):
        a, real = divmod(real, 256 ** 7)
        b, real = divmod(real, 256 ** 6)
        c, real = divmod(real, 256 ** 5)
        d, real = divmod(real, 256 ** 4)
        e, real = divmod(real, 256 ** 3)
        f, real = divmod(real, 256 ** 2)
        g, h = divmod(real, 256)

        return b"".join([char(a), char(b), char(c), char(d), char(e), char(f),
                         char(g), char(h)])

    if sys.version_info < (3, 0):
        @staticmethod
        def type_check(real):
            return type(real) is long
    else:
        @staticmethod
        def type_check(real):
            return type(real) is int

    @staticmethod
    def unsanitise(raw):
        a, b, c, d, e, f, g, h = raw

        return ((ordi(a) * 256 ** 7) + (ordi(b) * 256 ** 6) +
                (ordi(c) * 256 ** 5) + (ordi(d) * 256 ** 4) +
                (ordi(e) * 256 ** 3) + (ordi(f) * 256 ** 2) +
                (ordi(g) * 256) + ordi(h))


#------------------------------------------------------- Float helper functions
def float_sanitise(real):
    raw = str(abs(real))

    if "e" in raw:
        raw, ex = raw.split("e")
        ex = char(int(ex) + 127)
    else:
        ex = False

    if "." in raw:
        ex = 0
        d = False

        for i in raw:
            if i not in ["0", "."]:
                d = True

            if i != "." and d:
                ex += 1

            if (i == "." or ex < 0) and d:
                break

            if i in [".", "0"] and not d:
                ex -= 1

        if real < 0:
            if ex < 0:
                ex = 255 + ex
            else:
                ex = 127 - ex
        else:
            ex += 127

        ex = char(ex)
        w, d = raw.split(".")
    else:
        w = raw
        d = ""

    n = "".join([w, d]).lstrip("0")[:9]
    n = n.ljust(9, "0")
    n = int(n)

    # Convert to negative offset
    if real < 0:
        n = -n
        neg = b"\x00"
    else:
        neg = b"\x01"

    return n, neg, ex


def float_unsanitise(n, neg, ex):
    # Calc exponent
    if neg == 0:
        if ex > 128:
            ex -= 255
        else:
            ex = 127 - ex
    else:
        ex -= 127

    n = list(str(n))

    # Build exp and data
    if neg == 0:
        off = 1
    else:
        off = 0

    if ex < 0:
        for _ in range(0, abs(ex)):
            n.insert(off, "0")

        n.insert(abs(ex) + off, ".")
    elif ex > 0:
        while len(n) + off < ex:
            n.append("0")

        n.insert(ex + off, ".")

    return float("".join(n))


class FLOAT(DType):
    """
    FLOAT, single float, data type for HTD
    - Allows values of type float
    - Takes up 6 bytes of storage
    - Uses the identifier \x1e
    """
    can_range = True
    idn = b"\x1e"
    size = 6

    @staticmethod
    def sanitise(real):
        if real == 0.0:
            return b"\x01\x00\x00\x00\x00\x00"

        # Convert to raw int
        n, neg, ex = float_sanitise(real)

        # Convert to hex
        n += 2147483648

        a, n = divmod(n, 256 ** 3)
        b, n = divmod(n, 256 ** 2)
        c, d = divmod(n, 256)

        return b"".join([neg, ex, char(a), char(b), char(c), char(d)])

    @staticmethod
    def type_check(real):
        return type(real) is float

    @staticmethod
    def unsanitise(raw):
        if raw == b"\x01\x00\x00\x00\x00\x00":
            return 0.0

        neg, ex, a, b, c, d = raw

        n = ((ordi(a) * 256 ** 3) + (ordi(b) * 256 ** 2) + (ordi(c) * 256) +
             ordi(d) - 2147483648)

        return float_unsanitise(n, ordi(neg), ordi(ex))


class DFLOAT(DType):
    """
    FLOAT, double float, data type for HTD
    - Allows values of type float
    - Takes up 10 bytes of storage
    - Uses the identifier \x1f
    """
    can_range = True
    idn = b"\x1f"
    size = 10

    @staticmethod
    def sanitise(real):
        if real == 0.0:
            return b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00"

        # Convert to raw int
        n, neg, ex = float_sanitise(real)

        # Convert to hex
        n += 9223372036854775808

        a, n = divmod(n, 256 ** 7)
        b, n = divmod(n, 256 ** 6)
        c, n = divmod(n, 256 ** 5)
        d, n = divmod(n, 256 ** 4)
        e, n = divmod(n, 256 ** 3)
        f, n = divmod(n, 256 ** 2)
        g, h = divmod(n, 256)

        return b"".join([neg, ex, char(a), char(b), char(c), char(d), char(e),
                         char(f), char(g), char(h)])

    @staticmethod
    def type_check(real):
        return type(real) is float

    @staticmethod
    def unsanitise(raw):
        if raw == b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return 0.0

        neg, ex, a, b, c, d, e, f, g, h = raw

        n = ((ordi(a) * 256 ** 7) + (ordi(b) * 256 ** 6) +
             (ordi(c) * 256 ** 5) + (ordi(d) * 256 ** 4) +
             (ordi(e) * 256 ** 3) + (ordi(f) * 256 ** 2) +
             (ordi(g) * 256) + ordi(h)) - 9223372036854775808

        return float_unsanitise(n, ordi(neg), ordi(ex))


Register.register_dtype(BYTE)
Register.register_dtype(UBYTE)
Register.register_dtype(SHORT)
Register.register_dtype(USHORT)
Register.register_dtype(INTEGER)
Register.register_dtype(UINTEGER)
Register.register_dtype(LONG)
Register.register_dtype(ULONG)
Register.register_dtype(FLOAT)
Register.register_dtype(DFLOAT)


#------------------------------------------------- Python 3 specific data types
if sys.version_info < (3, 0):
    class UTF8(DType):
        """
        UTF8 data type for HTD (for python2)
        - Allows values of type unicode
        - Uses the identifier #
        """
        idn = b"#"
        can_range = False

        @staticmethod
        def sanitise(real):
            return real.encode("utf8")

        size = 0

        @staticmethod
        def type_check(real):
            return type(real) is unicode

        @staticmethod
        def unsanitise(raw):
            return raw.decode("utf8")

    Register.register_dtype(UTF8)

    class UTF16(DType):
        """
        UTF16 data type for HTD (for python2)
        - Allows values of type unicode
        - Uses the identifier $
        """
        idn = b"$"
        can_range = False

        @staticmethod
        def sanitise(real):
            return real.encode("utf16")

        size = 0

        @staticmethod
        def type_check(real):
            return type(real) is unicode

        @staticmethod
        def unsanitise(raw):
            return raw.decode("utf16")

    Register.register_dtype(UTF16)

    class UTF32(DType):
        """
        UTF32 data type for HTD (for python2)
        - Allows values of type unicode
        - Uses the identifier %
        """
        idn = b"%"
        can_range = False

        @staticmethod
        def sanitise(real):
            return real.encode("utf32")

        size = 0

        @staticmethod
        def type_check(real):
            return type(real) is unicode

        @staticmethod
        def unsanitise(raw):
            return raw.decode("utf32")

    Register.register_dtype(UTF32)


#----------------------------------------------------------- Preserve data type
class PRESERVE(DType):
    """
    PRESERVE data type for HTD
    - Allows any value that is valid in any registered data type
    - Uses the identifier \x0b
    """
    idn = b"\x0b"
    can_range = False

    @staticmethod
    def sanitise(real):
        for code in Register.dtype_codes:
            data_type = getattr(DataType, Register.dtype_codes[code])
            if (isinstance(data_type, PRESERVE) is False and
                    data_type.type_check(real)):
                try:
                    raw = b"".join([code, data_type.sanitise(real)])
                except (ValueError, OverflowError):
                    continue
                else:
                    return raw

        raise ValueError("Unknown type to preserve (%s)" % type(real))

    size = 0

    @staticmethod
    def type_check(real):
        for code in Register.dtype_codes:
            data_type = getattr(DataType, Register.dtype_codes[code])

            if (isinstance(data_type, PRESERVE) is False and
                    data_type.type_check(real)):
                return True

        return False

    @staticmethod
    def unsanitise(raw):
        code = raw[0:1]
        raw = raw[1:]

        if code in Register.dtype_codes:
            data_type = getattr(DataType, Register.dtype_codes[code])
        else:
            raise TypeError("Could not restore type.")

        return data_type.unsanitise(raw)


Register.register_dtype(PRESERVE)
Register.default_dtype = PRESERVE
