#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .. import ConnectionError, IntegrityError, Register, __version__
from .._base import convert_location, perm, ThreadLock
from .._base.config import Config
from .._base.data import sanitise, unsanitise
from .._base.trace import trace
from .._objects.attribute import Attribute
from weakref import ref


class Attributes(ThreadLock):
    """ Class that manages attribute settings in a table """
    def __init__(self, parent, location, table, file_obj=None):
        # Open file
        if file_obj is None:
            htdfile = Register.resolve_fhandler(location)
            self._file = htdfile(location, parent._access)
            self._opened = True
        else:
            self._file = file_obj
            self._opened = False

        # Set vars
        self.alive = True
        self.table = table

        self._parent = ref(parent)
        self.connection._children.append(self)

        self._access = self.connection._access  # Load user access

        ThreadLock.__init__(self)

    @trace
    def __contains__(self, key):
        """ Return True if attribute exists in db """

        self._file.lock()

        perm("r", self._access)

        out = False
        if key in self._file.header["ATT"]:
            out = True

        self._file.unlock()

        return out

    @trace
    def __delitem__(self, key):
        """ Delete attribute and data from db """

        self._file.lock(exclusive=True)

        perm("d", self._access)

        if key not in self._file.header["ATT"]:  # Check att exists
            raise KeyError('Could not find attribute to delete. (%s)' % key)

        if key in self._file.header["KEY"]:  # Check att not identifier
            raise ValueError('Cannot remove attribute that is an identifier')

        # Check att not in relationship
        if key in self._file.header["R1M"] or key in self._file.header["RM1"]:
            raise IntegrityError("Relationship(s) must be broken first")

        leng = len(self._file)  # Get number of records

        if leng > 0:  # Make temporary file if records exist
            tmp_file = self._make_tmp_file(self._file, self.location)

            # Truncate main file here because delete requires correct header
            list(self._file.delete(0, leng))

        header = self._file.header  # Get header info
        i = header["ATT"][key]._id  # Save ID to resort IDs later
        attribute = header["ATT"][key]  # Save Attribute for conv funcs later

        del header["ATT"][key]  # Remove attribute from header

        # Resort IDs
        for a in header["ATT"]:
            att = header["ATT"][a]
            if att._id > i:
                att._id -= 1

        self._file.header = header  # Save header changes

        if leng > 0:  # If records then read and convert data from copied table
            # Attribute is not selected in iterator so can just return record
            def conv(record, attribute):
                return record

            # Read from old, write to new
            self._run_data(tmp_file, conv, attribute, leng)

        self._file.unlock()

    @trace
    def __getitem__(self, key):
        """ Return attributes settings """

        self._file.lock()

        perm("r", self._access)

        attributes = self._file.header["ATT"]

        if key not in attributes:
            raise KeyError('Failed to find sanitise settings for attribute '
                           '(%s)' % (key))

        self._file.unlock()

        return attributes[key]

    @trace
    def __iter__(self):
        locked = self._acquire_lock()
        self._file.lock()

        perm("r", self._access)
        i = 0

        self._file.unlock()

        if locked:
            self._release()

        while 1:
            locked = self._acquire_lock()
            self._file.lock()

            try:
                key = self._file.header["ATTBYID"][i]["name"]
            except KeyError:
                raise StopIteration
            else:
                yield self._file.header["ATT"][key]
            finally:
                self._file.unlock()

                if locked:
                    self._release()

            i += 1

    @trace
    def __len__(self):
        """ The number of attributes in the table """

        self._file.lock()

        perm("r", self._access)

        leng = len(self._file.header["ATT"])

        self._file.unlock()

        return leng

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Attributes>" % __version__

        return "<HTD%s Attributes (%s)>" % (__version__, self.location)

    @trace
    def __setitem__(self, key, attribute):
        """ Add/Update the attribute key with the attribute """

        # Check the parameters are valid
        if type(attribute) is not Attribute:
            raise ValueError("Attribute must be a Attribute instance.")

        if key is None:
            raise ValueError("Attribute name cannot be NULL.")

        attribute.name = key  # Runs name check

        self._file.lock(exclusive=True)

        # Get number of records, header info and if the attribute exists
        leng = len(self._file)

        header = self._file.header
        exists = key in header["ATT"]

        if exists:
            # Overwrite current attribute
            perm("o", self._access)

            # Check attribute not in relationship
            if (key in self._file.header["R1M"] or
                    key in self._file.header["RM1"]):
                raise IntegrityError("Relationship(s) must be broken first")

            # Set the id from the old attribute to the new
            attribute._id = header["ATT"][key]._id

            # Confirm old data
            attr = {key: attribute}

            for i in self._file.select(key, 0, leng):
                sanitise(unsanitise(i, attr), attr)
        else:
            # Append a new attribute, set id to maxid + 1
            perm("a", self._access)
            attribute._id = len(header["ATT"])

        if leng > 0:  # Make temporary file if records exist
            tmp_file = self._make_tmp_file(self._file, self.location)

            # Truncate main file here because delete requires correct header
            list(self._file.delete(0, leng))

        header["ATT"][key] = attribute  # Write the changes to the tmp header
        self._file.header = header  # Write header changes

        if leng > 0:  # If records then read and convert data from copied table
            if exists:
                def conv(record, attribute):
                    # Convert the data to the new attribute settings
                    name = attribute.name
                    atts = {name: attribute}
                    data = sanitise(unsanitise({name: record[name]}, atts),
                                    atts)
                    record[name] = data[name]

                    return record
            else:
                def conv(record, attribute):
                    record[attribute.name] = None  # Set new attribute to NULL

                    return record

            # Read from old, write to new
            self._run_data(tmp_file, conv, attribute, leng)

        self._file.unlock()

    @staticmethod
    def _make_tmp_file(file, location):
        """ Make a copy of the table at the location to use for conversion """

        tmp_loc = convert_location(location, ".~", "backup.old")

        tmp_file = file.copy(tmp_loc, replace=False)
        tmp_file.flush(Config.force_sync)

        return tmp_file

    def _run_data(self, tmp_file, conv, attribute, leng):
        """
        Clear data in file
        Iterate through _gen_data using tmp_file to insert the data into file
        """

        tmp_file.lock()  # Lock tmp_file for reading

        # Create iterator to insert
        gen = self._file.insert()
        next(gen)

        # Insert data into old
        for record in tmp_file.select(self._file.header["ATT"].keys(), 0,
                                      leng):
            gen.send(conv(record, attribute))

        # Flush file and delete temporary file
        self._file.flush(Config.force_sync)
        tmp_file.close(delete_files=True)

        return True

    @trace
    def close(self):
        self.alive = False

        if self._opened:  # Close file if I opened it
            self._file.close()

        # Remove from parent
        try:
            if self in self.connection._children:
                self.connection._children.remove(self)
        except ConnectionError:
            pass

        return True

    @property
    def connection(self):
        """ Return the parent or None """

        par = self._parent()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @trace
    def index(self, attribute):
        """ Retrieve the index of a specific attribute """

        self._file.lock()

        perm("r", self._access)

        if attribute not in self._file.header["ATT"]:
            raise KeyError("Attribute (%s) doesn't not exist." % attribute)

        i = self._file.header["ATT"][attribute]._id

        self._file.unlock()

        return i

    @trace
    def keys(self):
        """ List the keys (attribute names) in the table """

        self._file.lock()

        perm("r", self._access)

        l = [a["name"] for a in self._file.header["ATTBYID"].values()]

        self._file.unlock()

        return l

    @property
    def location(self):
        """ The location of the file """

        return self._file.location

    @trace
    def rename(self, old, new):
        """ Rename the attribute `old` to `new` """

        self._file.lock(exclusive=True)

        # Check old exists and new doesn't
        if new in self._file.header["ATT"]:
            raise KeyError("Attribute name already exists")

        if old not in self._file.header["ATT"]:
            raise KeyError("Could not find attribute to rename (%s)" % old)

        # Run name test
        tmp = Attribute()
        tmp.name = new

        header = self._file.header

        # Update attribute
        header["ATT"][new] = header["ATT"][old]
        del header["ATT"][old]

        # Update identifier
        if old in header["KEY"]:
            header["KEY"][header["KEY"].index(old)] = new

        # Update relationships
        if old in header["R1M"]:
            for foreign_table, foreign_att in header["R1M"][old]:
                cur = self.connection[foreign_table]
                cur._file.lock(exclusive=True)

                tmp_header = cur._file.header
                tmp_header["RM1"][foreign_att].remove([self.table, old])
                tmp_header["RM1"][foreign_att].append([self.table, new])
                cur._file.header = tmp_header

                cur._file.unlock()
                cur.close()

            header["R1M"][new] = header["R1M"][old]
            del header["R1M"][old]

        if old in header["RM1"]:
            for foreign_table, foreign_att in header["RM1"][old]:
                cur = self.connection[foreign_table]
                cur._file.lock(exclusive=True)

                tmp_header = cur._file.header
                tmp_header["R1M"][foreign_att].remove([self.table, old])
                tmp_header["R1M"][foreign_att].append([self.table, new])
                cur._file.header = tmp_header

                cur._file.unlock()
                cur.close()

            header["RM1"][new] = header["RM1"][old]
            del header["RM1"][old]

        self._file.header = header

        self._file.unlock()

        return True

    @property
    def version(self):
        """ Version of the open table """

        self._file.lock()

        out = self._file.header["VER"]

        self._file.unlock()

        return out
