#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .. import ConnectionError, Register, __version__
from .._base import perm, ThreadLock
from .._base.trace import trace
from .._objects.user import User
from weakref import ref


class Auth(ThreadLock):
    """ Authentication object manages users and access for a table """
    def __init__(self, parent, location, file_obj=None):
        # Open file
        if file_obj is None:
            htdfile = Register.resolve_fhandler(location)
            self._file = htdfile(location, parent._access)  # Open file
            self._opened = True
        else:
            self._file = file_obj
            self._opened = False

        # Store variables
        self.alive = True
        self._parent = ref(parent)
        self.connection._children.append(self)

        self._access = parent._access  # Load access

        ThreadLock.__init__(self)  # Start inherit

    @trace
    def __contains__(self, key):
        self._file.lock()

        perm("rx", self._access)

        out = key in self._file.header["USR"]

        self._file.unlock()

        return out

    @trace
    def __delitem__(self, key):
        self._file.lock(exclusive=True)

        perm("dx", self._access)

        if key not in self._file.header["USR"]:
            raise KeyError('No User, "%s", found to delete' % key)

        if len(self._file.header["USR"]) == 1:
            raise ValueError("Must be at least 2 users to delete a user")

        if self.connection.user == key:
            raise ValueError("Cannot delete your own user.")

        users = self._file.header["USR"]
        del users[key]

        x = sum(1 for i in users.values() if i.access == "roadx")

        if x < 1:
            raise ValueError("At least 1 remaining user must have full access")

        # Update header
        header = self._file.header
        header["USR"] = users

        self._file.header = header

        self._file.unlock()

    @trace
    def __getitem__(self, key):
        self._file.lock()

        perm("rx", self._access)

        try:
            usr = self._file.header["USR"][key]
        except KeyError:
            usr = False
            raise KeyError('No User, "%s", found' % key)

        usr.username = key  # Load username

        self._file.unlock()

        return usr

    @trace
    def __iter__(self):
        locked = self._acquire_lock()
        self._file.lock()

        perm("rx", self._access)

        user_list = [u for u in self._file.header["USR"]]
        user_list.sort()

        self._file.unlock()

        if locked:
            self._release()

        for user in user_list:
            if locked:
                self._acquire()

            self._file.lock()

            try:
                obj = self._file.header["USR"][user]
            except KeyError:
                continue
            else:
                yield obj
            finally:
                self._file.unlock()

                if locked:
                    self._release()

        raise StopIteration

    @trace
    def __len__(self):
        self._file.lock()

        perm("rx", self._access)

        out = len(self._file.header["USR"])

        self._file.unlock()

        return out

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Auth>" % __version__

        return "<HTD%s Auth (%s)>" % (self.version, self.location)

    @trace
    def __setitem__(self, key, v):
        if isinstance(v, User) is False:
            raise ValueError("User must be an User instance.")

        self._file.lock(exclusive=True)

        v.username = key

        if key in self._file.header["USR"]:
            perm("ox", self._access)

            if len(self._file.header["USR"]) == 1 and v.access != "roadx":
                raise ValueError("At least one User must have full access")
        else:
            perm("ax", self._access)

        header = self._file.header
        header["USR"][key] = v

        self._file.header = header

        self._file.unlock()

    @trace
    def close(self):
        """ Close the cursor """

        self.alive = False  # Set closed

        if self._opened:  # Close file if I opened it
            self._file.close()

        # Remove from parent
        try:
            if self in self.connection._children:
                self.connection._children.remove(self)
        except ConnectionError:
            pass

        return True

    @property
    def connection(self):
        """ Return the parent or None """

        par = self._parent()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @trace
    def index(self, username):
        """ Retrieve the index of a specific username """

        self._file.lock()

        perm("rx", self._access)

        if username not in self._file.header["USR"]:
            raise KeyError("User (%s) doesn't not exist." % username)

        l = [i for i in self._file.header["USR"]]
        l.sort()
        i = l.index(username)

        self._file.unlock()

        return i

    @trace
    def keys(self):
        """ List the keys (attribute names) in the table """

        self._file.lock()

        perm("rx", self._access)

        l = [a for a in self._file.header["USR"]]
        l.sort()

        self._file.unlock()

        return l

    @property
    def location(self):
        """ The location of the file """

        return self._file.location

    @property
    def version(self):
        """ Version of the open table """

        self._file.lock()

        out = self._file.header["VER"]

        self._file.unlock()

        return out
