#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .._objects.auth import Auth
from .._objects.cursor import Cursor
from .._objects.user import Authenticate, User
from .._base import ThreadLock, perm
from .._base.trace import trace
from .. import (__version__, AuthenticationError, ConnectionError,
                DatabaseError, IntegrityError, Register)
from copy import copy


class Connection(ThreadLock):
    """ Class that provides a connection to a table or db """
    def __init__(self, location, user=None, file_obj=None):
        ThreadLock.__init__(self)  # Load subclasses

        # Set vars
        self.alive = True
        self._children = []

        # Confirm location is str
        if not isinstance(location, str):
            self.alive = False
            raise ValueError("location (%s) must be a str object." % location)

        # Load user
        if user is None:
            user = Authenticate("guest", None)

        self.user = user.username

        if file_obj is None:  # Open file if not passed
            htdfile = Register.resolve_fhandler(location)
            file_obj = htdfile(location, "r")

        self._file = file_obj
        header = self._file.header

        # Load db mode
        self._isdb = False

        if header["TAB"] is False:
            self._isdb = True
        elif header["TAB"]:
            self.close()
            raise ConnectionError("This table is part of a db, please connect "
                                  "to the db")

        # Load and check auth
        if (user.username not in header["USR"] or
                header["USR"][user.username].password != user.password):
            self.close()
            raise AuthenticationError()

        # Load access
        self._file.access = header["USR"][user.username].access
        self._access = header["USR"][user.username].access

    @trace
    def __contains__(self, name):
        """ Return if the key is in the database """

        perm("r", self._access)

        if not self.is_db:
            raise ValueError("This feature requires a connection to a "
                             "database")

        cur = self[None]
        exists = name in cur
        cur.close()

        return exists

    @trace
    def __delitem__(self, name):
        """ Split table with parent - returns connection to child """

        perm("xd", self._access)

        if not self.is_db:
            raise ValueError("This feature requires a connection to a "
                             "database")

        # Check no relationships
        cur = self[name]

        if (len(cur._file.header["R1M"]) > 0 or
                len(cur._file.header["RM1"]) > 0):
            raise IntegrityError("Relationship(s) must be broken first")

        cur.close()

        cur = self[None]

        if name in cur:
            # Split table from parent
            sub_cur = self[name]
            del cur[name]

            # Update header
            sub_cur._file.lock(exclusive=True)

            header = sub_cur._file.header
            header["TAB"] = None

            sub_cur._file.header = header

            sub_cur._file.unlock()

            sub_cur.close()
            cur.close()
        else:
            cur.close()
            raise KeyError("No table to split named '%s'." % name)

    @trace
    def __getitem__(self, key):
        """ Open a cursor on the table """

        f = self._file

        if f.persistent:
            f = None

        if key is None:
            return Cursor(self, self.location, None, file_obj=f)

        # Get table data from db
        cur = Cursor(self, self.location, None, file_obj=f)
        tbd = cur[key]
        cur.close()

        return Cursor(self, tbd['location'], key)

    @trace
    def __iter__(self):
        """ If a db - return the tables """

        perm("r", self._access)

        # Check connections is DB
        if not self._isdb:
            raise StopIteration

        # Open cursor, list table, close
        cur = self[None]

        for record in cur.select(["name", "location"]):
            yield record

        cur.close()

    def __len__(self):
        # Check for db mode
        if self._isdb:
            cur = self[None]
            leng = len(cur)
            cur.close()
            return leng
        else:
            return 1

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Connection>" % __version__

        return ("<HTD%s Connection (%s@%s), DB Mode %s>" %
                (__version__, self.user, self.location, self.is_db))

    @trace
    def __setitem__(self, name, child):
        """ Join table (connection) with parent """

        perm('xa', self._access)

        if not self.is_db:
            raise DatabaseError("Must be connected to a DB to set a child")

        if isinstance(child, self.__class__) is False:
            raise ValueError("Child must be a Connection instance")

        cur = child[None]

        # Check that child is a table
        cur._file.lock()
        header = cur._file.header
        cur._file.unlock()

        cur.close()

        if header["TAB"] is not None:
            raise ValueError("Child must be a table")

        # Add location to parent
        cur = self[None]

        if name in cur:  # If already exists check no relationships linked
            old_cur = self[name]

            if (len(old_cur._file.header["R1M"]) > 0 or
                    len(old_cur._file.header["RM1"]) > 0):
                old_cur.close()
                raise IntegrityError("Relationship(s) must be broken first")

            old_cur.close()

        cur[name] = {"location": child.location}
        cur.close()

        child.close()

        cur = self[name]

        # Update child header
        cur._file.lock(exclusive=True)

        header = cur._file.header
        header["TAB"] = True
        header["USR"] = {"guest": User(None, "roadx")}  # Clear users

        cur._file.header = header

        cur._file.unlock()

        cur.close()

        return self

    @property
    @trace
    def authentication(self):
        """ Create a new authentication object """

        perm("x", self._access)

        f = self._file

        if f.persistent:
            f = None

        return Auth(self, self.location, file_obj=f)

    @property
    @trace
    def children(self):
        """ Open children """

        for cur in copy(self._children):
            if cur is None:
                self._children.remove(cur)
            else:
                if not cur.alive:
                    del self._children._parent
                    self._children.remove(cur)

        return self._children

    @trace
    def close(self):
        """ Close the connection """

        for cur in copy(self.children):
            cur.close()  # Close children

        self._file.close()

        self.alive = False

        return True

    @trace
    def delete(self, children=False):
        """ Delete connection db (and tables) and disconnect """

        perm("d", self._access)

        for cur in copy(self.children):
            cur.close()  # Close cursors

        if self.is_db and children:
            for i in list(self):
                # Delete children
                htdfile = Register.resolve_fhandler(i["location"])
                f = htdfile(i["location"], self._access)
                f.close(delete_files=True)

        self._file.close(delete_files=True)  # Delete main file

        self.alive = False

        return True

    @property
    def is_db(self):
        """ Is the connection to a DB or table """

        return self._isdb

    @trace
    def link(self, name1, name2):
        """ Link tbl1.att1 to tbl2.att2 """

        perm("x", self._access)

        (many, att), (one, key) = self._relation_check(name1, name2)

        one._file.lock(exclusive=True)
        many._file.lock(exclusive=True)

        # Update many table
        header = many._file.header

        if att[1] not in header["RM1"]:
            header["RM1"][att[1]] = []

        if key not in header["RM1"][att[1]]:
            header["RM1"][att[1]].append(key)

        many._file.header = header

        # Update one table
        header = one._file.header

        if key[1] not in header["R1M"]:
            header["R1M"][key[1]] = []

        if att not in header["R1M"][key[1]]:
            header["R1M"][key[1]].append(att)

        one._file.header = header

        one._file.unlock()
        many._file.unlock()

        return True

    @property
    def location(self):
        """ The location of the file """

        return self._file.location

    @trace
    def keys(self):
        """ List the table names in the db """

        if not self._isdb:
            raise ValueError("This feature requires a connection to a "
                             "database")

        cur = self[None]

        for i in cur.keys():
            yield i

        cur.close()

    @property
    @trace
    def relation(self):
        """ List relations in db returns [(many, one)] """

        perm("r", self._access)

        for table in self.keys():
            cur = self[table]

            cur._file.lock()

            for rel in cur._file.header["RM1"]:
                for item in cur._file.header["RM1"][rel]:
                    yield ((table, rel), tuple(item))

            cur._file.unlock()

            cur.close()

    def _relation_check(self, name1, name2):
        """ Check that two attributes can form a relationship """

        if not self._isdb:
            raise ValueError("This feature requires a connection to a "
                             "database")

        for i in (name1, name2):
            if "." not in i:
                raise ValueError("The table and attribute split by a '.' is "
                                 "required (%s)" % i)

        # Load attributes
        tbl1, att1 = name1.rsplit(".", 1)
        tbl2, att2 = name2.rsplit(".", 1)

        for i in (tbl1, tbl2):
            if i not in self.keys():  # Check tables exist
                raise ValueError("No table found with name (%s)" % i)

        # Connect to tables
        cur1 = self[tbl1]
        cur2 = self[tbl2]

        # Check one is KEY and one is ATT
        key = None
        att = None

        for a, t, c in [(att1, tbl1, cur1), (att2, tbl2, cur2)]:
            if c.composite:
                att = [t, a]
            else:
                if a == c.identifier:
                    key = [t, a]
                else:
                    att = [t, a]

        cur1.close()
        cur2.close()

        if key is None:
            raise ValueError("One link must be to a Primary Key")
        else:
            one = self[key[0]]

        if att is None:
            raise ValueError("One link must be to an Attribute")
        else:
            many = self[att[0]]

        # Check attributes are the same
        tmp1 = one._file.header["ATT"][key[1]].to_dict()
        tmp2 = many._file.header["ATT"][att[1]].to_dict()

        del tmp1["id"]  # Strip IDs (could be different)
        del tmp2["id"]

        if tmp1 != tmp2:
            raise ValueError("Attributes to link, must be the same.")

        # Check 'many' values are in '1'
        for i in many.select(att[1]):
            if i not in one:
                raise KeyError("Value in child didn't exist in parent table "
                               "(%s)." % i)

        return (many, att), (one, key)

    @trace
    def unlink(self, name1, name2):
        """ Unlink tbl1.att1 to tbl2.att2 """

        perm("x", self._access)

        (many, att), (one, key) = self._relation_check(name1, name2)

        one._file.lock(exclusive=True)
        many._file.lock(exclusive=True)

        # Update many table
        header = many._file.header

        if att[1] in header["RM1"] and key in header["RM1"][att[1]]:
            header["RM1"][att[1]].remove(key)

            if len(header["RM1"][att[1]]) == 0:
                del header["RM1"][att[1]]

        many._file.header = header

        # Update one table
        header = one._file.header

        if key[1] in header["R1M"] and att in header["R1M"][key[1]]:
            header["R1M"][key[1]].remove(att)

            if len(header["R1M"][key[1]]) == 0:
                del header["R1M"][key[1]]

        one._file.header = header

        one._file.unlock()
        many._file.unlock()

        return True


# def load_connection(*args, **kwargs):
#    return Connection(*args, **kwargs)
