#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .. import ConnectionError, IntegrityError, Register, __version__
from .._base import perm, ThreadLock
from .._base.data import (get_key_as_dict, get_key_from_data, sanitise,
                          sanitise_key, unsanitise, update_record,
                          where_compile)
from .._base.progress import Progress, ProgressItem
from .._base.trace import trace
from .._objects.attributes import Attributes
from .._objects.record import Record
from threading import current_thread
from weakref import ref


def integrity_check(data, header, parent):
    """ Relationship check """

    if len(header["RM1"]) == 0:
        return

    for att in header["RM1"]:
        for tbl, foreign_att in header["RM1"][att]:
            cur = parent[tbl]

            if data[att] not in cur:
                cur.close()
                raise IntegrityError("Value (%s) not in foreign table (%s)"
                                     % (data[att], tbl))

            cur.close()


def parse_record(record):
    """ Copy record, preserving version, so that users isn't modified """

    if type(record) is Record:
        version = record._version
    else:
        version = None

    record = dict(record or {})
    record[".version"] = version

    return record


class Cursor(ThreadLock):
    """ Class that manages a cursor on a table """
    def __init__(self, parent, location, table, header=None, file_obj=None):
        # Open file
        if file_obj is None:
            htdfile = Register.resolve_fhandler(location)
            self._file = htdfile(location, parent._access, header)
            self._opened = True
        else:
            self._file = file_obj
            self._opened = False

        # Store variables
        self.alive = True
        self.table = table

        self._parent = ref(parent)
        self.connection._children.append(self)

        ThreadLock.__init__(self)  # Start inherit

    @trace
    def __contains__(self, key):
        self._file.lock()  # Lock file

        hex_key = sanitise_key(key, self._file.header)  # Convert key to hex
        exists = hex_key in self._file  # Check if the key exists

        self._file.unlock()  # Unlock

        return exists

    @trace
    def __delitem__(self, key):
        self._file.lock(exclusive=True)

        if len(self._file.header["R1M"]) > 0:
            err = False

            for att in self._file.header["R1M"]:
                for tbl, foreign_att in self._file.header["R1M"][att]:
                    cur = self.connection[tbl]

                    if key in cur.select(foreign_att):
                        err = True
                        cur.close()
                        break

                    cur.close()

            if err:
                raise IntegrityError("Relationship must be broken first")

        if isinstance(key, slice):  # delete a slice
            start = key.start

            if start is None:
                start = 0

            stop = key.stop

            if stop is None:
                stop = len(self._file)

            if start >= stop:
                raise ValueError("Stop index must be greater than start.")

            list(self._file.delete(start=start, stop=stop))

            self._file.unlock()
        else:
            hex_key = sanitise_key(key, self._file.header)  # Conv key to hex

            try:
                del self._file[hex_key]
            except KeyError:
                raise KeyError('No record, %s, was found to delete.' %
                               str(key))
            finally:
                self._file.unlock()

    @trace
    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.select(start=key.start or 0, stop=key.stop or False)

        self._file.lock()  # Lock file

        hex_key = sanitise_key(key, self._file.header)  # Convert key to hex

        try:
            data = self._file[hex_key]
        except KeyError:
            raise KeyError('No record, %s, was found.' % str(key))

        data = unsanitise(data, self._file.header["ATT"])

        self._file.unlock()  # Unlock

        return Record(data, con=self.connection, table=self.table)

    @trace
    def __iter__(self):
        """ If a db - return the tables """

        for record in self.select():
            yield record

    @trace
    def __len__(self):
        """ The number of records in the table """

        self._file.lock()

        length = len(self._file)  # Get length

        self._file.unlock()

        return length

    def __repr__(self):
        if self.alive is False:
            return "<HTD%s Closed Cursor>" % __version__

        return "<HTD%s Cursor (%s)>" % (__version__, self.location)

    @trace
    def __setitem__(self, key, data):
        self._file.lock(exclusive=True)

        key = get_key_as_dict(key, self._file.header["KEY"])
        atts = self._file.header["ATT"]

        data = parse_record(data)
        data.update(key)  # Add key to data

        exists, index, line = self._file.is_record(sanitise(key, atts))

        if exists:  # Record exists get current data
            line = unsanitise(line, atts)
            data = update_record(data, line)

        integrity_check(data, self._file.header, self.connection)
        data = sanitise(data, atts)  # Run filters

        if exists:
            self._file[index] = data
        else:
            self._file.insert_record(index, data)

        self._file.unlock()

    @property
    def attributes(self):
        """ Manage the attribute settings """

        if self._opened:
            htdfile = None
        else:
            htdfile = self._file

        return Attributes(self.connection, self.location, self.table, htdfile)

    @trace
    def close(self):
        """ Close the cursor """

        self.alive = False  # Set closed

        if self._opened:  # Close file if I opened it
            self._file.close()

        # Remove from parent
        try:
            if self in self.connection._children:
                self.connection._children.remove(self)
        except ConnectionError:
            pass

        return True

    @property
    def composite(self):
        """ Is the table using a composite key """

        return isinstance(self.identifier, (list, tuple))

    @property
    def connection(self):
        """ Return the parent or None """

        par = self._parent()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @trace
    def convert(self, location, file_format):
        """ Convert this file to a new file in a different format """

        self._file.lock()

        if file_format not in Register.fhandlers:  # Check file_format exists
            raise ValueError("htd_file has not been registered.")

        # Copy header and break relationships
        header = self._file.header

        header["TAB"] = None
        header["R1M"] = {}
        header["RM1"] = {}

        # Create new file
        new_file = file_format(location, self.connection._access, header)
        new_file.lock(exclusive=True)

        inserter = new_file.insert()
        next(inserter)

        # Copy data into new file from old
        for record in self._file.select(list(header["ATT"].keys()), 0,
                                        len(self._file)):
            inserter.send(record)

        new_file.close()

        self._file.unlock()

        return True

    @trace
    def copy(self, location, replace=False):
        """ Copy the table to a new location breaking relationships """

        self._file.lock()

        new_file = self._file.copy(location, replace)

        # Set to standalone table and clear relationships
        header = self._file.header

        header["TAB"] = None
        header["R1M"] = {}
        header["RM1"] = {}

        new_file.header = header

        new_file.close()

        self._file.unlock()

        return True

    @trace
    def delete(self, where=None, args={}):
        """ Delete items in the table which meet a criteria """

        self._file.lock(exclusive=True)

        if len(self._file.header["R1M"]) > 0:
            raise IntegrityError("Relationship must be broken first")

        if where is None:  # Load where
            c = sum(self._file.delete(0, len(self._file)))

            self._file.unlock()

            return c
        else:
            where = where_compile(where, args, self._file.header["ATT"])

        c = 0  # Count number of deleted items
        gen = self._file.delete(None, None)

        for record in gen:
            if where is not None:  # Check where clause
                if not eval(where, dict(record)):
                    gen.send(False)
                    continue

            gen.send(True)
            c += 1

        self._file.unlock()

        return c

    @property
    @trace
    def identifier(self):
        """ The identifier for the table """

        self._file.lock()

        perm('r', self._file.access)

        key = self._file.header["KEY"][:]

        if len(key) == 1:
            key = key[0]

        self._file.unlock()

        return key

    @identifier.setter
    @trace
    def identifier(self, key):
        """ Set the identifier for the table """

        self._file.lock(exclusive=True)

        perm("o", self._file.access)

        # Confirm data
        if isinstance(key, str):
            key = [key]

        if isinstance(key, (list, tuple)) is False:
            raise ValueError("Identifier must be a list or list of strings")

        if len(self._file) > 0:
            raise ValueError("The table must be empty before changing the"
                             " identifier.")

        # Check old attributes are not in relationship
        for k in self._file.header["KEY"]:
            if k in self._file.header["R1M"] or k in self._file.header["RM1"]:
                raise IntegrityError("Relationship(s) must be broken first")

        # Check new attributes are not in relationship
        for k in key:
            if k in self._file.header["R1M"] or k in self._file.header["RM1"]:
                raise IntegrityError("Relationship(s) must be broken first")

        header = self._file.header
        header["KEY"] = key

        self._file.header = header

        self._file.unlock()

    @trace
    def index(self, key):
        """ Retrieve the index of a key """

        self._file.lock()  # Lock file

        hex_key = sanitise_key(key, self._file.header)  # Convert key to hex

        try:
            index = self._file.index(hex_key)
        except KeyError:
            raise KeyError('No record, %s, was found.' % str(key))
        finally:
            self._file.unlock()  # Unlock

        return index

    @trace
    def insert(self, generator):
        """ Insert a generator into the table """

        locked = self._acquire_lock()

        self._file.lock(exclusive=True)

        gen = self._file.insert()  # Load file generator

        next(gen)

        self._file.unlock()

        if locked:
            self._release()

        for record in generator:  # Loop and send
            locked = self._acquire_lock()

            self._file.lock(exclusive=True)

            record = parse_record(record)
            integrity_check(record, self._file.header, self.connection)
            record = sanitise(record, self._file.header["ATT"])

            gen.send(record)

            self._file.unlock()

            if locked:
                self._release()

        gen.close()

    @trace
    def keys(self):
        """ List the keys in the table """

        return self.select(self._file.header["KEY"])

    @property
    def location(self):
        """ The location of the file """

        return self._file.location

    def lock_to_thread(self, thread=None):
        """ Lock the cursor to the thread """

        if thread is None:
            thread = current_thread()

        self._hold_thread_lock(thread)
        self._file.lock(exclusive=True, idn=thread)

    @trace
    def rename(self, location, replace=False):
        """ Rename the table to the new location """

        self._file.lock(exclusive=True)

        self._file.unlock()
        self._file.rename(location, replace)  # rename the file

        # Open the new renamed file
        htdfile = Register.resolve_fhandler(location)
        self._file = htdfile(location, self.connection._access)

        # Update the reference in the parent if db
        if self.connection.is_db:
            if self.table is not None:
                cur = self.connection[None]
                cur[self.table] = {"location": location}
                cur.close()
            else:
                self.connection._file.close()
                self.connection._file = htdfile(location,
                                                self.connection._access)

        return True

    @trace
    def select(self, attribute=None, start=0, stop=False, where=None, args={}):
        """ List selected attribute(s) """

        # Lock the thread
        locked = self._acquire_lock()

        self._file.lock()

        if attribute is None:  # Select all atts
            attbyid = self._file.header["ATTBYID"]
            attribute = [attbyid[i]["name"] for i in attbyid]

        if isinstance(attribute, str):
            attribute = [attribute]

        if not isinstance(attribute, (list, tuple)):
            raise ValueError("Attribute must be a list, tuple or str.")

        numatts = len(attribute)  # Count atts

        # Check atts exist
        if sum([1 for att in attribute if att not in
                self._file.header["ATT"]]) > 0:
            raise KeyError('No find attribute, %s, to list.' % attribute)

        if stop is False:  # load stop
            stop = len(self._file)

        if start > stop:
            raise ValueError("Stop index must be greater than start.")

        if where is not None:  # Load where
            where = where_compile(where, args, self._file.header["ATT"])

        gen = self._file.select(attribute, start, stop)  # load generator

        # Load progress
        calling = "Calling htd.Cursor.select"
        prog = ProgressItem(current_thread(), stop - start, calling)
        Progress.add(str(id(prog)), prog)

        for i in gen:
            if where is not None:  # Check where clause
                if not eval(where, dict(i)):
                    continue

            tmp = unsanitise(i, self._file.header["ATT"])

            self._file.unlock()

            if locked:
                self._release()

            prog.count()

            if numatts == 1:  # 1 att doesn't return record
                yield tmp[attribute[0]]
            else:
                yield Record(tmp, self.connection, self.table)

            locked = self._acquire_lock()

            self._file.lock()

        Progress.remove(str(id(prog)))

        self._file.unlock()

        if locked:
            self._release()

    def unlock_from_thread(self, thread=None):
        """ Unlock the cursor from the thread """

        if thread is None:
            thread = current_thread()

        self._release_thread_lock(thread)
        self._file.unlock(idn=thread)

    @trace
    def update(self, start=0, stop=False, where=None, args={}):
        """ Update a generator into the table """

        locked = self._acquire_lock()

        self._file.lock(exclusive=True)

        if stop is False:  # load stop
            stop = len(self._file)

        if where is not None:  # Load where
            where = where_compile(where, args, self._file.header["ATT"])

        gen = self._file.update(start, stop)

        # Load progress
        calling = "Calling htd.Cursor.update"
        prog = ProgressItem(current_thread(), stop - start, calling)
        Progress.add(str(id(prog)), prog)

        for old_record in gen:
            prog.count()

            if where is not None:  # Check where clause
                if not eval(where, dict(old_record)):
                    gen.send({})
                    continue

            old_record = Record(unsanitise(old_record,
                                           self._file.header["ATT"]),
                                self.connection, self.table)
            key = get_key_from_data(old_record, self._file.header["KEY"])

            self._file.unlock()

            if locked:
                self._release()

            record = (yield old_record)  # Send/Recieve new record

            locked = self._acquire_lock()

            self._file.lock(exclusive=True)

            record = parse_record(record)
            record = update_record(record, old_record)

            if key != get_key_from_data(record, self._file.header["KEY"]):
                raise KeyError("The key must remain the same.")

            integrity_check(record, self._file.header, self.connection)

            gen.send(sanitise(record, self._file.header["ATT"]))

            yield 1

        Progress.remove(str(id(prog)))

        self._file.unlock()

        if locked:
            self._release()

    @property
    @trace
    def version(self):
        """ Version of the open table """

        self._file.lock()

        out = self._file.header["VER"]

        self._file.unlock()

        return out
