#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .. import ConnectionError
from weakref import ref


class Record(dict):
    """ Class that provides read-only record and foreign key support """
    def __init__(self, data={}, con=None, table=None):
        # Version can come from attribute in data
        if ".version" in data:
            version = data[".version"]
            del data[".version"]
        else:
            version = None

        self._version = version
        self.table = table

        if con is not None:
            self._con = ref(con)
        else:
            self._con = None

        super(Record, self).__init__(data)

    def __getitem__(self, key):
        if "." not in key:  # Check if foreign key lookup needed
            return super(Record, self).__getitem__(key)

        # Split table and attribute
        key = key.split(".")
        table = key[-1]
        key = key[:-1]

        if self.table is None:
            raise ValueError("Record is not referenced to table.")

        # Check the object exists
        try:
            cur = self.connection[table]  # Open cursor
        except KeyError:
            raise KeyError("No table found (%s)" % table)

        # Load relationships
        r1m = cur._file.header["R1M"]
        rm1 = cur._file.header["RM1"]
        r1m_link = []
        rm1_link = []

        # Find relationships that match key.table
        for att in r1m:
            for tbl, foreign_att in r1m[att]:
                if tbl == self.table and foreign_att in key:
                    r1m_link.append((tbl, foreign_att))

        for att in rm1:
            for tbl, local_att in rm1[att]:
                if tbl == self.table and local_att in key:
                    rm1_link.append((tbl, att, local_att))

        # Check that relationships are valid
        if len(r1m_link) == 0 and len(rm1_link) == 0:
            cur.close()
            raise ValueError("Attribute (%s) has no relation to table (%s)." %
                             (key, table))

        if len(r1m_link) > 0 and len(rm1_link) > 0:
            cur.close()
            raise ValueError("Attribute (%s) has too many relationships to"
                             " table (%s)." % (key, table))

        if len(r1m_link) > 0:  # For 1-m relationships
            tbl, att = r1m_link[0]  # Pick first relation

            if not super(Record, self).__contains__(att):
                cur.close()
                raise ValueError("This record does not have the required "
                                 "attribute (%s), please add the attribute "
                                 "to the query." % att)

            data = cur[super(Record, self).__getitem__(att)]

        if len(rm1_link) > 0:  # For m-1 relationships
            tbl, att, local_att = rm1_link[0]  # Pick first relation

            if not super(Record, self).__contains__(local_att):
                cur.close()
                raise ValueError("This record does not have the required "
                                 "attribute (%s), please add the attribute "
                                 "to the query." % att)

            val = super(Record, self).__getitem__(local_att)
            data = list(cur.select(where="%s=={val}" % att,
                                   args={"val": (att, val)}))

        cur.close()  # Close cursor

        return data  # Return data

    def clear_version(self):
        """ Remove the version, forcing this as the latest version """

        self._version = None

    @property
    def connection(self):
        """ Return the parent or None """

        if self._con is None:
            return None

        par = self._con()

        if par is None:
            raise ConnectionError("Parent Connection has closed.")

        return par

    @property
    def version(self):
        """ Get the version of the record """

        return self._version
