#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .. import __version__
from hashlib import sha256


class Authenticate(object):
    def __init__(self, username=None, password=None, user_obj=None):
        if username is None and password is None and user_obj is None:
            raise ValueError("A username and password must be parsed.")

        if user_obj is None:
            user_obj = User(password=password)

            if username is None:
                raise ValueError("A username must be parsed.")

            user_obj.username = username

        if not isinstance(user_obj, User):
            raise ValueError("user_obj must be a User instance.")

        self.username = user_obj.username
        self.password = user_obj.password

    def __repr__(self):
        return ("<HTD%s Authenticated User (%s)>" %
                (__version__, self.username))


class User(object):
    """ User object """
    def __init__(self, password=None, access="roadx"):
        self.username = None
        self.password = password
        self.access = access

    def __repr__(self):
        return ("<HTD%s User (%s with '%s' permissions)>" %
                (__version__, self.username, self.access))

    @property
    def access(self):
        return self.__access

    @access.setter
    def access(self, access):
        """
        Set the access level for the user
        NOTE: Only used for creating new users, not authenticating connections
        """

        if not isinstance(access, str):
            raise ValueError("Access must be a string")

        for i in access:
            if i not in "roadx":
                raise ValueError("Invalid permission given, '%s'." % i)

        self.__access = access

    @property
    def password(self):
        return self.__password

    @password.setter
    def password(self, password):
        """ Set the password for the user """

        if password is None:
            self.__password = None
        else:
            if not isinstance(password, str):
                raise ValueError("Password must be a string or None")

            self.__password = sha256(password.encode("ISO-8859-1")).hexdigest()

    def to_dict(self):
        return {"pass": self.password, "acc": self.access}

    @property
    def username(self):
        return self.__username

    @username.setter
    def username(self, username):
        if username is None:
            self.__username = None
            return

        if isinstance(username, str) is False:
            raise ValueError("Username must a string")

        self.__username = username
