#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .._base import perm
from .._base.data import check_record_version, get_key_as_dict
from .._workers import Locking


class BaseFile(Locking):
    """
    Base class for a format
    Provides methods for the format, can override if required
    """
    def __init__(self):
        self.build_data = None
        self.escape_char = b"\x01"
        self.read_data = None
        self.searcher = None

        Locking.__init__(self)

    def __contains__(self, key):
        """ Return True or False if the key exists in the file """

        perm("r", self.access)

        try:
            self.searcher.get_line(self, key, self.header["KEY"])
        except KeyError:
            return False
        else:
            return True

    def __delitem__(self, key):
        """ Delete the key from the file (and the version) """

        perm("d", self.access)

        line = self.searcher.get_line(self, key, self.header["KEY"])

        del self._file[line["index"]]  # Delete from file

    def __getitem__(self, key):
        """ Get the data and version (as .version) for the key """

        perm("r", self.access)

        line = self.searcher.get_line(self, key, self.header["KEY"])

        data = self.read_data(self._file[line["index"]],
                              self.header["ATTBYID"])

        data[".version"] = self._file.get_version(line["index"])

        return data

    def __len__(self):
        """ Get the number of records in the file """

        return len(self._file)

    def __setitem__(self, index, record):
        """
        Set the data and version (record.version) to the index
        - This function is for overwrite so the index must exist already
        """

        perm("o", self.access)

        version = record[".version"]
        del record[".version"]

        version = check_record_version(version, self._file.get_version(index))
        string = self.build_data(record, self._file, index,
                                 self.header["ATTBYID"], self.escape_char)

        self._file[index] = string
        self._file.set_version(index, version + 1)

    def calc_file_settings(self):
        """ Update header size and # of attributes etc in file object """

        self._file.update_file_settings(self.header)

    def close(self, delete_files=False):
        """ Close file, with option to delete files, and sub classes """

        if delete_files:
            perm("d", self.access)

            self._file.lock_writing()  # Wait for others tasks to complete

        self._file.close(delete_files)  # Close file

        if self._file in self.files:
            self.files.remove(self._file)

        return True

    def copy(self, location, replace):
        """ Copy file to location with option of replacing existing files """

        perm("r", self.access)

        self._file.copy(location, replace)

        return self.__class__(location, self.access)

    def delete(self, start=None, stop=None):
        perm("d", self.access)

        if start is not None and stop is not None:
            self._file.delete_region(start, stop)

            yield stop - start

            raise StopIteration
        else:
            i = 0
            leng = len(self._file)

            while i < leng:
                output = self.read_data(self._file[i], self.header["ATTBYID"])

                delete = (yield output)

                if delete:  # delete i
                    del self._file[i]
                    leng -= 1
                else:
                    i += 1  # only +1 if not deleted (data is shifted)

                yield 1

            raise StopIteration

    def flush(self, fsync=False):
        """ Flush the file with option of fsyncing the file """

        return self._file.flush(fsync)

    def get_index(self, index):
        """ Get the record in the file at index """

        return self.read_data(self._file[index], self.header["ATTBYID"])

    def get_left(self, lower, index, upper):
        return (lower + upper) // 2, index - 1

    def get_right(self, lower, index, upper):
        return (lower + upper) // 2, index + 1

    def get_top(self, length=None):
        return length // 2

    def index(self, key):
        """ Find the index of a key """

        perm("r", self.access)

        line = self.searcher.get_line(self, key, self.header["KEY"])

        return line["index"]

    def insert(self):
        while 1:
            data = (yield)

            exists, index, line = self.is_record(data)

            if exists:
                raise KeyError("Key already exists (%s)" %
                               get_key_as_dict(data, self.header["KEY"]))
            else:
                self.insert_record(index, data)

    def insert_record(self, index, record):
        """
        Insert a record into the file
        - If the key exists then return the index
        - Else append the record and version
        """

        perm("a", self.access)

        version = record[".version"]
        del record[".version"]

        string = self.build_data(record, self._file, index,
                                 self.header["ATTBYID"], self.escape_char,
                                 True)

        self._file.insert(index, string)
        self._file.set_version(index, version)

    def is_record(self, key):
        """
        Return the index and data if the record exists
        otherwise return the index of the new record
        """

        line = self.searcher.search(self, key, self.header["KEY"], new=True)

        if line["found"]:
            return (True, line["index"], line["data_line"])
        else:
            return (False, line["index"], {})

    def rename(self, location, replace):
        """ Rename to location with option to replace any existing files """

        self._file.rename(location, replace)

        self.close()  # Close open fd

    def select(self, attribute, start, stop):
        """
        List the attributes given and version of each record from start to stop
        """

        perm("r", self.access)

        exclude = [att for att in self.header["ATT"] if att not in attribute]

        for record in self._file.iterate(start, stop):
            version, record = record
            output = self.read_data(record, self.header["ATTBYID"], exclude)

            if output is False:
                raise StopIteration

            output[".version"] = version

            yield output

            start += 1  # Use start as index

        raise StopIteration

    def update(self, start, stop):
        index = start

        for old_record in self._file.iterate(start, stop):
            version, old_record = old_record

            old_record = self.read_data(old_record, self.header["ATTBYID"])
            old_record[".version"] = version

            new_record = (yield old_record)

            if new_record == {}:
                yield 1
                continue

            # Get and remove version
            version = new_record[".version"]
            del new_record[".version"]

            version = check_record_version(version,
                                           self._file.get_version(index))

            string = self.build_data(new_record, self._file, index,
                                     self.header["ATTBYID"], self.escape_char)

            self._file[index] = string
            self._file.set_version(index, version + 1)  # Update version
            index += 1

            yield 1
