#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from ..._base.file import location_is_file, FileTools, LockFile
import os
from os import name, rename, unlink
from os.path import exists
from struct import pack, unpack
from sys import platform

# Load fsync
if platform == 'linux2':
    fsync = os.fdatasync
else:
    fsync = os.fsync


class FileWriter(LockFile):
    def __init__(self, location, header, new):
        LockFile.__init__(self)

        self.location = location
        self.offset = 0  # Offset until data
        self.width = 0  # Width of a record
        self.update_file_settings(header)

    def __delitem__(self, index):
        """ Remove the data or mark as deleted """

        self.delete_region(index, index + 1)

    def __getitem__(self, index):
        """ Get the data for the index """

        self.file.seek((index * self.width) + self.offset + 8)
        return self.file.read(self.width - 8)

    def __setitem__(self, index, data):
        """ Set the data for the index """

        self.file.seek((index * self.width) + self.offset + 8)
        self.file.write(data)

    def __len__(self):
        """ Number of items in the file """

        self.file.seek(0, 2)
        return (self.file.tell() - self.offset) // self.width

    def close(self, delete_files):
        if exists(self.location):
            self.flush()

        self.close_lock_file(delete_files)  # Deletes this file too

    def copy(self, location, replace):
        if replace is False and exists(location):
            raise IOError("Path already exists")

        if not location_is_file(location):
            raise ValueError("Location is not valid for this file format")

        with open(location, "wb") as f:
            self.file.seek(0)
            f.write(self.file.read())

    def delete_region(self, start, stop):
        """ Delete a region of data """

        FileTools.delete_region(self.file, (start * self.width) + self.offset,
                                (stop * self.width) + self.offset)

    def flush(self, fsync=False):
        if self.file is not None:
            self.file.flush()

            if fsync:
                fsync(self.file)

    def get_version(self, index):
        """ Get the version of the index """

        self.file.seek((index * self.width) + self.offset)

        return unpack("Q", self.file.read(8))[0]

    def insert(self, index, data):
        """ Insert the item into an index """

        position = (index * self.width) + self.offset

        data = b"".join([b"\x00" * 8, data])  # Insert blank version

        FileTools.insert_region(self.file, position, data)

    def iterate(self, i, stop):
        """ Iterate over a region in the file """

        reading = True

        while reading and i < stop:
            self.file.seek((i * self.width) + self.offset)
            data = self.file.read(self.width)
            reading = data != b""
            i += 1

            yield unpack("Q", data[:8])[0], data[8:]

        raise StopIteration

    def left(self, index, binary_search=False):
        """
        Get the index to the left of the index
        binary_search=True get the real left
        binary_search=False get the data left
        """

        if index - 1 > 0:
            return index - 1
        else:
            raise IndexError()

    def rename(self, location, replace):
        if replace is False and exists(location):
            raise IOError("Cannot rename because path already exists")

        if replace and name == "nt":  # Windows only rename if dst !exist
            unlink(location)

        rename(self.location, location)  # Rename the file

    def right(self, index, binary_search=False):
        """
        Get the index to the right of the index
        binary_search=True get the real right
        binary_search=False get the data right
        """

        if index + 1 < len(self):
            return index + 1
        else:
            raise IndexError()

    def set_left(self, index):
        """ Set the index to the left of the item """

        return True  # File is ordered so no need for setting positioning

    def set_right(self, index):
        """ Set the index to the right of the item """

        return True  # File is ordered so no need for setting positioning

    def set_version(self, index, version):
        """ Set the version of an index """

        self.file.seek((index * self.width) + self.offset)
        self.file.write(pack("Q", version or 0))

        return True

    def update_file_settings(self, header):
        """ Update the file settings from the header info """

        self.offset = header["SIZ"]
        self.width = sum([header["ATT"][i].size + 1
                          for i in header["ATT"]]) + 8
