#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from ... import Register
from ..._base import location_is_file
from ..base import BaseFile
from .file import FileWriter
from ..header.inline_json import JSONHeader
from os.path import exists
from ..search import BinarySearch
from struct import error, unpack, pack


def build_data(data, file, index, atts, esc, new=False):
    string = []
    offset = 0

    for att in atts:
        width = 0

        if not new:  # If record exists then read it
            old_data = file[index]

        if data[atts[att]["name"]] is None:  # Value is Null
            if new:  # New record
                string.append(b"\xff\xff\xff\xff")
            else:  # Record exists so read old
                d = old_data[offset:offset + 4]

                if d[:4] == b"\xff\xff\xff\xff":
                    string.append(d)
                else:
                    width = unpack("I", d)[0]
                    val = old_data[offset + 4:offset + 4 + width]
                    string.extend([d, val])
        else:  # Value is not null
            val = data[atts[att]["name"]]
            string.extend([pack("I", len(val)), val])

            if not new:  # Record exists so overwrite
                d = old_data[offset:offset + 4]

                if d != b"\xff\xff\xff\xff":
                    width = unpack("I", d)[0]

        offset += width + 4

    return b"".join(string)


def read_data(data, atts, exclude=[]):
    output = {}
    offset = 0

    # Read atts
    for att in atts:
        try:
            width = unpack("I", data[offset:offset + 4])[0]
        except error:
            output = False
            break

        if width == 4294967295:
            width = 0
            value = None
        else:
            value = data[offset + 4:offset + 4 + width]

        if atts[att]["name"] not in exclude:
            output[atts[att]["name"]] = value

        offset += width + 4

    return output


class VARIABLE_WIDTH_FILE(BaseFile, JSONHeader):
    escape_char = b"\x00"
    idn = b"\x01"
    persistent = True

    def __init__(self, location, access, header=None):
        # Set variables
        self.access = access
        self.files = []
        self.location = location

        # Load classes
        BaseFile.__init__(self)
        JSONHeader.__init__(self)

        self.build_data = build_data
        self.read_data = read_data
        self.searcher = BinarySearch

        # Check for file
        new_file = not exists(location)

        if new_file:
            if header is None:
                raise IOError("No HTD file found (%s)" % location)

            open(location, "wb").close()
            self.header = header  # Set and load header
        else:
            if header is not None:
                raise IOError('File already exists (%s)' % location)

            self.header_confirm()  # Load header

        self._file = FileWriter(location, self.header, new_file)

        self.files.append(self._file)  # Index run from File.flush()
        self._header_update_funcs.append(self.calc_file_settings)
        self._confirm_funcs.extend([self.header_confirm,
                                    self._file.index.index_confirm])


Register.register_fhandler(VARIABLE_WIDTH_FILE, location_is_file)
Register.default_fhandler = VARIABLE_WIDTH_FILE
