#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from ..._base.file import location_is_file, FileTools, LockFile
from ..index.width_store import WidthStore
import os
from os import name, rename, unlink
from os.path import exists
from sys import platform
from struct import pack, unpack


# Load fsync
if platform == 'linux2':
    fsync = os.fdatasync
else:
    fsync = os.fsync


class FileWriter(LockFile):
    def __init__(self, location, header, new):
        LockFile.__init__(self)

        self.location = location
        self.num_atts = 0  # Width of a record
        self.offset = 0  # Offset until data
        self.update_file_settings(header)

        self.lock_writing()
        self.index = WidthStore(location, self.index_builder)

        if new:
            self.index.clear()

        self.unlock()

    def __delitem__(self, index):
        """ Remove the data or mark as deleted """

        position, width = self.index[index]

        FileTools.delete_region(self.file, position + self.offset,
                                position + width + self.offset)

        del self.index[index]

    def __getitem__(self, index):
        """ Get the data for the index """

        position, width = self.index[index]

        self.file.seek(position + self.offset + 8)
        return self.file.read(width - 8)

    def __setitem__(self, index, data):
        """ Set the data for the index """

        position, width = self.index[index]

        FileTools.overwrite_region(self.file, position + self.offset + 8,
                                   width - 8, data)
        self.index[index] = len(data) + 8

    def __len__(self):
        """ Number of items in the file """

        return len(self.index)

    def close(self, delete_files):
        if exists(self.location):
            self.flush()

        self.index.close(delete_files)
        self.close_lock_file(delete_files)  # Deletes this file too

    def copy(self, location, replace):
        if replace is False and exists(location):
            raise IOError("Path already exists")

        if not location_is_file(location):
            raise ValueError("Location is not valid for this file format")

        with open(location, "wb") as f:
            self.file.seek(0)
            f.write(self.file.read())

    def delete_region(self, start, stop):
        """ Delete a region of data """

        start_position = self.index.get_position(start) + self.offset
        stop_position = self.index.get_position(stop) + self.offset

        FileTools.delete_region(self.file, start_position, stop_position)
        self.index.clear_slice(start, stop)

    def flush(self, fsync=False):
        self.index.flush(fsync)

        # with open(self.location, "rb") as f:
        if self.file is not None:
            self.file.flush()

            if fsync:
                fsync(self.file)

    def get_version(self, index):
        """ Get the version of the index """

        self.file.seek(self.index.get_position(index) + self.offset)

        return unpack("Q", self.file.read(8))[0]

    def index_builder(self):
        """ Present the record widths to the index """

        self.file.seek(self.offset)

        reading = True

        while reading:
            self.file.seek(8, 1)  # Seek over version

            record_width = 8
            x = 0

            while x < self.num_atts:
                width = self.file.read(4)  # Get width

                if width == b"":
                    reading = False
                else:
                    width = unpack("I", width)[0]  # unpack width

                    if width == 4294967295:  # Check not NULL
                        width = 0

                    record_width += width + 4  # Add width and struct width
                    self.file.seek(width, 1)  # Seek over data

                x += 1

            yield record_width

        raise StopIteration

    def insert(self, index, data):
        """ Insert the item into an index """

        position = self.index.get_position(index) + self.offset

        data = b"".join([b"\x00" * 8, data])  # Insert blank version

        FileTools.insert_region(self.file, position, data)
        self.index.insert(index, len(data))

    def iterate(self, start, stop):
        """ Iterate over a region in the file """

        reading = True
        i = int(start)

        while reading and i < stop:
            position, width = self.index[i]
            self.file.seek(position + self.offset)
            data = self.file.read(width)
            reading = data != b""
            i += 1

            yield unpack("Q", data[:8])[0], data[8:]

        raise StopIteration

    def left(self, index, binary_search=False):
        """
        Get the index to the left of the index
        binary_search=True get the real left
        binary_search=False get the data left
        """

        if index > 0:
            return index - 1
        else:
            raise IndexError()

    def rename(self, location, replace):
        if replace is False and exists(location):
            raise IOError("Cannot rename because path already exists")

        self.index.close(delete_files=True)  # Delete index - rebuilt later

        if replace and name == "nt":  # Windows only rename if dst !exist
            unlink(location)

        rename(self.location, location)  # Rename the file

    def right(self, index, binary_search=False):
        """
        Get the index to the right of the index
        binary_search=True get the real right
        binary_search=False get the data right
        """

        if index + 1 < len(self):
            return index + 1
        else:
            raise IndexError()

    def set_left(self, index):
        """ Set the index to the left of the item """

        return True  # File is ordered so no need for setting positioning

    def set_right(self, index):
        """ Set the index to the right of the item """

        return True  # File is ordered so no need for setting positioning

    def set_version(self, index, version):
        """ Set the version of an index """

        self.file.seek(self.index.get_position(index) + self.offset)
        self.file.write(pack("Q", version or 0))

        return True

    def update_file_settings(self, header):
        """ Update the file settings from the header info """

        self.offset = header["SIZ"]
        self.num_atts = len(header["ATT"])
