#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from ... import DataType, Register
from ..._base.data import get_attributes_by_id
from ..._base.file import FileTools
from ..._objects.attribute import Attribute
from ..._objects.user import User
from json import dumps, loads
from struct import pack, unpack


class JSONHeader():
    def __init__(self, header=None):
        self._header = {"IDN": 0}
        self._header_update_funcs = []

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, header):
        """ Set new header """

        # Build inital str
        write = [self.idn, pack("Q", self._header["IDN"] + 1)]
        data = {}

        for att in header:
            if att == "ATT":  # Add header atts
                data["ATT"] = {}

                for att in header["ATT"]:
                    attr = header["ATT"][att]
                    attr = attr.to_dict()

                    for i in ["dtype", "size"]:
                        attr[i] = attr[i].decode("ISO-8859-1")

                    data["ATT"][att] = attr
            elif att == "USR":
                data["USR"] = {}

                for user in header["USR"]:
                    data["USR"][user] = header["USR"][user].to_dict()
            elif att in ["IDN", "FOR"]:
                continue
            else:
                data[att] = header[att]

        write.append(dumps(data).encode("ISO-8859-1"))

        write = b"".join(write)
        write = b"".join([pack("Q", len(write) + 8), write])  # Add length

        # Write to file
        width = 0

        if self._header["IDN"] != 0:
            width = self._header["SIZ"]

        if hasattr(self, "_file") and self._file.file is not None:
            FileTools.overwrite_region(self._file.file, 0, width, write)
            self.__load(self._file.file)
        else:
            with open(self.location, "rb+") as writer:
                FileTools.overwrite_region(writer, 0, width, write)
                self.__load(writer)

        for i in self._header_update_funcs:
            i()

        return True

    def header_confirm(self):
        """ Check for changes in the header """

        # Use open file it is exists and is open
        if hasattr(self, "_file") and self._file.file is not None:
            self._file.file.seek(9)

            idn = unpack("Q", self._file.file.read(8))[0]

            if idn != self.header["IDN"]:
                self.__load(self._file.file)

                for i in self._header_update_funcs:
                    i()
        else:
            with open(self.location, "rb") as f:
                f.seek(9)
                idn = unpack("Q", f.read(8))[0]

                if idn != self.header["IDN"]:
                    self.__load(f)

                    for i in self._header_update_funcs:
                        i()

    def __load(self, file):
        """ Convert and read the data in the file """

        # Check data exists
        file.seek(0)
        data = file.read(8)

        if data == b"":
            return False

        # Read data
        length = unpack("Q", data)[0]
        file_format = file.read(1)
        idn = unpack("Q", file.read(8))[0]

        data = loads(file.read(length - 17).decode("ISO-8859-1"))
        data["FOR"] = file_format
        data["IDN"] = idn
        data["SIZ"] = length

        for att in data["ATT"]:  # Load Attribute object
            val = data["ATT"][att]

            dtype = Register.dtype_codes[val["dtype"].encode("ISO-8859-1")]
            dtype = getattr(DataType, dtype)

            size = unpack("I", val["size"].encode("ISO-8859-1"))[0]

            tmp = Attribute(dtype=dtype, size=size)
            tmp._id = val["id"]
            tmp._name = att
            tmp._limit_list = val["limit_list"]
            tmp._range = val["range"]
            tmp._reg_expr = val["reg_expr"]
            tmp._required = val["required"]
            tmp._default = val["default"]

            data["ATT"][att] = tmp

        for user in data["USR"]:  # Load User object
            val = data["USR"][user]
            data["USR"][user] = User(access=val["acc"])
            data["USR"][user]._User__username = user
            data["USR"][user]._User__password = val["pass"]  # sha256

        # Add calculated
        data["ATTBYID"] = get_attributes_by_id(data["ATT"])

        self._header = data  # Update var
