#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide

Header object - Manages the I/O for the header
'''

from ... import DataType, Register
from ..._base.data import get_attributes_by_id
from ..._base.file import FileTools
from ..._objects.attribute import Attribute
from ..._objects.user import User
from struct import pack, unpack


class VariableWidthHeader():
    def __init__(self):
        self._header = {"IDN": 0}
        self._header_update_funcs = []

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, header):
        """ Set new header """

        self.__update(header)

    def __load(self, file):
        """ Load header into cache """

        # Check data exists
        file.seek(0)
        data = file.read(8)

        if data == b"":
            return False

        # Read data
        length = unpack("Q", data)[0]
        file_format = file.read(1)
        idn = unpack("Q", file.read(8))[0]
        header = {"FOR": file_format, "IDN": idn}
        pos = file.tell()

        while pos < length:
            wid = unpack("Q", file.read(8))[0]
            data = file.read(wid)
            header[str(data[0:3].decode("ISO-8859-1"))] = data[3:]
            pos += 8 + wid

        # Convert data

        # Convert Version
        header["VER"] = float(header["HTD"])
        del header["HTD"]

        # Convert each
        for att in header:
            if att not in ["VER", "FOR", "IDN"]:
                header[att] = eval(header[att])

            if att == "ATT":
                for a in header[att]:
                    val = header[att][a]

                    dtype = getattr(DataType,
                                    Register.dtype_codes[val["dtype"]])

                    size = val["size"]

                    if not isinstance(size, bytes):
                        size = size.encode("ISO-8859-1")

                    size = unpack("I", size)[0]

                    tmp = Attribute(dtype=dtype, size=size)
                    tmp._id = val["id"]
                    tmp._name = a
                    tmp._limit_list = val["limit_list"]
                    tmp._range = val["range"]
                    tmp._reg_expr = val["reg_expr"]
                    tmp._required = val["required"]
                    tmp._default = val["default"]

                    header[att][a] = tmp

            if att == "USR":
                for a in header[att]:
                    val = header[att][a]
                    header[att][a] = User(access=val["acc"])
                    header[att][a]._User__username = a
                    header[att][a]._User__password = val["pass"]  # sha256

        # Add calculated
        header["SIZ"] = length
        header["ATTBYID"] = get_attributes_by_id(header["ATT"])

        self._header = header  # Update var

    def __update(self, header):
        # Build inital str
        leng = pack("Q", len(str(header["VER"])) + 3)
        write = [self.idn, pack("Q", self._header["IDN"] + 1), leng,
                 b"HTD", str(header["VER"]).encode("ISO-8859-1")]

        # Add header atts
        for att in ["KEY", "USR", "TAB", "ATT", "R1M", "RM1"]:
            if att in ["ATT", "USR"]:
                tmp = {}
                for i in header[att]:
                    tmp[i] = header[att][i].to_dict()
            else:
                tmp = header[att]

            length = len(str(tmp)) + 3
            write.extend([pack("Q", length), att.encode("ISO-8859-1"),
                          str(tmp).encode("ISO-8859-1")])

        write = b"".join(write)
        write = b"".join([pack("Q", len(write) + 8), write])  # Add length

        # Write to file
        width = 0

        if self._header["IDN"] != 0:
            width = self._header["SIZ"]

        if hasattr(self, "_file") and self._file.file is not None:
            FileTools.overwrite_region(self._file.file, 0, width, write)
            self.__load(self._file.file)
        else:
            with open(self.location, "rb+") as writer:
                FileTools.overwrite_region(writer, 0, width, write)
                self.__load(writer)

        for i in self._header_update_funcs:
            i()

        return True

    def header_confirm(self):
        """ Check for changes in the header """

        # Use open file it is exists and is open
        if hasattr(self, "_file") and self._file.file is not None:
            self._file.file.seek(9)
            idn = unpack("Q", self._file.file.read(8))[0]

            if idn != self.header["IDN"]:
                self.__load(self._file.file)

                for i in self._header_update_funcs:
                    i()
        else:
            with open(self.location, "rb") as f:
                f.seek(9)
                idn = unpack("Q", f.read(8))[0]

                if idn != self.header["IDN"]:
                    self.__load(f)

                    for i in self._header_update_funcs:
                        i()

        return
