#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from ..._base.file import convert_location, FileTools
from os import unlink
from os.path import exists
from struct import pack, unpack


class WidthStore():
    def __init__(self, location, builder):
        self.alive = True

        self.builder = builder
        self.cache = []
        self.cache_size = 53
        self.location = convert_location(location, ".~index_")
        self.revision = -1

        # Create location
        if not exists(self.location):
            open(self.location, "w").close()
            self.build()

        self.index_confirm()  # Build cache and get revision

    def __delitem__(self, index):
        """ Delete a record from index """

        position = (index * 8) + 8
        with open(self.location, "rb+") as writer:
            FileTools.delete_region(writer, position, position + 8)

        self.update_revision()
        self.create_cache()

    def __getitem__(self, index):
        """ Get the position and width of the record """

        return (self.get_position(index), self.get_width(index))

    def __len__(self):
        """ Return number of records """

        with open(self.location, "rb") as f:
            f.seek(0, 2)
            return (f.tell() - 8) // 8

    def __setitem__(self, index, width):
        """ Change record width to value """

        with open(self.location, "rb+") as f:
            f.seek((index * 8) + 8)
            f.write(pack("Q", width))

        self.update_revision()
        self.create_cache()

    def build(self):
        """ Build the file """

        self.clear()

        for i, width in enumerate(self.builder()):
            self.insert(i, width)

    def clear(self):
        """ Clear the index file and internal cache """

        if not self.alive:
            return

        with open(self.location, "rb+") as f:
            f.seek(8)
            f.truncate()

        self.update_revision()
        self.create_cache()  # Create cache

    def clear_slice(self, start, stop):
        """ Clear record(s) from index """

        if not self.alive:
            return

        with open(self.location, "rb+") as writer:
            FileTools.delete_region(writer, (start * 8) + 8, (stop * 8) + 8)

        self.update_revision()
        self.create_cache()

    def close(self, delete_files=False):
        """ Disconnect from index """

        self.alive = False
        self.cache = []
        self.cache_size = 50
        self.revision = None

        if delete_files and exists(self.location):
            unlink(self.location)

    def create_cache(self):
        """
        Create cache from connected file
        - Cache holds blocks of *cache_size* records
        """

        if not self.alive:
            return

        # Calculate Cache Size y = 0.01033x + 52.77
        length = len(self)
        self.cache_size = int((length * 0.01033) + 52.77)
        length //= self.cache_size  # Work out number of blocks

        # Set vars
        x = 0
        self.cache = []  # Clear current cache
        block_fmt = "Q" * self.cache_size  # Build struct for block
        block_width = self.cache_size * 8  # Calc width of block

        with open(self.location, "rb") as f:
            f.seek(8)

            while x < length:  # Read file
                self.cache.append(sum(unpack(block_fmt, f.read(block_width))))
                x += 1

    def flush(self, fsync=False):
        """ Check the file has been flushed """

        if exists(self.location):
            with open(self.location, "rb") as f:
                f.flush()

    def get_position(self, index):
        """ Get position of record in index """

        blocks, remainder = divmod(index, self.cache_size)

        if blocks == 0:
            length = 0
        else:
            length = sum(self.cache[:blocks])

        with open(self.location, "rb") as f:
            f.seek(blocks * (self.cache_size * 8) + 8)
            data = f.read(remainder * 8)  # Read width

        if remainder != 0:
            length += sum(unpack("Q" * remainder, data))

        return length

    def get_width(self, index):
        """ Get length of record """

        with open(self.location, "rb") as f:
            f.seek((8 * index) + 8)
            data = f.read(8)

        if len(data) == 0:
            return 0
        else:
            return unpack("Q", data)[0]

    def index_confirm(self):
        """ Check for changes in the header """

        with open(self.location, "rb") as f:
            revision = unpack("Q", f.read(8))[0]

        if revision != self.revision:
            self.revision = revision
            self.create_cache()

    def insert(self, index, value):
        """ Insert record into position """

        data = pack("Q", value)

        with open(self.location, "rb+") as writer:
            FileTools.insert_region(writer, (index * 8) + 8, data)

        self.update_revision()
        self.create_cache()

    def update_revision(self):
        """ Update the revision of the index """

        self.revision += 1

        with open(self.location, "rb+") as f:
            f.write(pack("Q", self.revision))
