#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
from struct import pack
import unittest


class AttributeTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test init
        - None-None, dtype-None, None-size, dtype-size
        - Check dtype, size, dict
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)

        #------------------------------------------------------------ None-None
        att = htd.Attribute()

        self.assertEqual(att.dtype, default, "Dtype should be default")
        self.assertEqual(att.size, 255, "Default size should be 255")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

        #----------------------------------------------------------- dtype-None
        att = htd.Attribute(dtype=htd.DataType.STRING)

        self.assertEqual(att.dtype, htd.DataType.STRING,
                         "Dtype should be default")
        self.assertEqual(att.size, 255, "Default size should be 255")

        att_dict = {"dtype": htd.DataType.STRING.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

        #------------------------------------------------------------ None-size
        att = htd.Attribute(size=10)

        self.assertEqual(att.dtype, default, "Dtype should be default")
        self.assertEqual(att.size, 10, "Default size should be 255")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 10)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

        #----------------------------------------------------------- dtype-size
        att = htd.Attribute(dtype=htd.DataType.STRING, size=10)

        self.assertEqual(att.dtype, htd.DataType.STRING,
                         "Dtype should be default")
        self.assertEqual(att.size, 10, "Default size should be 255")

        att_dict = {"dtype": htd.DataType.STRING.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 10)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_repr(self):
        """
        Test repr
        - name, dtype, size
        """

        #------------------------------------------------------------ Blank att
        att = htd.Attribute()

        att_repr = ("<HTD%s Attribute (<name '%s'> <type '%s'> <len %s>)>" %
                    (htd.__version__, None, htd.Register.default_dtype, "255"))

        self.assertEqual(repr(att), att_repr, "repr is incorrect")

        #------------------------------------------------------------- Real att
        att = htd.Attribute(dtype=htd.DataType.STRING, size=50)
        att.name = "id"

        att_repr = ("<HTD%s Attribute (<name '%s'> <type '%s'> <len %s>)>" %
                    (htd.__version__, "id", "STRING", "50"))

        self.assertEqual(repr(att), att_repr, "repr is incorrect")

    def test_default(self):
        """
        Test default
        - Set/Get
        """

        att = htd.Attribute(dtype=htd.DataType.STRING)

        # Get
        self.assertEqual(att.default, None, "Default is not None")

        # Set
        att.default = "abc"

        # Get
        self.assertEqual(att.default, "abc", "Default no abc")

        att_dict = {"dtype": htd.DataType.STRING.idn, "id": None,
                    "limit_list": [False, []], "default": b"abc",
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_dtype(self):
        """
        Test dtype
        - set/get
        - bad dtype
        """

        att = htd.Attribute()

        # Bad dtype Set
        try:
            att.dtype = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Bad dtype (%s)" % e)
        else:
            raise ValueError("No errors using Bad dtype")

        # Dtype set
        att.dtype = htd.DataType.STRING

        self.assertEqual(att.dtype, htd.DataType.STRING, "Dtype is not STRING")

        att_dict = {"dtype": htd.DataType.STRING.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_limit_list(self):
        """
        Test limit list
        - set/get
        - not a list
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        # Not a list
        try:
            att.limit_list = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Not a list (%s)" % e)
        else:
            raise ValueError("No errors using Not a list")

        # Limit list set
        att.limit_list = [1, 2, 3]

        self.assertListEqual(att.limit_list, [True, [1, 2, 3]],
                             "Limit list is not 1,2,3")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [True, [1, 2, 3]], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_name(self):
        """
        Test name
        - set/get
        - not str, includes .
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        # Not str
        try:
            att.name = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, not str (%s)" % e)
        else:
            raise ValueError("No errors using not str")

        # Includes a .
        try:
            att.name = "abc.def"
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, includes . (%s)" % e)
        else:
            raise ValueError("No errors using includes .")

        # Name set
        att.name = "abc"

        self.assertEqual(att.name, "abc", "Name is not abc")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_range(self):
        """
        Test range
        - set/get
        - range not list, lower > upper
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        # Range not list
        try:
            att.range = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, not list (%s)" % e)
        else:
            raise ValueError("No errors using not list")

        # Lower > upper
        try:
            att.range = [10, 5]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, lower > upper (%s)" % e)
        else:
            raise ValueError("No errors using lower > upper")

        # Range Set
        att.range = [1, 5]

        self.assertEqual(att.range, [1, 5], "Range is not 1 < x < 5")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": [1, 5], "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_reg_expr(self):
        """
        Test reg_expr
        - set/get
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        # Reg expr set
        att.reg_expr = "[^0-9]"

        self.assertEqual(att.reg_expr, b"[^0-9]", "Reg expr not [^0-9]")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": b"[^0-9]",
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_required(self):
        """
        Test required
        - set/get
        - Not True/False
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        # Required not True/False
        try:
            att.required = "abc"
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, not bool (%s)" % e)
        else:
            raise ValueError("No errors using not bool")

        # Required Set
        att.required = True

        self.assertEqual(att.required, True, "Required is not True")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": True, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

    def test_size(self):
        """
        Test size
        - set/get
        - size not int, size to large or < 0
        - size using fixed dtype (eg Integer)
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        # Size not int
        try:
            att.size = "abc"
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, size not int (%s)" % e)
        else:
            raise ValueError("No errors using size not int")

        # Size < 0
        try:
            att.size = -1
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, size < 0 (%s)" % e)
        else:
            raise ValueError("No errors using size < 0")

        # Size > 2^32
        try:
            att.size = 2 ** 32
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, size > 2^32 (%s)" % e)
        else:
            raise ValueError("No errors using size > 2^32")

        # Size Set
        att.size = 10

        self.assertEqual(att.size, 10, "Size is not 10")

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 10)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")

        # Size using fixed dtype
        att.dtype = htd.DataType.INTEGER

        att.size = 15

        self.assertEqual(att.size, 4, "Size should come from attribute")

    def test_to_dict(self):
        """
        Test to_dict
        """

        default = getattr(htd.DataType, htd.Register.default_dtype)
        att = htd.Attribute()

        att_dict = {"dtype": default.idn, "id": None,
                    "limit_list": [False, []], "default": None,
                    "range": False, "reg_expr": None,
                    "required": False, "size": pack("I", 255)}

        self.assertEqual(att.to_dict(), att_dict, "Dict is incorrect")
