#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest

names.append("/tmp/.~tbl1.backup.old.htd")


class AttributesTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test test
        - Open from Cursor object
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertTrue(atts.alive, "Atts not alive")
        self.assertEqual(con.location, atts.location, "Location not correct")
        self.assertEqual(atts.connection, con, "Auth connection incorrect")

        atts.close()
        con.close()

    def test_contains(self):
        """
        Test contains
        - exists, !exist
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertTrue("id" in atts, "id should be in attributes")
        self.assertFalse("abc" in atts, "abc should not be in attributes")

        con.close()

    def test_del(self):
        """
        Test del
        - exist, !exist, att is KEY
        - data
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        # !exist
        try:
            del atts["attr"]
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, att !exist (%s)" % e)
        else:
            raise ValueError("No errors using att !exist")

        # att is KEY
        try:
            del atts["id"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, att is KEY (%s)" % e)
        else:
            raise ValueError("No errors using att is KEY")

        # prepare
        atts["name"] = htd.Attribute()
        self.assertTrue("name" in atts, "Could not prepare test, name !att")

        # Add data
        atts["tmp"] = htd.Attribute()

        cur["1"] = {"tmp": "1"}
        cur["2"] = {"tmp": "2"}
        cur["3"] = {"tmp": "3"}

        del atts["tmp"]

        l = [{"id": "1", "name": None}, {"id": "2", "name": None},
             {"id": "3", "name": None}]

        self.assertListEqual(list(cur), l, "List incorrect")

        cur.delete()

        self.assertEqual(len(cur), 0, "Table not empty")

        # exist
        del atts["name"]

        self.assertFalse("name" in atts, "Attribute not deleted")

        con.close()

    def test_get(self):
        """
        Test get
        - att exists, !exist
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        # exists
        att = atts["id"]

        self.assertEqual(att.name, "id", "Id name incorrect")

        # !exist
        try:
            atts["data"]
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, att !exist (%s)" % e)
        else:
            raise ValueError("No errors using att !exist")

        con.close()

    def test_iter(self):
        """
        Test iter
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes
        atts["name"] = htd.Attribute()

        items = list(atts)

        self.assertEqual(len(items), 2, "Length incorrect")

        con.close()

    def test_len(self):
        """
        Test len
        - 1 att, 2 atts
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertEqual(len(atts), 1, "Length should be 1")

        atts["name"] = htd.Attribute()

        self.assertEqual(len(atts), 2, "Length should be 2")

        con.close()

    def test_repr(self):
        """
        Test repr
        - open, closed
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        repr_str = "<HTD%s Attributes (%s)>" % (htd.__version__, names[0])
        self.assertEqual(repr(atts), repr_str, "Repr incorrect for open atts")

        atts.close()

        repr_str = "<HTD%s Closed Attributes>" % (htd.__version__)
        self.assertEqual(repr(atts), repr_str, "Repr incorrect closed atts")

        con.close()

    def test_set(self):
        """
        Test set
        - ! Attribute instance, name is None, normal
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        # !Attribute
        try:
            atts["abc"] = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, !Attribute (%s)" % e)
        else:
            raise ValueError("No errors using index !Attribute")

        # Name is None
        try:
            atts[None] = htd.Attribute()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, name is NULL (%s)" % e)
        else:
            raise ValueError("No errors using name is NULL")

        # Normal
        self.assertFalse("name" in atts, "Prepare failed, name in atts")

        atts["name"] = htd.Attribute()

        self.assertTrue("name" in atts, "Name has not been added to atts")

        # Add data
        cur["1"] = {"name": "1"}
        cur["2"] = {"name": "2"}
        cur["3"] = {"name": "3"}

        # Overwrite attribute
        att = htd.Attribute()
        att.size = 1024

        atts["name"] = att

        l = [{"id": "1", "name": "1"}, {"id": "2", "name": "2"},
             {"id": "3", "name": "3"}]
        self.assertListEqual(list(cur), l, "List incorrect")

        con.close()

    def test_close(self):
        """
        Test Close
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertTrue(atts.alive, "*Prepare* atts not alive")

        atts.close()

        self.assertFalse(atts.alive, "Atts should be dead")

        con.close()

    def test_connnection(self):
        """
        Test connection
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertEqual(atts.connection, con, "Connection is not correct")

        con.close()

    def test_index(self):
        """
        Test index
        - User exists, user !exist
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        # Attribute exists
        self.assertEqual(atts.index("id"), 0, "Index incorrect")

        # Attribute !exist
        try:
            atts.index("abc")
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, index !att (%s)" % e)
        else:
            raise ValueError("No errors using index !att")

        con.close()

    def test_keys(self):
        """
        Test key
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute()

        attributes = atts.keys()

        self.assertEqual(len(attributes), 2, "Number of atts is incorrect")
        self.assertListEqual(attributes, ["id", "data"], "List is incorrect")

        con.close()

    def test_location(self):
        """
        Test location
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertEqual(atts.location, names[0], "Location is incorrect")

        con.close()

    def test_rename(self):
        """
        Test rename
        """

        # Prepare
        con = htd.connect(names[0])

        cur = con[None]

        attributes = cur.attributes
        attributes["name"] = htd.Attribute()

        cur["1"] = {"name": "a"}
        cur["2"] = {"name": "b"}
        cur["3"] = {"name": "c"}

        # Test rename
        self.assertTrue(attributes.rename("name", "data"),
                        "Rename should have occurred")

        self.assertListEqual(list(cur.select("data")), ["a", "b", "c"],
                             "Data has not been converted")

        # Test rename - att exists
        try:
            attributes.rename("data", "id")
        except KeyError:
            pass
        else:
            raise ValueError("Rename should fail")

        # Test rename identifier
        self.assertTrue(attributes.rename("id", "id0"), "Rename identifier")
        self.assertEqual(cur.identifier, "id0", "Identifier incorrect")

        con.close()

        # Test rename relationships
        con = htd.connect(names[1])

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        # key-att
        con.link("b.id", "a.parent")

        cur_a = con["a"]
        cur_b = con["b"]

        cur_b["1"] = {}
        cur_b["2"] = {}
        cur_b["3"] = {}

        cur_a["0"] = {"parent": "1"}

        # Test M1 rename
        attributes = cur_a.attributes

        attributes.rename("parent", "parent_2")
        self.assertEqual(cur_a["0"]["parent_2.b"]["id"], "1",
                         "M1 is incorrect")

        attributes.close()

        # Test 1M rename
        attributes = cur_b.attributes

        attributes.rename("id", "id_2")
        self.assertListEqual(list(cur_b["1"]["id_2.a"]),
                             [{"id": "0", "name": None, "parent_2": "1"}],
                             "M1 is incorrect")

        attributes.close()

        con.close()

    def test_version(self):
        """
        Test version
        """

        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes

        self.assertEqual(atts.version, htd.__version__, "Version is incorrect")

        con.close()
