#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class AuthTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test test
        - Open from Connection object
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertTrue(auth.alive, "Auth not alive")
        self.assertEqual(con.location, auth.location, "Location not correct")
        self.assertEqual(auth.connection, con, "Auth connection incorrect")

        auth.close()

        con.close()

    def test_contains(self):
        """
        Test __contains__
        - Exists, !Exists
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertTrue("guest" in auth, "Guest should be in auth")
        self.assertFalse("abc" in auth, "abc should not exist in auth")

        auth.close()

        con.close()

    def test_delitem(self):
        """
        Test __delitem__
        - User !exist, User exist, Delete last user, Delete self,
        - 1 user left without roadx
        """

        con = htd.connect(names[1])

        auth = con.authentication

        # User !exist
        try:
            del auth["abc"]
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, delete bad user (%s)" % e)
        else:
            raise ValueError("No errors using delete bad user")

        # Delete Last user
        try:
            del auth["guest"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, delete last user (%s)" % e)
        else:
            raise ValueError("No errors using delete last user")

        # Prepare
        auth["user"] = htd.User(password="abc", access="dx")
        self.assertTrue("user" in auth, "user not in auth, prepare fail")

        con1 = htd.connect(names[1],
                           user=htd.Authenticate(username="user",
                                                 password="abc"))

        auth1 = con1.authentication

        # Delete self
        try:
            del auth1["user"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, delete self (%s)" % e)
        else:
            raise ValueError("No errors using delete self")

        # No user with full perm
        try:
            del auth1["guest"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, no full perm (%s)" % e)
        else:
            raise ValueError("No errors using no full perm")

        auth1.close()
        con1.close()

        # User Exist
        del auth["user"]
        self.assertFalse("user" in auth, "User not deleted from auth")

        con.close()

    def test_get(self):
        """
        Test get
        - User exists, User !exist
        """

        con = htd.connect(names[1])

        auth = con.authentication

        # User exists
        user = auth["guest"]

        self.assertIsInstance(user, htd.User, "User not User instance")
        self.assertEqual(user.username, "guest", "Username is incorrect")
        self.assertEqual(user.access, "roadx", "Permission is incorrect")

        # User !exist
        try:
            user = auth["abc"]
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, user !exist (%s)" % e)
        else:
            raise ValueError("No errors using user !exist")

        con.close()

    def test_iter(self):
        """
        Test iter
        """

        con = htd.connect(names[1])

        auth = con.authentication
        auth["user"] = htd.User(password="abc", access="dx")

        users = list(auth)

        self.assertEqual(len(users), 2, "Length is incorrect")

        con.close()

    def test_len(self):
        """
        Test len
        - one user, two users
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertEqual(len(auth), 1, "Length of users not 1")

        auth["user"] = htd.User(password="abc", access="dx")

        self.assertEqual(len(auth), 2, "Length of users not 2")

        con.close()

    def test_repr(self):
        """
        Test repr
        - Open, closed
        """

        con = htd.connect(names[1])

        auth = con.authentication

        repr_str = "<HTD%s Auth (%s)>" % (htd.__version__, names[1])
        self.assertEqual(repr(auth), repr_str, "Repr incorrect open Auth")

        auth.close()

        repr_str = "<HTD%s Closed Auth>" % htd.__version__
        self.assertEqual(repr(auth), repr_str, "Repr incorrect closed Auth")

        con.close()

    def test_set(self):
        """
        Test set
        - User, Not user instance, overwrite perms
        """

        con = htd.connect(names[1])

        auth = con.authentication

        # Not user instance
        try:
            auth["abc"] = "user!"
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, ! user instance (%s)" % e)
        else:
            raise ValueError("No errors using ! user instance")

        # Overwrite perms
        try:
            auth["guest"] = htd.User(access="rd")
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, overwrite perm (%s)" % e)
        else:
            raise ValueError("No errors using overwrite perm")

        # User
        auth["def"] = htd.User(password="abc", access="r")

        self.assertIn("def", auth, "def not in users")
        self.assertEqual(len(auth), 2, "Length incorrect")

        con.close()

        return

    def test_close(self):
        """
        Test Close
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertTrue(auth.alive, "*Prepare* auth not alive")

        auth.close()

        self.assertFalse(auth.alive, "Auth should be dead")

        con.close()

    def test_connnection(self):
        """
        Test connection
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertEqual(auth.connection, con, "Connection is not correct")

        con.close()

    def test_index(self):
        """
        Test index
        - User exists, user !exist
        """

        con = htd.connect(names[1])

        auth = con.authentication

        # User exists
        self.assertEqual(auth.index("guest"), 0, "Index incorrect")

        # User !exist
        try:
            auth.index("abc")
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, index !user (%s)" % e)
        else:
            raise ValueError("No errors using index !user")

        con.close()

    def test_keys(self):
        """
        Test key
        """

        con = htd.connect(names[1])

        auth = con.authentication
        auth["def"] = htd.User(password="abc", access="r")

        users = auth.keys()

        self.assertEqual(len(users), 2, "Number of users is incorrect")
        self.assertListEqual(users, ["def", "guest"], "List is incorrect")

        con.close()

    def test_location(self):
        """
        Test location
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertEqual(auth.location, names[1], "Location is incorrect")

        con.close()

    def test_version(self):
        """
        Test version
        """

        con = htd.connect(names[1])

        auth = con.authentication

        self.assertEqual(auth.version, htd.__version__, "Version is incorrect")

        con.close()
