#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class ConnectionTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test Connection
        - DB, Table, ChildTable

        alive, user, is_db, location
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        # Alive
        self.assertIs(con.alive, True, "Connection should be alive")

        # User
        self.assertIsInstance(con.user, str, "User should be str instance")
        self.assertEqual(con.user, "guest", "User should be guest")

        # Is_db
        self.assertIs(con.is_db, False, "Should be a table")

        # Location
        self.assertEqual(con.location, names[0], "Location is incorrect")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Alive
        self.assertIs(con.alive, True, "Connection should be alive")

        # User
        self.assertIsInstance(con.user, str, "User should be str instance")
        self.assertEqual(con.user, "guest", "User should be guest")

        # Is_db
        self.assertIs(con.is_db, True, "Should be a DB")

        # Location
        self.assertEqual(con.location, names[1], "Location is incorrect")

        # Prepare test
        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        con.close()

        #----------------------------------------------------------- ChildTable

        #
        # BUG lp:1085278
        #

        try:
            con = htd.connect(names[3])
            con.close()
        except htd.ConnectionError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, child table (%s)" % e)
        else:
            raise ValueError("No errors using child table")

    def test_init_extreme_location(self):
        """
        __init__ -> Location
        NULL, not a str, incorrect location
        """

        # NULL location
        try:
            con = htd.connect(None)
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, NULL location (%s)" % e)
        else:
            raise ValueError("No errors using NULL location a table")

        # Not a str
        try:
            con = htd.connect(123)
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, not str location (%s)" % e)
        else:
            raise ValueError("No errors using not str location a table")

        # Incorrect location
        try:
            con = htd.connect(names[3])
            con.close()
        except htd.ConnectionError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, bad location (%s)" % e)
        else:
            raise ValueError("No errors using bad location a table")

    def test_contains(self):
        """
        Test __contains__
        table, db
        - exists, !exist
        """

        #----------------------------------------------------------- Table test

        # __contains__ should not be available
        con = htd.connect(names[0])

        try:
            "a" in con
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table contains table (%s)"
                             % e)
        else:
            raise ValueError("No errors using table contains table")
        finally:
            con.close()

        #-------------------------------------------------------------- DB Test

        # !Exists
        con = htd.connect(names[1])

        self.assertFalse("abc" in con, "Table does exist")

        # Prepare test
        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        # Exists
        self.assertTrue("abc" in con, "Table does not exist")

        con.close()

    def test_del(self):
        """
        Test __delitem__
        table, db
        - exists, !exist, check header.TAB

        BUG lp:1084683
        """

        #----------------------------------------------------------- Table test

        # __delitem__ should not be available
        con = htd.connect(names[0])

        try:
            del con[None]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table del table (%s)" % e)
        else:
            raise ValueError("No errors using table del table")
        finally:
            con.close()

        #-------------------------------------------------------------- DB Test

        # !Exists
        con = htd.connect(names[1])

        self.assertFalse("abc" in list(con.keys()),
                         "Table exists - prepare fail")

        try:
            del con["abc"]
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, del !exist table (%s)" % e)
        else:
            raise ValueError("No errors using del !exist table")

        self.assertFalse("abc" in list(con.keys()),
                         "Table exists, when it shouldn't")

        con.close()

        # Exists
        con = htd.connect(names[1])

        # Prepare test
        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        self.assertTrue("abc" in list(con.keys()), "Table has not been added")

        # Check TAB before
        cur0 = con[None]
        self.assertFalse(cur0._file.header["TAB"], "Should not be a Table")
        cur0.close()

        cur1 = con["abc"]
        self.assertTrue(cur1._file.header["TAB"], "Should be a table")
        cur1.close()

        del con["abc"]

        self.assertFalse("abc" in list(con.keys()),
                         "Table has not been deleted")

        # Check TAB after
        cur0 = con[None]
        self.assertFalse(cur0._file.header["TAB"], "Should not be a Table")
        cur0.close()

        con1 = htd.connect(names[3])
        cur1 = con1[None]
        self.assertIs(cur1._file.header["TAB"], None, "Should be a table")
        cur1.close()
        con1.close()

        con.close()

    def test_del_extreme(self):
        """
        Test __delitem__ extreme
        - Incorrect data type, Null
        """

        con = htd.connect(names[1])

        # NULL
        try:
            del con[None]
        except TypeError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, del NULL (%s)" % e)
        else:
            raise ValueError("No errors using del NULL")

        # Bad dtype
        try:
            del con[123]
        except ValueError:  # TODO: Should be TypeError?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, del bad dtype (%s)" % e)
        else:
            raise ValueError("No errors using del bad dtype")

        con.close()

    def test_get(self):
        """
        Test __getitem__
        table, db
        - None, exists, !exists - check cursor.location
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        # !exists
        try:
            cur = con["abc"]
            cur.close()
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table get !exist (%s)" % e)
        else:
            raise ValueError("No errors using table get !exist")

        # should return master for None
        cur = con[None]

        self.assertEqual(cur.location, names[0], "None should be master")

        cur.close()

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # !exists
        try:
            cur = con["abc"]
            cur.close()
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, db get !exist (%s)" % e)
        else:
            raise ValueError("No errors using db get !exist")

        # should return master for None
        cur = con[None]

        self.assertEqual(cur.location, names[1], "None should be master")

        cur.close()

        # Prepare test
        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        self.assertTrue("abc" in list(con.keys()), "Table has not been added")

        # should return child table
        cur = con["abc"]

        self.assertEqual(cur.location, names[3], "Subtable is incorrect")

        cur.close()

        con.close()

    def test_get_extreme(self):
        """
        Test __gelitem__ extreme
        Table, DB
        - Incorrect data type
        """

        #----------------------------------------------------------- Test Table
        con = htd.connect(names[0])

        # Bad dtype
        try:
            cur = con[123]
            cur.close()
        except ValueError:  # TODO: Should be TypeError?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table get bad dtype (%s)" %
                             e)
        else:
            raise ValueError("No errors using table get bad dtype")

        con.close()

        #-------------------------------------------------------------- Test DB
        con = htd.connect(names[1])

        # Bad dtype
        try:
            cur = con[123]
            cur.close()
        except ValueError:  # TODO: Should be TypeError?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, get DB bad dtype (%s)" % e)
        else:
            raise ValueError("No errors using DB get bad dtype")

        con.close()

    def test_iter(self):
        """
        Test iter
        - table, db
        DB -> no tables, tables
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        self.assertEqual(list(con), [], "List should be empty, table.")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # No tables
        self.assertListEqual(list(con), [], "List should be empty, no tables.")

        # Prepare test
        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        #
        # BUG lp:1084645
        #

        # Tables
        self.assertListEqual(list(con),
                             [{"name": "abc", "location": names[3]}],
                             "List is incorrect, db - one table")

        con.close()

    def test_len(self):
        """
        Test len
        - table (always 1), db (no tables = 0, tables (1) = 1)
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        self.assertEqual(len(con), 1, "Table connection len should equal 1")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # No tables
        self.assertEqual(len(con), 0, "Len should be 0, no tables.")

        # Prepare test
        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        #
        # BUG lp:1084645
        #

        # Tables
        self.assertEqual(len(con), 1, "Len should be 1, db - one table")

        con.close()

    def test_repr(self):
        """
        Test repr
        - table, db
        - open/closed
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        # Open
        self.assertEqual(repr(con), "<HTD%s Connection (guest@%s), DB Mode"
                         " False>" % (htd.__version__, names[0]),
                         "Open table repr incorrect")

        con.close()

        # Closed
        self.assertEqual(repr(con), "<HTD%s Closed Connection>" %
                         htd.__version__, "Closed table repr incorrect")

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Open
        self.assertEqual(repr(con), "<HTD%s Connection (guest@%s), DB Mode"
                         " True>" % (htd.__version__, names[1]),
                         "Open DB repr incorrect")

        con.close()

        # Closed
        self.assertEqual(repr(con), "<HTD%s Closed Connection>" %
                         htd.__version__, "Closed DB repr incorrect")

    def test_set(self):
        """
        Test set
        - table, db
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        tbl = htd.create(names[6], identifier="id", attributes=atts)

        try:
            con["abc"] = tbl
        except htd.DatabaseError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table set child (%s)" % e)
        else:
            raise ValueError("No errors using table set child")

        tbl.close()

        self.assertListEqual(list(con), [], "Table should have no children")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        child = htd.create(names[4], identifier="id", attributes=atts)

        con["abc"] = child

        # Check child closed
        self.assertFalse(child.alive, "Child should have been closed")

        # Check child has been added
        self.assertEqual(len(con), 1, "Should be 1 table in DB")

        #
        # BUG lp:1084645
        #

        # Tables
        self.assertListEqual(list(con),
                             [{"name": "abc", "location": names[4]}],
                             "List is incorrect, db - one table")

        con.close()

    def test_set_extreme(self):
        """
        Test set
        - db
        child - None, wrong data type, is a DB
        name - None, wrong data type, already exists
        """

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # CHILD TESTS

        # Child is None
        try:
            con["abc"] = None
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, db set None (%s)" % e)
        else:
            raise ValueError("No errors using db set None")

        # Child wrong data type
        try:
            con["abc"] = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, db set bad dtype (%s)" % e)
        else:
            raise ValueError("No errors using db set bad dtype")

        # Child is DB
        tbl = htd.connect(names[1])

        try:
            con["abc"] = tbl
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, db set db (%s)" % e)
        else:
            raise ValueError("No errors using db set db")

        tbl.close()

        # NAME TESTS

        # Name is None
        tbl = htd.connect(names[0])

        try:
            con[None] = tbl
        except TypeError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, db set name None (%s)" % e)
        else:
            raise ValueError("No errors using db set name None")

        tbl.close()

        #
        # BUG lp:1086149
        #

        # Name wrong data type
        tbl = htd.connect(names[0])

        try:
            con[123] = tbl
        except ValueError:
            self.assertTrue(tbl.alive, "Child table should be alive")
        except Exception as e:
            self.assertTrue(tbl.alive, "Child table should be alive")
            raise ValueError("Incorrect exception, db set bad dtype (%s)" % e)
        else:
            raise ValueError("No errors using db set name bad dtype")

        tbl.close()

        # Name exists should overwrite
        con["abc"] = htd.connect(names[0])

        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        child = htd.create(names[4], identifier="id", attributes=atts)

        con["abc"] = child

        child.close()
        con.close()

        return

    def test_auth(self):
        """
        Test auth
        - table, db
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        auth = con.authentication

        self.assertTrue(auth.alive, "Auth is not alive")
        self.assertEqual(auth.location, names[0], "Auth location incorrect")

        con.close()

        self.assertFalse(auth.alive, "Auth has not closed")

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        auth = con.authentication

        self.assertTrue(auth.alive, "Auth is not alive")
        self.assertEqual(auth.location, names[1], "Auth location incorrect")

        con.close()

        self.assertFalse(auth.alive, "Auth has not closed")

    def test_children(self):
        """
        Test children
        - table, db
        - empty, cursors, auths
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        # Empty
        self.assertListEqual(con.children, [], "No children")

        # Open cursors
        cur0 = con[None]
        cur1 = con[None]

        # Open Auths
        auth0 = con.authentication
        auth1 = con.authentication

        # Children exist
        self.assertEqual(len(con.children), 4, "Children size is incorrect")

        # Close cursor & auth
        cur0.close()
        auth0.close()

        self.assertEqual(len(con.children), 2, "Children size is incorrect")

        con.close()

        self.assertFalse(cur1.alive, "Cursor not closed")
        self.assertFalse(auth1.alive, "Auth not closed")

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Empty
        self.assertListEqual(con.children, [], "No children")

        # Open cursors
        cur0 = con[None]
        cur1 = con[None]

        # Open Auths
        auth0 = con.authentication
        auth1 = con.authentication

        # Children exist
        self.assertEqual(len(con.children), 4, "Children size is incorrect")

        # Close cursor & auth
        cur0.close()
        auth0.close()

        self.assertEqual(len(con.children), 2, "Children size is incorrect")

        con.close()

        self.assertFalse(cur1.alive, "Cursor not closed")
        self.assertFalse(auth1.alive, "Auth not closed")

    def test_close(self):
        """
        Test close
        - Table, db
        - alive
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        con.close()

        self.assertIs(con.alive, False, "Cursor is still alive")

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        con.close()

        self.assertIs(con.alive, False, "Cursor is still alive")

    def test_delete(self):
        """
        Test delete
        - Table, DB (+/- children)
        - Alive, Path exists
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        con.delete()

        self.assertFalse(con.alive, "Connection not closed")
        self.assertFalse(exists(names[0]), "File should not exist")

        #-------------------------------------------------------------- DB Test

        # No tables
        con = htd.connect(names[1])

        con.delete()

        self.assertFalse(con.alive, "Connection not closed")
        self.assertFalse(exists(names[1]), "File should not exist")

        # Prepare
        self.tearDown()
        self.setUp()

        # Tables - delete children
        con = htd.connect(names[1])

        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        con.delete(children=True)

        self.assertFalse(con.alive, "Connection not closed")
        self.assertFalse(exists(names[1]), "File should not exist")
        self.assertFalse(exists(names[3]), "File should not exist")

        # Prepare
        self.tearDown()
        self.setUp()

        # Tables - keep children
        con = htd.connect(names[1])

        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        con.delete(children=False)

        self.assertFalse(con.alive, "Connection not closed")
        self.assertFalse(exists(names[1]), "File should not exist")
        self.assertTrue(exists(names[3]), "File should exist")

    def test_is_db(self):
        """
        Test is_db
        - table, db
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        self.assertFalse(con.is_db, "Should not be a DB")

        con.close()

        #-------------------------------------------------------------- DB Test

        con = htd.connect(names[1])

        self.assertTrue(con.is_db, "Should be a DB")

        con.close()

    def test_link(self):
        """
        Test link
        - table, db
        - !key-key, key-att, !att-att, att-key
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        try:
            con.link("tbl1.att1", "tbl2.att1")
        except ValueError:  # TODO: Value or Database error?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table list keys (%s)" % e)
        else:
            raise ValueError("No errors using table list keys")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="a", attributes=atts)

        atts = {"d": htd.Attribute(htd.DataType.INTEGER),
                "e": htd.Attribute(htd.DataType.INTEGER),
                "a": htd.Attribute(htd.DataType.INTEGER)}
        con["dea"] = htd.create(names[4], identifier="d", attributes=atts)

        # Key-Key
        try:
            con.link("abc.a", "dea.d")
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, link key-key (%s)" % e)
        else:
            raise ValueError("No errors using link key-key")

        # key-att
        con.link("abc.a", "dea.a")

        self.assertListEqual(list(con.relation), [(("dea", "a"),
                                                   ("abc", "a"))],
                             "Relation is not correct")

        con.unlink("abc.a", "dea.a")

        self.assertListEqual(list(con.relation), [], "Relation not cleared")

        # att-att
        try:
            con.link("abc.a", "dea.d")
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, link att-att (%s)" % e)
        else:
            raise ValueError("No errors using link att-att")

        # att-key
        con.link("dea.a", "abc.a")

        self.assertListEqual(list(con.relation), [(("dea", "a"),
                                                   ("abc", "a"))],
                             "Relation is not correct")

        con.unlink("dea.a", "abc.a")

        self.assertListEqual(list(con.relation), [], "Relation not cleared")

        con.close()

    def test_link_extreme(self):
        """
        Test link extreme
        - Incorrect table/att name, different attribute settings, data exists
        """

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="a", attributes=atts)

        atts = {"d": htd.Attribute(htd.DataType.INTEGER),
                "e": htd.Attribute(htd.DataType.INTEGER),
                "a": htd.Attribute(htd.DataType.INTEGER)}
        con["dea"] = htd.create(names[4], identifier="d", attributes=atts)

        # Bad table name
        try:
            con.link("123.a", "dea.d")
        except ValueError:  # TODO: keyerror?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Bad tbl name (%s)" % e)
        else:
            raise ValueError("No errors using Bad tbl name")

        # Bad att name
        try:
            con.link("abc.1", "dea.d")
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Bad att name (%s)" % e)
        else:
            raise ValueError("No errors using Bad att name")

        # Different att settings
        att = htd.Attribute(htd.DataType.STRING)
        att.size = 100

        atts = {"f": htd.Attribute(htd.DataType.INTEGER),
                "g": htd.Attribute(htd.DataType.INTEGER), "a": att}
        con["fga"] = htd.create(names[5], identifier="f", attributes=atts)

        try:
            con.link("abc.a", "fga.a")
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, att1 != att2 (%s)" % e)
        else:
            raise ValueError("No errors using att1 != att2")

        # Data exists in many but not one
        cur = con["dea"]

        cur[1] = {"e": 1, "a": 1}
        cur[2] = {"e": 2, "a": 2}
        cur[3] = {"e": 3, "a": 3}

        cur.close()

        try:
            con.link("abc.a", "dea.a")
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, data exists (%s)" % e)
        else:
            raise ValueError("No errors using data exists")

        # Data exists both
        cur = con["abc"]

        cur[1] = {"b": 1, "c": 1}
        cur[2] = {"b": 2, "c": 2}
        cur[3] = {"b": 3, "c": 3}

        cur.close()

        con.link("abc.a", "dea.a")

        self.assertListEqual(list(con.relation), [(("dea", "a"),
                                                   ("abc", "a"))],
                             "Relation is not correct")

        con.unlink("abc.a", "dea.a")

        con.close()

    def test_location(self):
        """
        Test location
        - table, db
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        self.assertEqual(con.location, names[0], "Table location not same")

        con.close()

        #-------------------------------------------------------------- DB Test

        con = htd.connect(names[1])

        self.assertEqual(con.location, names[1], "DB location not same")

        con.close()

    def test_keys(self):
        """
        Test keys
        - table, db
        - no tables, many tables
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        try:
            list(con.keys())
        except ValueError:  # TODO: Value or Database error?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table list keys (%s)" % e)
        else:
            raise ValueError("No errors using table list keys")

        con.close()

        #-------------------------------------------------------------- DB Test

        con = htd.connect(names[1])

        # No tables
        self.assertListEqual(list(con.keys()), [], "Should be no tables")

        atts = {"id": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)

        # Many tables
        self.assertListEqual(list(con.keys()), ["abc"], "Should be tables")
        con.close()

    def test_relation(self):
        """
        Test relation
        - db
        - relation exists, no relations exist
        """

        con = htd.connect(names[1])

        # No Relations
        self.assertListEqual(list(con.relation), [], "Should be no relations")

        # Prepare - make relation
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="a", attributes=atts)

        atts = {"d": htd.Attribute(htd.DataType.INTEGER),
                "e": htd.Attribute(htd.DataType.INTEGER),
                "a": htd.Attribute(htd.DataType.INTEGER)}
        con["dea"] = htd.create(names[4], identifier="d", attributes=atts)

        con.link("abc.a", "dea.a")

        # Many tables
        self.assertListEqual(list(con.relation), [(("dea", "a"),
                                                   ("abc", "a"))],
                             "Relation is not correct")

        con.close()

        return

    def test_unlink(self):
        """
        Test unlink
        - table, db
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        try:
            con.unlink("tbl1.att1", "tbl2.att1")
        except ValueError:  # TODO: Value or Database error?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, table list keys (%s)" % e)
        else:
            raise ValueError("No errors using table list keys")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        self.assertListEqual(list(con.relation), [], "Should be no relations")

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="a", attributes=atts)

        atts = {"d": htd.Attribute(htd.DataType.INTEGER),
                "e": htd.Attribute(htd.DataType.INTEGER),
                "a": htd.Attribute(htd.DataType.INTEGER)}
        con["dea"] = htd.create(names[4], identifier="d", attributes=atts)

        con.link("abc.a", "dea.a")

        self.assertListEqual(list(con.relation), [(("dea", "a"),
                                                   ("abc", "a"))],
                             "Relation is not correct")

        con.unlink("abc.a", "dea.a")

        self.assertListEqual(list(con.relation), [],
                             "Unlink has not worked")

        con.close()

    def test_unlink_extreme(self):
        """
        Test unlink extreme
        - Incorrect table/att name, different attribute settings, data exists
        """

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="a", attributes=atts)

        atts = {"d": htd.Attribute(htd.DataType.INTEGER),
                "e": htd.Attribute(htd.DataType.INTEGER),
                "a": htd.Attribute(htd.DataType.INTEGER)}
        con["dea"] = htd.create(names[4], identifier="d", attributes=atts)

        # Bad table name
        try:
            con.unlink("123.a", "dea.d")
        except ValueError:  # TODO: keyerror?
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Bad tbl name (%s)" % e)
        else:
            raise ValueError("No errors using Bad tbl name")

        # Bad att name
        try:
            con.unlink("abc.1", "dea.d")
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Bad att name (%s)" % e)
        else:
            raise ValueError("No errors using Bad att name")

        # Data exists
        cur = con["dea"]

        cur[1] = {"e": 1, "a": 1}
        cur[2] = {"e": 2, "a": 2}
        cur[3] = {"e": 3, "a": 3}

        cur.close()
        cur = con["abc"]

        cur[1] = {"b": 1, "c": 1}
        cur[2] = {"b": 2, "c": 2}
        cur[3] = {"b": 3, "c": 3}

        cur.close()

        con.link("abc.a", "dea.a")
        con.unlink("abc.a", "dea.a")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        con.close()
