#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class STRING():
    dtype = htd.DataType.STRING

    def unsan_san(self, value):
        return value

    def comp_values(self):
        return ((chr(i), chr(i)) for i in range(0, 256))

    def values(self):
        return (chr(i) for i in range(0, 256))

STRING = STRING()


class CursorTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        htd.create(names[1],
                   identifier=["id1", "id2"],
                   attributes={"id1": htd.Attribute(htd.DataType.STRING),
                               "id2": htd.Attribute(htd.DataType.STRING)},
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[2]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test test
        - Open from Cursor object
        """

        #----------------------------------------------------------- Test table
        con = htd.connect(names[0])

        cur = con[None]

        self.assertTrue(cur.alive, "Cur not alive")
        self.assertEqual(cur.location, con.location, "Location not correct")
        self.assertEqual(cur.table, None, "Table is incorrect")
        self.assertEqual(cur.connection, con, "Auth connection incorrect")
        self.assertIn(cur, con.children, "Cur not in connection children")

        cur.close()
        con.close()

        #-------------------------------------------------------------- Test DB
        con = htd.connect(names[2])

        # Prepare
        atts = {"id": htd.Attribute(htd.DataType.STRING)}
        con["abc"] = htd.create(names[3], identifier="id", attributes=atts)
        con["def"] = htd.create(names[4], identifier="id", attributes=atts)

        con_list = [{'location': names[3], 'name': 'abc'},
                    {'location': names[4], 'name': 'def'}]

        self.assertListEqual(list(con), con_list, "Prepare failed")

        # Test table abc
        cur_abc = con["abc"]
        self.assertTrue(cur_abc.alive, "Cur not alive")
        self.assertEqual(cur_abc.location, names[3], "Location not correct")
        self.assertEqual(cur_abc.table, "abc", "Table is incorrect")
        self.assertEqual(cur_abc.connection, con, "Connection incorrect")
        self.assertIn(cur_abc, con.children, "Cur not in connection children")
        cur_abc.close()

        # Test table def
        cur_def = con["def"]
        self.assertTrue(cur_def.alive, "Cur not alive")
        self.assertEqual(cur_def.location, names[4], "Location not correct")
        self.assertEqual(cur_def.table, "def", "Table is incorrect")
        self.assertEqual(cur_def.connection, con, "Connection incorrect")
        self.assertIn(cur_def, con.children, "Cur not in connection children")
        cur_def.close()

        # Test table deleted
        unlink(names[4])

        try:
            cur_def = con["def"]
        except htd.ConnectionError:
            pass
        else:
            raise ValueError("Test failed table should not exist")

        con.close()

    def test_contains(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        for i in STRING.values():  # Test all values !exist
            self.assertFalse(i in cur, "Key should not exist in table")

        cur["b"] = {}  # Add one value

        for i in STRING.values():  # Test all values
            if i == "b":
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        cur["d"] = {}  # Add one value

        for i in STRING.values():  # Test all values
            if i in ("b", "d"):
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        # Test NULL value
        try:
            (None in cur)
        except:
            pass
        else:
            raise KeyError("NULL value check should fail.")

        # Test Empty value
        cur[""] = {}

        l = list(STRING.values())
        l.append("")

        for i in l:  # Test all values
            if i in ("b", "d", ""):
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        cur.close()
        con.close()

        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        for i in STRING.comp_values():  # Test all values !exist
            self.assertFalse(i in cur, "Key should not exist in table")

        cur[("c", "c")] = {}  # Add one value

        for i in STRING.comp_values():  # Test all values
            if i == ("c", "c"):
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        cur[("e", "e")] = {}  # Add one value

        for i in STRING.comp_values():  # Test all values
            if i in [("c", "c"), ("e", "e")]:
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        # Test NULL value
        try:
            ((None, None) in cur)
        except:
            pass
        else:
            raise KeyError("NULL value check should fail.")

        # Test Empty value
        cur[("", "")] = {}

        l = list(STRING.comp_values())
        l.append(("", ""))

        for i in l:  # Test all values
            if i in [("c", "c"), ("e", "e"), ("", "")]:
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        cur.close()
        con.close()

    def test_delitem(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        for i in STRING.values():  # Test all values !exist
            try:
                del cur[i]
            except KeyError:
                pass
            else:
                raise ValueError("Key should not exist in table")

        cur["a"] = {}
        cur["b"] = {}
        cur["c"] = {}

        for i in STRING.values():  # Test all values
            if i in ("a", "b", "c"):
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        self.assertEqual(len(cur), 3, "Length of table incorrect")

        del cur["b"]

        for i in STRING.values():  # Test all values
            if i in ("a", "c"):
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        self.assertEqual(len(cur), 2, "Length of table incorrect")

        # Test NULL value
        try:
            del cur[None]
        except:
            pass
        else:
            raise KeyError("NULL value check should fail.")

        # Test Empty value
        cur[""] = {}

        l = list(STRING.values())
        l.append("")

        self.assertTrue("" in cur, "Key should exist in table")
        self.assertEqual(len(cur), 3, "Length of table incorrect")

        del cur[""]

        for i in l:  # Test all values
            if i in ("a", "c"):
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        self.assertEqual(len(cur), 2, "Length of table incorrect")

        cur.close()
        con.close()

        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        for i in STRING.comp_values():  # Test all values !exist
            try:
                del cur[i]
            except KeyError:
                pass
            else:
                raise ValueError("Key should not exist in table")

        cur[("a", "a")] = {}
        cur[("b", "b")] = {}
        cur[("c", "c")] = {}

        for i in STRING.comp_values():  # Test all values
            if i in [("a", "a"), ("b", "b"), ("c", "c")]:
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        self.assertEqual(len(cur), 3, "Length of table incorrect")

        del cur[("b", "b")]

        for i in STRING.comp_values():  # Test all values
            if i in [("a", "a"), ("c", "c")]:
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        self.assertEqual(len(cur), 2, "Length of table incorrect")

        # Test NULL value
        try:
            del cur[(None, None)]
        except:
            pass
        else:
            raise KeyError("NULL value check should fail.")

        # Test Empty value
        cur[("", "")] = {}

        l = list(STRING.comp_values())
        l.append(("", ""))

        self.assertTrue(("", "") in cur, "Key should exist in table")
        self.assertEqual(len(cur), 3, "Length of table incorrect")

        del cur[("", "")]

        for i in l:  # Test all values
            if i in [("a", "a"), ("c", "c")]:
                self.assertTrue(i in cur, "Key should exist in table")
            else:
                self.assertFalse(i in cur, "Key should not exist in table")

        self.assertEqual(len(cur), 2, "Length of table incorrect")

        cur.close()
        con.close()

    def test_delslice(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        # Test with start > end 1:0 (should error)
        try:
            del cur[1:0]
        except ValueError:
            pass
        else:
            raise ValueError("Error should have occurred start > stop")

        # Test with start == end 1:1 (should error)
        try:
            del cur[1:1]
        except ValueError:
            pass
        else:
            raise ValueError("Error should have occurred start == stop")

        self.assertEqual(len(cur), 0, "Table not empty")

        cur["a"] = {}
        cur["b"] = {}
        cur["c"] = {}
        cur["d"] = {}
        cur["e"] = {}

        self.assertEqual(len(cur), 5, "Table should have 5 items")

        del cur[1:2]

        self.assertEqual(len(cur), 4, "Table should have 4 items")
        self.assertNotIn("b", cur, "b should be deleted")

        del cur[3:]

        self.assertEqual(len(cur), 3, "Table should have 3 items")
        self.assertNotIn("e", cur, "e should be deleted")

        del cur[:1]

        self.assertEqual(len(cur), 2, "Table should have 2 items")
        self.assertNotIn("a", cur, "a should be deleted")

        del cur[:]

        self.assertEqual(len(cur), 0, "Table not empty")
        self.assertNotIn("c", cur, "c should be deleted")
        self.assertNotIn("d", cur, "d should be deleted")

        cur.close()
        con.close()

        #--------------------------------------------------------- Mulikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertEqual(len(cur), 0, "Table not empty")

        cur[("a", "a")] = {}
        cur[("b", "b")] = {}
        cur[("c", "c")] = {}
        cur[("d", "d")] = {}
        cur[("e", "e")] = {}

        self.assertEqual(len(cur), 5, "Table should have 5 items")

        del cur[1:2]

        self.assertEqual(len(cur), 4, "Table should have 4 items")
        self.assertNotIn(("b", "b"), cur, "b should be deleted")

        del cur[3:]

        self.assertEqual(len(cur), 3, "Table should have 3 items")
        self.assertNotIn(("e", "e"), cur, "e should be deleted")

        del cur[:1]

        self.assertEqual(len(cur), 2, "Table should have 2 items")
        self.assertNotIn(("a", "a"), cur, "a should be deleted")

        del cur[:]

        self.assertEqual(len(cur), 0, "Table not empty")
        self.assertNotIn(("c", "c"), cur, "c should be deleted")
        self.assertNotIn(("d", "d"), cur, "d should be deleted")

        cur.close()
        con.close()

    def test_getitem(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        for i in STRING.values():
            try:
                cur[i]
            except KeyError:
                pass
            else:
                raise ValueError("Key should not exist")

        for i in STRING.values():
            cur[i] = {}
            self.assertEqual(cur[i], {"id": i}, "Record is not correct")
            self.assertEqual(cur[i].version, 0, "Version is incorrect")

        # Test NULL value
        try:
            cur[None]
        except:
            pass
        else:
            raise KeyError("NULL value check should fail.")

        # Test empty
        cur[""] = {}
        self.assertEqual(cur[""], {"id": ""}, "Empty record is not correct")

        cur.close()
        con.close()

        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        for i in STRING.comp_values():
            try:
                cur[i]
            except KeyError:
                pass
            else:
                raise ValueError("Key should not exist")

        for a, b in STRING.comp_values():
            cur[(a, b)] = {}
            self.assertEqual(cur[(a, b)], {"id1": a, "id2": b},
                             "Record is not correct")
            self.assertEqual(cur[(a, b)].version, 0, "Version is incorrect")

        # Test NULL value
        try:
            cur[(None, None)]
        except:
            pass
        else:
            raise KeyError("NULL value check should fail.")

        # Test empty
        cur[("", "")] = {}
        self.assertEqual(cur[("", "")], {"id1": "", "id2": ""},
                         "Empty record is not correct")

        cur.close()
        con.close()

    def test_iter(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertListEqual(list(cur), [], "List should be empty")

        for i in STRING.values():
            cur[i] = {}

        self.assertListEqual(list(cur), list(STRING.values()),
                             "List is no correct")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertListEqual(list(cur), [], "List should be empty")

        for i in STRING.comp_values():
            cur[i] = {}

        self.assertListEqual(list(cur), [{"id1": a, "id2": b} for a, b in
                                         STRING.comp_values()],
                             "List is no correct")

        cur.close()
        con.close()

    def test_len(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertEqual(len(cur), 0, "Length should be 0")

        cur["a"] = {}
        cur["b"] = {}

        self.assertEqual(len(cur), 2, "Length should be 2")

        del cur["a"]

        self.assertEqual(len(cur), 1, "Length should be 1")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertEqual(len(cur), 0, "Length should be 0")

        cur[("a", "b")] = {}
        cur[("c", "d")] = {}

        self.assertEqual(len(cur), 2, "Length should be 2")

        del cur[("a", "b")]

        self.assertEqual(len(cur), 1, "Length should be 1")

        cur.close()
        con.close()

    def test_repr(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        repr_string = "<HTD%s Cursor (%s)>" % (htd.__version__, names[0])

        self.assertEqual(repr(cur), repr_string, "Repr incorrect")

        cur.close()

        repr_string = "<HTD%s Closed Cursor>" % htd.__version__

        self.assertEqual(repr(cur), repr_string, "Repr incorrect")

        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        repr_string = "<HTD%s Cursor (%s)>" % (htd.__version__, names[1])

        self.assertEqual(repr(cur), repr_string, "Repr incorrect")

        cur.close()

        repr_string = "<HTD%s Closed Cursor>" % htd.__version__

        self.assertEqual(repr(cur), repr_string, "Repr incorrect")

        con.close()

    def test_setitem(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        att = cur.attributes
        att["data"] = htd.Attribute(htd.DataType.INTEGER)
        att["1"] = htd.Attribute()
        att["2"] = htd.Attribute()
        att["3"] = htd.Attribute()

        # Test overwrite NULLs exist
        cur["1"] = {}
        self.assertEqual(cur["1"].version, 0, "Version is incorrect")

        cur["1"] = {"data": 123, "1": "abc", "2": None, "3": "def"}
        self.assertEqual(cur["1"].version, 1, "Version is incorrect")

        self.assertEqual(cur["1"], {"id": "1", "data": 123, "1": "abc",
                                    "2": None, "3": "def"}, "Record incorrect")

        cur.delete()

        del att["1"]
        del att["2"]
        del att["3"]

        self.assertEqual(len(cur), 0, "Length should be 0")

        for i, key in enumerate(STRING.values()):
            cur[key] = {"data": i}

            self.assertEqual(cur[key].version, 0, "Version is incorrect")
            self.assertEqual(cur[key], {"id": key, "data": i},
                             "Record incorrect")

        for i, key in enumerate(STRING.values()):
            cur[key] = {"data": i * 2}

            self.assertEqual(cur[key].version, 1, "Version is incorrect")
            self.assertEqual(cur[key], {"id": key, "data": i * 2},
                             "Record incorrect")

        # Test sending just key
        cur["aa"] = {}

        self.assertEqual(cur["aa"], {"id": "aa", "data": None},
                         "Record is incorrect")

        # Test record out of data
        cur["test"] = {"data": 123}
        self.assertEqual(cur["test"].version, 0, "Version is incorrect")

        record0 = cur["test"]

        record1 = cur["test"]
        record1["data"] = 456

        cur["test"] = record1
        self.assertEqual(cur["test"].version, 1, "Version is incorrect")

        try:
            cur["test"] = record0
        except htd.ConcurrencyError:
            pass
        else:
            raise ValueError("Setting old revision")

        record0.clear_version()
        cur["test"] = record0

        self.assertEqual(cur["test"].version, 2, "Version is incorrect")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        att = cur.attributes
        att["data"] = htd.Attribute(htd.DataType.INTEGER)

        self.assertEqual(len(cur), 0, "Length should be 0")

        for i, (a, b) in enumerate(STRING.comp_values()):
            cur[(a, b)] = {"data": i}

            self.assertEqual(cur[(a, b)].version, 0, "Version is incorrect")
            self.assertEqual(cur[(a, b)], {"id1": a, "id2": b, "data": i},
                             "Record incorrect")

        for i, (a, b) in enumerate(STRING.comp_values()):
            cur[(a, b)] = {"data": i * 2}

            self.assertEqual(cur[(a, b)].version, 1, "Version is incorrect")
            self.assertEqual(cur[(a, b)], {"id1": a, "id2": b, "data": i * 2},
                             "Record incorrect")

        cur.close()
        con.close()

    def test_attributes(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        atts = cur.attributes

        self.assertEqual(atts.connection, con, "Connection incorrect")
        self.assertEqual(atts.location, names[0], "Location incorrect")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        atts = cur.attributes

        self.assertEqual(atts.connection, con, "Connection incorrect")
        self.assertEqual(atts.location, names[1], "Location incorrect")

        cur.close()
        con.close()

    def test_close(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertTrue(cur.alive, "Cursor not alive")
        self.assertIn(cur, con.children, "Cursor not in connection children")

        cur.close()

        self.assertFalse(cur.alive, "Cursor not dead")
        self.assertNotIn(cur, con.children, "Cursor in connection children")

        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertTrue(cur.alive, "Cursor not alive")
        self.assertIn(cur, con.children, "Cursor not in connection children")

        cur.close()

        self.assertFalse(cur.alive, "Cursor not dead")
        self.assertNotIn(cur, con.children, "Cursor in connection children")

        con.close()

    def test_composite(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertFalse(cur.composite, "Table is single key")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertTrue(cur.composite, "Table is multikey")

        cur.close()
        con.close()

    def test_connection(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertEqual(cur.connection, con, "Connection is not correct")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertEqual(cur.connection, con, "Connection is not correct")

        cur.close()
        con.close()

    def test_convert(self):
        #----------------------------------------------------------- Table Test
        # Prepare
        con = htd.connect(names[0])

        cur = con[None]

        for i in range(100):
            cur[str(i)] = {}

        if isinstance(cur._file, htd.FileType.VARIABLE_WIDTH_FILE):
            cur.convert(names[5], htd.FileType.FIXED_WIDTH_FILE)
            cls = htd.FileType.FIXED_WIDTH_FILE
        elif isinstance(cur._file, htd.FileType.FIXED_WIDTH_FILE):
            cur.convert(names[5], htd.FileType.VARIABLE_WIDTH_FILE)
            cls = htd.FileType.VARIABLE_WIDTH_FILE
        else:
            cur.convert(names[5], htd.FileType.FIXED_WIDTH_FILE)
            cls = htd.FileType.FIXED_WIDTH_FILE

        cur.close()

        con5 = htd.connect(names[5])
        cur = con5[None]

        for i in range(100):
            self.assertIn(str(i), cur, "Value not in converted file")

        self.assertIsInstance(cur._file, cls, "Bad Instance (%s)" % cls)

        cur.close()

        con5.close()
        con.close()

        #-------------------------------------------------------------- DB Test
        # Prepare
        con = htd.connect(names[2])

        con["abc"] = htd.create(names[3], identifier="id",
                                attributes={"id": htd.Attribute()})

        con["def"] = htd.create(names[4], identifier="id",
                                attributes={"id": htd.Attribute(),
                                            "name": htd.Attribute()})

        con.link("abc.id", "def.name")

        cur = con["abc"]
        cur[1] = {}
        cur.close()

        cur = con["def"]

        for i in range(100):
            cur[str(i)] = {"name": 1}

        if isinstance(cur._file, htd.FileType.VARIABLE_WIDTH_FILE):
            cur.convert(names[6], htd.FileType.FIXED_WIDTH_FILE)
            cls = htd.FileType.FIXED_WIDTH_FILE
        elif isinstance(cur._file, htd.FileType.FIXED_WIDTH_FILE):
            cur.convert(names[6], htd.FileType.VARIABLE_WIDTH_FILE)
            cls = htd.FileType.VARIABLE_WIDTH_FILE
        else:
            cur.convert(names[6], htd.FileType.FIXED_WIDTH_FILE)
            cls = htd.FileType.FIXED_WIDTH_FILE

        cur.close()

        con5 = htd.connect(names[6])
        cur = con5[None]

        for i in range(100):
            self.assertIn(str(i), cur, "Value not in converted file")

        self.assertIsInstance(cur._file, cls, "Bad Instance (%s)" % cls)

        cur.close()

        con5.close()
        con.close()

    def test_copy(self):
        #----------------------------------------------------------- Table Test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertTrue(cur.copy(names[6]), "Copy should have occurred")

        con6 = htd.connect(names[6])

        self.assertEqual(con6.location, names[6], "Location is incorrect")

        con6.close()
        con.close()

        #-------------------------------------------------------------- DB Test

        # Prepare
        con = htd.connect(names[2])

        con["abc"] = htd.create(names[3], identifier="id",
                                attributes={"id": htd.Attribute()})

        con["def"] = htd.create(names[4], identifier="id",
                                attributes={"id": htd.Attribute(),
                                            "name": htd.Attribute()})

        con.link("abc.id", "def.name")

        cur = con["def"]

        # Test copy
        self.assertTrue(cur.copy(names[7]), "Copy should have occurred")

        con7 = htd.connect(names[7])

        self.assertEqual(con7.location, names[7], "Location is incorrect")

        con7.close()

        # Test copy - replace
        try:
            cur.copy(names[7])
        except IOError:
            pass
        else:
            raise ValueError("File should exist")

        self.assertTrue(cur.copy(names[7], True), "Copy should have occurred")

        con7 = htd.connect(names[7])

        self.assertEqual(con7.location, names[7], "Location is incorrect")

        con7.close()

        con.close()

    def test_delete(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.INTEGER)

        for i, key in enumerate(STRING.values()):
            cur[key] = {"data": i}

        deleted = cur.delete("data > {min}", dict(min=("data", 10)))

        self.assertGreater(deleted, 0, "Should have deleted something")

        self.assertEqual(len(cur), 11, "Length of table incorrect")

        deleted = cur.delete()

        self.assertEqual(deleted, 11, "Deleted should be 11")
        self.assertEqual(len(cur), 0, "Table should be empty")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.INTEGER)

        for i, key in enumerate(STRING.comp_values()):
            cur[key] = {"data": i}

        deleted = cur.delete("data > {min}", dict(min=("data", 10)))

        self.assertGreater(deleted, 0, "Should have deleted something")
        self.assertEqual(len(cur), 11, "Length of table incorrect")

        deleted = cur.delete()

        self.assertEqual(deleted, 11, "Deleted should be 11")
        self.assertEqual(len(cur), 0, "Table should be empty")

        cur.close()
        con.close()

    def test_identifier(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.STRING)

        self.assertEqual(cur.identifier, "id", "Identifier incorrect")

        cur.identifier = "data"

        self.assertEqual(cur.identifier, "data", "Identifier incorrect")

        # Test type
        try:
            cur.identifier = {}
        except ValueError:
            pass
        else:
            raise ValueError("Identifier should only allow lists")

        # Test with data in table
        cur["a"] = {}

        try:
            cur.identifier = "id"
        except ValueError:
            pass
        else:
            raise ValueError("Identifier should only allow when table empty")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.STRING)

        self.assertEqual(cur.identifier, ["id1", "id2"],
                         "Identifier incorrect")

        cur.identifier = ["id1", "data"]

        self.assertEqual(cur.identifier, ["id1", "data"],
                         "Identifier incorrect")

        cur.close()
        con.close()

    def test_index(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        cur["0"] = {}
        cur["1"] = {}
        cur["2"] = {}

        self.assertEqual(cur.index("0"), 0, "Index incorrect")

        try:
            cur.index("a")
        except KeyError:
            pass
        else:
            raise ValueError("Key should not exist")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        cur[("0", "0")] = {}
        cur[("1", "1")] = {}
        cur[("2", "2")] = {}

        self.assertEqual(cur.index(("0", "0")), 0, "Index incorrect")

        try:
            cur.index(("a", "a"))
        except KeyError:
            pass
        else:
            raise ValueError("Key should not exist")

        cur.close()
        con.close()

    def test_insert(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.STRING)

        self.assertEqual(len(cur), 0, "Table not empty")

        list_1 = [{"id": "a"}, {"id": "b", "data":"1"},
                  {"id": "c", "data": "2"}]

        cur.insert(list_1)

        list_1 = [{"id": "a", "data": None}, {"id": "b", "data":"1"},
                  {"id": "c", "data": "2"}]

        self.assertListEqual(list(cur), list_1, "Lists are not equal")
        self.assertEqual(len(cur), 3, "Length should be 3")

        try:
            cur.insert([{"id": "b"}])
        except KeyError:
            pass
        else:
            raise ValueError("Insert same id should fail")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.STRING)

        self.assertEqual(len(cur), 0, "Table not empty")

        list_1 = [{"id1": "a", "id2": "z"},
                  {"id1": "b", "id2": "y", "data":"1"},
                  {"id1": "c", "id2": "x", "data": "2"}]

        cur.insert(list_1)

        list_1 = [{"id1": "a", "id2": "z", "data": None},
                  {"id1": "b", "id2": "y", "data":"1"},
                  {"id1": "c", "id2": "x", "data": "2"}]

        self.assertListEqual(list(cur), list_1, "Lists are not equal")
        self.assertEqual(len(cur), 3, "Length should be 3")

        try:
            cur.insert([{"id1": "b", "id2": "y"}])
        except KeyError:
            pass
        else:
            raise ValueError("Insert same id should fail")

        cur.close()
        con.close()

    def test_keys(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertEqual(list(cur.keys()), [], "Keys should be empty")

        cur["a"] = {}
        cur["b"] = {}
        cur["c"] = {}

        keys = ["a", "b", "c"]

        self.assertEqual(list(cur.keys()), keys, "Keys incorrect")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertEqual(list(cur.keys()), [], "Keys should be empty")

        cur[("a", "0")] = {}
        cur[("b", "1")] = {}
        cur[("c", "2")] = {}

        keys = [{"id1": "a", "id2": "0"},
                {"id1": "b", "id2": "1"},
                {"id1": "c", "id2": "2"}]

        self.assertEqual(list(cur.keys()), keys, "Keys incorrect")

        cur.close()
        con.close()

    def test_location(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertEqual(cur.location, names[0], "Location is incorrect")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertEqual(cur.location, names[1], "Location is incorrect")

        cur.close()
        con.close()

    def test_rename(self):
        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])
        cur = con[None]

        # Test rename no replace
        cur.rename(names[7], replace=False)

        con7 = htd.connect(names[7])
        self.assertEqual(con7.location, names[7], "Location is incorrect")
        self.assertEqual(cur.location, names[7], "Location is incorrect")

        con7.close()

        cur.rename(names[0], replace=False)

        open(names[7], "w+").close()  # Create file

        # Test rename no replace - file exists
        try:
            cur.rename(names[7], replace=False)
        except IOError:
            pass
        else:
            raise ValueError("File exists and replace is False")

        # Test rename replace - file exists
        cur.rename(names[7], replace=True)

        con7 = htd.connect(names[7])
        self.assertEqual(con7.location, names[7], "Location is incorrect")
        self.assertEqual(cur.location, names[7], "Location is incorrect")

        con7.close()

        cur.rename(names[0], replace=False)

        con.close()

        #-------------------------------------------------------------- DB Test
        # Prepare
        con = htd.connect(names[2])

        con["abc"] = htd.create(names[3], identifier="id",
                                attributes={"id": htd.Attribute()})

        con["def"] = htd.create(names[4], identifier="id",
                                attributes={"id": htd.Attribute(),
                                            "name": htd.Attribute()})

        con.link("abc.id", "def.name")

        cur = con["def"]

        # Test rename no replace
        cur.rename(names[6], replace=False)

        self.assertEqual(cur.location, names[6], "Location is incorrect")

        cur8 = con["def"]
        self.assertEqual(cur8.location, names[6], "Location is incorrect")
        cur8.close()

        cur.rename(names[4], replace=False)

        open(names[6], "w+").close()  # Create file

        # Test rename no replace - file exists
        try:
            cur.rename(names[6], replace=False)
        except IOError:
            pass
        else:
            raise ValueError("File exists and replace is False")

        # Test rename replace - file exists
        cur.rename(names[6], replace=True)

        cur8 = con["def"]
        self.assertEqual(cur8.location, names[6], "Location is incorrect")
        self.assertEqual(cur.location, names[6], "Location is incorrect")

        cur8.close()

        # Test rename db file
        cur10 = con[None]
        cur10.rename(names[5])
        self.assertEqual(cur10.location, names[5], "Location is incorrect")
        self.assertEqual(con.location, names[5], "Location is incorrect")
        cur10.close()

        con.close()

    def test_select(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.INTEGER)

        cur["a"] = {"data": 1}
        cur["b"] = {"data": 2}
        cur["c"] = {"data": 3}
        cur["d"] = {"data": 4}
        cur["e"] = {"data": 5}

        l = cur.select("data")

        self.assertListEqual(list(l), [1, 2, 3, 4, 5], "List not correct")

        l = cur.select("data", start=1, stop=3)

        self.assertListEqual(list(l), [2, 3], "List not correct")

        l = cur.select("data", where="data < {max}",
                       args=dict(max=("data", 4)))

        self.assertListEqual(list(l), [1, 2, 3], "List not correct")

        l = cur.select(["id", "data"])

        db_list = [{"id": "a", "data": 1},
                   {"id": "b", "data": 2},
                   {"id": "c", "data": 3},
                   {"id": "d", "data": 4},
                   {"id": "e", "data": 5}]

        self.assertListEqual(list(l), db_list, "List not correct")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.INTEGER)

        cur[("a", "a")] = {"data": 1}
        cur[("b", "b")] = {"data": 2}
        cur[("c", "c")] = {"data": 3}
        cur[("d", "d")] = {"data": 4}
        cur[("e", "e")] = {"data": 5}

        l = cur.select("data")

        self.assertListEqual(list(l), [1, 2, 3, 4, 5], "List not correct")

        l = cur.select("data", start=1, stop=3)

        self.assertListEqual(list(l), [2, 3], "List not correct")

        l = cur.select("data", where="data < {max}",
                       args=dict(max=("data", 4)))

        self.assertListEqual(list(l), [1, 2, 3], "List not correct")

        l = cur.select(["id1", "id2", "data"])

        db_list = [{"id1": "a", "id2": "a", "data": 1},
                   {"id1": "b", "id2": "b", "data": 2},
                   {"id1": "c", "id2": "c", "data": 3},
                   {"id1": "d", "id2": "d", "data": 4},
                   {"id1": "e", "id2": "e", "data": 5}]

        self.assertListEqual(list(l), db_list, "List not correct")

        cur.close()
        con.close()

    def test_update(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.INTEGER)

        cur["a"] = {"data": 1}
        cur["b"] = {"data": 2}
        cur["c"] = {"data": 3}
        cur["d"] = {"data": 4}
        cur["e"] = {"data": 5}

        # Double everything
        updater = cur.update()

        for record in updater:
            self.assertEqual(record.version, 0, "Version is incorrect")
            record["data"] *= 2
            updater.send(record)

        # Confirm database
        db_list = [{"id": "a", "data": 2},
                   {"id": "b", "data": 4},
                   {"id": "c", "data": 6},
                   {"id": "d", "data": 8},
                   {"id": "e", "data": 10}]

        self.assertListEqual(list(cur), db_list, "List not correct")

        # Double < 5
        updater = cur.update(where="data < {max}",
                             args=dict(max=("data", 5)))

        for record in updater:
            self.assertEqual(record.version, 1, "Version is incorrect")
            record["data"] *= 2
            updater.send(record)

        # Confirm database
        db_list = [{"id": "a", "data": 4},
                   {"id": "b", "data": 8},
                   {"id": "c", "data": 6},
                   {"id": "d", "data": 8},
                   {"id": "e", "data": 10}]

        self.assertListEqual(list(cur), db_list, "List not correct")

        # Double Update indexes 2->4
        updater = cur.update(start=2, stop=4)

        for record in updater:
            record["data"] *= 2
            updater.send(record)

        # Confirm database
        db_list = [{"id": "a", "data": 4},
                   {"id": "b", "data": 8},
                   {"id": "c", "data": 12},
                   {"id": "d", "data": 16},
                   {"id": "e", "data": 10}]

        self.assertListEqual(list(cur), db_list, "List not correct")

        cur.close()
        con.close()

        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        atts = cur.attributes
        atts["data"] = htd.Attribute(htd.DataType.INTEGER)

        cur[("a", "a")] = {"data": 1}
        cur[("b", "b")] = {"data": 2}
        cur[("c", "c")] = {"data": 3}
        cur[("d", "d")] = {"data": 4}
        cur[("e", "e")] = {"data": 5}

        # Double everything
        updater = cur.update()

        for record in updater:
            self.assertEqual(record.version, 0, "Version is incorrect")
            record["data"] *= 2
            updater.send(record)

        # Confirm database
        db_list = [{"id1": "a", "id2": "a", "data": 2},
                   {"id1": "b", "id2": "b", "data": 4},
                   {"id1": "c", "id2": "c", "data": 6},
                   {"id1": "d", "id2": "d", "data": 8},
                   {"id1": "e", "id2": "e", "data": 10}]

        self.assertListEqual(list(cur), db_list, "List not correct")

        # Double < 5
        updater = cur.update(where="data < {max}",
                             args=dict(max=("data", 5)))

        for record in updater:
            self.assertEqual(record.version, 1, "Version is incorrect")
            record["data"] *= 2
            updater.send(record)

        # Confirm database
        db_list = [{"id1": "a", "id2": "a", "data": 4},
                   {"id1": "b", "id2": "b", "data": 8},
                   {"id1": "c", "id2": "c", "data": 6},
                   {"id1": "d", "id2": "d", "data": 8},
                   {"id1": "e", "id2": "e", "data": 10}]

        self.assertListEqual(list(cur), db_list, "List not correct")

        # Double Update indexes 2->4
        updater = cur.update(start=2, stop=4)

        for record in updater:
            record["data"] *= 2
            updater.send(record)

        # Confirm database
        db_list = [{"id1": "a", "id2": "a", "data": 4},
                   {"id1": "b", "id2": "b", "data": 8},
                   {"id1": "c", "id2": "c", "data": 12},
                   {"id1": "d", "id2": "d", "data": 16},
                   {"id1": "e", "id2": "e", "data": 10}]

        self.assertListEqual(list(cur), db_list, "List not correct")

        cur.close()
        con.close()

    def test_version(self):
        #------------------------------------------------------ Single key test
        con = htd.connect(names[0])
        cur = con[None]

        self.assertEqual(cur.version, htd.__version__, "Version is incorrect")

        cur.close()
        con.close()
        #-------------------------------------------------------- Multikey test
        con = htd.connect(names[1])
        cur = con[None]

        self.assertEqual(cur.version, htd.__version__, "Version is incorrect")

        cur.close()
        con.close()
