#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from __future__ import division
from .base import htd_engine as htd, names
from datetime import date, datetime, time
from struct import pack, unpack
from sys import version_info
import unittest


#----------------------------------------------------------- Data types classes
class BOOL():
    dtype = htd.DataType.BOOL

    def unsan_san(self, value):
        return value

    def values(self):
        for i in (True, False):
            yield i


class BYTES():
    dtype = htd.DataType.BYTES

    def unsan_san(self, value):
        return value

    if version_info < (3, 0):
        def values(self):
            return (bytes(chr(i)) for i in range(0, 256))
    else:
        def values(self):
            return (chr(i).encode("ISO-8859-1") for i in range(0, 256))


class STRING():
    dtype = htd.DataType.STRING

    def unsan_san(self, value):
        return value

    def values(self):
        return (chr(i) for i in range(0, 256))


class DATE():
    dtype = htd.DataType.DATE

    def unsan_san(self, value):
        return value

    def values(self):
        for y in range(1970, 2013):
            for m in range(1, 13):
                for d in range(1, 28):
                    yield date(y, m, d)


class DATETIME():
    dtype = htd.DataType.DATETIME

    def unsan_san(self, value):
        return value

    def values(self):
        for y in [2012]:
            for m in range(1, 13):
                for d in range(1, 10):
                    for h in range(1, 12):
                        for m in range(1, 15):
                            for s in range(1, 15):
                                yield datetime(y, m, d, h, m, s)


class TIME():
    dtype = htd.DataType.TIME

    def unsan_san(self, value):
        return value

    def values(self):
        for h in range(1, 24):
            for m in range(1, 60):
                for s in range(1, 60):
                    yield time(h, m, s)


class BYTE():
    dtype = htd.DataType.BYTE

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(-128, 128):
            yield i


class SHORT():
    dtype = htd.DataType.SHORT

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(-32768, 32768):
            yield i


class INTEGER():
    dtype = htd.DataType.INTEGER

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(-2147483648, -2147480000):
            yield i

        for i in range(-100, 100):
            yield i

        for i in range(2147480000, 2147483648):
            yield i


class LONG():
    dtype = htd.DataType.LONG

    def unsan_san(self, value):
        return value

    if version_info < (3, 0):
        def values(self):
            for i in range(-9223372036854775808, -9223372036854770000):
                yield long(i)

            for i in range(-100, 100):
                yield long(i)

            for i in range(9223372036854770000, 9223372036854775807):
                yield long(i)
    else:
        def values(self):
            for i in range(-9223372036854775808, -9223372036854770000):
                yield i

            for i in range(-100, 100):
                yield i

            for i in range(9223372036854770000, 9223372036854775807):
                yield i


class ULONG():
    dtype = htd.DataType.ULONG

    def unsan_san(self, value):
        return value

    if version_info < (3, 0):
        def values(self):
            for i in range(0, 1000):
                yield long(i)

            for i in range(18446744073709550000, 18446744073709551615):
                yield long(i)
    else:
        def values(self):
            for i in range(0, 1000):
                yield i

            for i in range(18446744073709550000, 18446744073709551615):
                yield i


class UINTEGER():
    dtype = htd.DataType.UINTEGER

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(0, 1000):
            yield i

        for i in range(4294960000, 4294967296):
            yield i


class USHORT():
    dtype = htd.DataType.USHORT

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(0, 65536):
            yield i


class UBYTE():
    dtype = htd.DataType.UBYTE

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(0, 256):
            yield i


class FLOAT():
    dtype = htd.DataType.FLOAT

    def unsan_san(self, value):
        return unpack("f", pack("f", value))[0]

    def values(self):
        for i in range(1000, 2000):
            yield i / 3


class DFLOAT():
    dtype = htd.DataType.DFLOAT

    def unsan_san(self, value):
        return value

    def values(self):
        for i in range(1000, 2000):
            yield i / 3


class PRESERVE():
    dtype = htd.DataType.PRESERVE

    def unsan_san(self, value):
        return value

    def values(self):
        for i in [True, b"\x01", "a", date(2012, 12, 12),
                  datetime(2012, 12, 12, 21, 50, 42), time(20, 49, 36), -12,
                  - (2 ** 15), -(2 ** 31), -(2 ** 63), 2 ** 63, 2 ** 31,
                  2 ** 15, 1.2345678]:
            yield i


#---------------------------------------------------- Unit tests for Data types
class DataTypeTest(unittest.TestCase):
    def test_attributes(self):
        """ Test Data Types have required attributes """

        for dtype in htd.Register.dtypes:
            a = getattr(htd.DataType, dtype)

            a.idn
            a.can_range
            a.sanitise
            a.size
            a.type_check
            a.unsanitise

    def test_formatting(self):
        """ Test Data Types have data formatted correctly """

        for dtype in htd.Register.dtypes:
            a = getattr(htd.DataType, dtype)

            self.assertIsInstance(a.idn, bytes, "IDN Should be BYTES")
            self.assertEqual(len(a.idn), 1, "IDN should be 1 byte")
            num = ord(a.idn)
            self.assertLessEqual(num, 255, "IDN should be <= 255")
            self.assertGreaterEqual(num, 0, "IDN should be >= 0")

            self.assertIn(a.can_range, (True, False),
                          "Can range should be True or False")
            self.assertGreaterEqual(a.size, 0, "Size should be >= 0")

    def test_data(self):
        """ Test the data type with Data """

        for i in (BOOL, BYTES, STRING, DATE, TIME, BYTE, SHORT, INTEGER, LONG,
                  UBYTE, USHORT, UINTEGER, ULONG, FLOAT, DFLOAT, PRESERVE):
            i = i()
            a = i.dtype

            for v in i.values():
                self.assertTrue(a.type_check(v), "Type check should be True")
                san = a.sanitise(v)
                self.assertIsInstance(san, bytes, "San data should be BYTES")

                t = a.unsanitise(a.sanitise(v))

                # TODO: fix FLOATING POINTS?
                if a in [htd.DataType.FLOAT, htd.DataType.DFLOAT]:
                    self.assertTrue(v == t or str(t) in str(v),
                                    "unsan(san(data)) == data %s" % a)
                else:
                    self.assertTrue(v == t, "unsan(san(data)) == data, type"
                                    " %s value %s -> %s" % (a, v, t))

    def test_compare(self):
        """ test the ordering of the data type is the same before and after """

        # PRESERVE NOT IN THIS TEST

        for i in (BOOL, BYTES, STRING, DATE, TIME, BYTE, SHORT, INTEGER, LONG,
                  UBYTE, USHORT, UINTEGER, ULONG, FLOAT, DFLOAT):
            i = i()
            a = i.dtype

            l2 = list(i.values())
            l2.reverse()

            for k, i1 in enumerate(i.values()):
                i2 = l2[k]

                san1 = a.sanitise(i1)
                san2 = a.sanitise(i2)

                self.assertEqual(san1 < san2, i1 < i2,
                                 "Order should be preserved (%s, %s)" %
                                 (i1, i2))


if __name__ == "__main__":
    unittest.main(verbosity=2)
