#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class EventTest(unittest.TestCase):
    events = {}

    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        htd.create(names[1],
                   identifier=["id1", "id2"],
                   attributes={"id1": htd.Attribute(htd.DataType.STRING),
                               "id2": htd.Attribute(htd.DataType.STRING)},
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[2]).close()

        #------------------------------------------------ Connect event handler
        htd.Event.connect(htd.EventType.ALL_EVENTS, htd.EventType.ALL_SIGNALS,
                          self.all_handler)

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def all_handler(self, idn, event, signal, args, kwargs):
        if event not in self.events:
            self.events[event] = {}

        if signal not in self.events:
            self.events[event][signal] = []

        self.events[event][signal].append((args, kwargs))

    def autotest(self, parent, name, tests):
        for meth, args, kwargs, expected_result, result_type, prop in tests:
            func = getattr(parent, meth)

            if args == [] and kwargs == {}:
                if prop:
                    result = func
                else:
                    result = func()
            else:
                result = func(*args, **kwargs)

            if result_type is not None:
                self.assertEqual(result_type(result), expected_result,
                                 "Result incorrect")
            else:
                self.assertEqual(result, expected_result, "Result incorrect")

            # Add class to args
            args.insert(0, parent)

            key = "%s.%s" % (name, meth)

            for event in [htd.EventType.BEFORE_LOCK,
                          htd.EventType.BEFORE_RUN]:
                self.assertIn((tuple(args), kwargs), self.events[event][key],
                              "%s failed %s" % (event, key))

            for event in [htd.EventType.AFTER_RUN,
                          htd.EventType.AFTER_UNLOCK]:
                self.assertIn((result, kwargs), self.events[event][key],
                              "%s failed %s" % (event, key))

    def test_connection(self):
        con = htd.connect(names[0])

        tests = [("__iter__", [], {}, [], list, False),
                 ("children", [], {}, [], list, True)]

        self.autotest(con, "Connection", tests)

        con.close()

    def test_cursor(self):
        con = htd.connect(names[0])

        cur = con[None]

        atts = cur.attributes
        atts["name"] = htd.Attribute(htd.DataType.STRING)

        tests = [("__contains__", ["1"], {}, False, bool, False),
                 ("__iter__", [], {}, [], list, False),
                 ("__len__", [], {}, 0, int, False),
                 ("identifier", [], {}, "id", str, True),
                 ("keys", [], {}, [], list, False),
                 ("version", [], {}, htd.__version__, float, True),
                 # Add record
                 ("__setitem__", ["1", {"name": "a"}], {}, None, None, False),
                 ("__contains__", ["1"], {}, True, bool, False),
                 ("__getitem__", ["1"], {}, {"id": "1", "name": "a"},
                  dict, False),
                 ("__iter__", [], {}, [{"id": "1", "name": "a"}], list, False),
                 ("__len__", [], {}, 1, int, False),
                 ("select", ["name"], {}, ["a"], list, False),
                 # Delete record
                 ("__delitem__", ["1"], {}, None, None, False),
                 ("__contains__", ["1"], {}, False, bool, False),
                 ("__iter__", [], {}, [], list, False),
                 ("__len__", [], {}, 0, int, False)]

        self.autotest(cur, "Cursor", tests)

        con.close()
