#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os import unlink
from os.path import exists
import unittest


#---------------------------------------------------- Unit tests for File types
class FileTypeTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_meths_and_atts(self):
        """ Test File Types have required attributes """

        header = {"TAB": None, "ATT": {}, "USR": {}, "KEY": [], "VER": 123,
                  "R1M": {}, "RM1": {}}

        for ftype in htd.Register.fhandlers:
            f = getattr(htd.FileType, ftype.__name__)

            # Test for attributes
            f.idn
            f.persistent

            #-------------------------------------------------- Open blank file
            f = f(names[0], "roadx", header)
            f.location

            # Used by Locking
            # f._confirm_funcs
            # f._files

            # From threading used by Locking
            # f._thread_locked  # Property
            # f._acquire  # method
            # f._release  # method

            # From header
            f.header  # Property

            # From Locking
            f.locked  # Property
            f.lock  # method
            f.unlock  # method

            # Main methods
            f.__contains__
            f.__delitem__
            f.__getitem__
            f.__len__
            f.__setitem__
            f.close
            f.delete
            f.get_index
            f.index
            f.insert  # From GetLineAlgorithm
            f.insert_record
            f.select
            f.update

            f.close()

            self.tearDown()  # Clear for next file type

    def test_formatting(self):
        """ Test File Types have data formatted correctly """

        for ftype in htd.Register.fhandlers:
            f = getattr(htd.FileType, ftype.__name__)

            # File format check
            self.assertIsInstance(f.idn, bytes, "file_format should be BYTES")
            self.assertEqual(len(f.idn), 1, "file_format should be 1 byte")

            num = ord(f.idn)
            self.assertLessEqual(num, 255, "file_format should be <= 255")
            self.assertGreaterEqual(num, 0, "file_format should be >= 0")

            # Persistent check
            self.assertIn(f.persistent, (True, False),
                          "persistent should be True or False")

    def test_init(self):
        """
        Test init method
        no location exists
            - header None (should fail)
            - real header
        location exists
            - header None
            - real header (should fail)
        """

        header = {"TAB": None, "ATT": {}, "USR": {}, "KEY": [], "VER": 123,
                  "R1M": {}, "RM1": {}}

        for ftype in htd.Register.fhandlers:
            f = getattr(htd.FileType, ftype.__name__)

            # No location exists and header is None
            try:
                t = f(names[0], "roadx", None)
            except IOError:
                pass
            except Exception as e:
                raise ValueError("Incorrect exception, header is None. (%s)"
                                 % e)
            else:
                t.close()
                raise ValueError("Header is None")

            self.tearDown()  # Clear files

            # No location exists and real header
            try:
                t = f(names[0], "roadx", header)
            except Exception as e:
                raise ValueError("Should pass, header is real. (%s)" % e)
            else:
                t.close()

            self.tearDown()  # Clear files

            # Prepare
            t = f(names[0], "roadx", header)
            t.close()

            # Location exists and header is None
            try:
                t = f(names[0], "roadx", None)
            except Exception as e:
                raise ValueError("Should pass, header is None. (%s)" % e)
            else:
                t.close()

            # Location exists and header is real
            try:
                t = f(names[0], "roadx", header)
            except IOError:
                pass
            except Exception as e:
                raise ValueError("Incorrect exception, header is real. (%s)"
                                 % e)
            else:
                t.close()
                raise ValueError("Header is real")

            self.tearDown()

if __name__ == "__main__":
    unittest.main(verbosity=2)
