#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class HTDTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_defaults(self):
        """
        Test the default settings of HTD
        __build_id_
        __version__
        AuthenticationError
        ConectionError
        DatabaseError
        IntegrityError
        VersionError
        Register
            default_dtype
            default_fhandler
        Config
            debug
            force_sync
            progress
            errors
        """

        # test exist
        self.assertIsNotNone(htd.__build_id__, "Build ID should have a value")
        self.assertIsNotNone(htd.__version__, "Version should have a value")

        # test exceptions
        htd.AuthenticationError
        htd.ConnectionError
        htd.DatabaseError
        htd.IntegrityError
        htd.VersionError

        # test register TODO: not consistent?
        self.assertEqual(htd.Register.default_dtype, "PRESERVE",
                         "Default data type should be preserve")

        self.assertEqual(htd.Register.default_fhandler,
                         htd.FileType.VARIABLE_WIDTH_FILE,
                         "Default file type should be variable width")

        # test config
        self.assertFalse(htd.Config.debug, "Debug should be False")
        self.assertFalse(htd.Config.force_sync, "Force Sync should be False")
        self.assertFalse(htd.Config.progress, "Progress should be False")
        self.assertTrue(htd.Config.errors, "Errors should be True")

    def test_create_normal(self):
        """
        Test normal creation of DB/Table/CompTable
        DB, Table, Composite Table
            __class__
            is_db
            user
        """

        #------------------------------------------------ Create normal DB file
        con = htd.create(names[0])

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a DB
        self.assertTrue(con.is_db, "Connection is not a DB")

        # Check User
        self.assertEqual(con.user, "guest", "User is not guest")

        con.close()

        # Destroy
        self.tearDown()

        #-------------------------------------------------- Create normal table
        con = htd.create(names[1], identifier="id",
                         attributes={"id": htd.Attribute(htd.DataType.BYTES)})

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertFalse(con.is_db, "Connection is not a table")

        # Check User
        self.assertEqual(con.user, "guest", "User is not guest")

        con.close()

        # Destroy
        self.tearDown()

        #----------------------------------------------- Create composite table

        con = htd.create(names[2], identifier=["id1", "id2"],
                         attributes={"id1": htd.Attribute(htd.DataType.BYTES),
                                     "id2": htd.Attribute(htd.DataType.BYTES)})

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertFalse(con.is_db, "Connection is not a table")

        # Check User
        self.assertEqual(con.user, "guest", "User is not guest")

        con.close()

        # Destroy
        self.tearDown()

        # TODO: test other file types

    def test_create_exteme_location(self):
        """
        Test location
        location is not str
        location is None
        """

        # Location of wrong type
        try:
            con = htd.create(123)
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, location is not str. (%s)"
                             % e)
        else:
            raise ValueError("Location is not str")

        # Location is None
        try:
            con = htd.create(None)
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, location is None. (%s)"
                             % e)
        else:
            raise ValueError("Location is None")

    def test_create_exteme_attributes(self):
        """
        Test attributes
        attributes is not dict
        attributes is None
        """

        # Attributes of wrong type
        try:
            con = htd.create(names[0], attributes=123)
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, attributes not dict. (%s)"
                             % e)
        else:
            raise ValueError("Attributes is not dict")

        # Attributes is None
        try:
            con = htd.create(names[1], attributes=None)
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, attributes is None. (%s)"
                             % e)
        else:
            raise ValueError("Attributes is None")

    def test_create_exteme_identifier(self):
        """
        Test create table with extreme identifier values
        identifier value is None
        identifier is not in attributes
        identifier is in attribute but of wrong type
        identifier is not in attribute and of wrong type
        """

        # Identifier value is None
        try:
            con = htd.create(names[2], identifier=[None],
                             attributes={"id":
                                         htd.Attribute(htd.DataType.BYTES)})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, identifier is None(%s)" % e)
        else:
            raise ValueError("Identifier is None and attributes isn't empty")

        # Identifier is not in attributes
        try:
            con = htd.create(names[2], identifier=["notinatrributes"],
                             attributes={"id":
                                         htd.Attribute(htd.DataType.BYTES)})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, identifier not in "
                             "attributes. (%s)" % e)
        else:
            raise ValueError("Identifier not in attributes")

        # Identifier is not in attributes and of wrong type
        try:
            con = htd.create(names[2], identifier=[1],
                             attributes={"id":
                                         htd.Attribute(htd.DataType.BYTES)})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, identifier not in"
                             " attribute and of wrong type. (%s)" % e)
        else:
            raise ValueError("Identifier not in attribute and of wrong type")

        # Identifier is in attributes but of wrong type
        try:
            con = htd.create(names[2], identifier=[1],
                             attributes={1: htd.Attribute(htd.DataType.BYTES)})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, identifier wrong type (%s)"
                             % e)
        else:
            raise ValueError("Identifier wrong type")

    def test_create_normal_users(self):
        """
        Test create with normal users
        DB, Table, Composite Table
        - Single User, Multiuser
            __class__
            is_db
            user
        """
        #------------------------------------------------ Create normal DB file

        #
        # Single user
        #

        con = htd.create(names[0], users={"test1": htd.User("abc", "roadx")})

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertTrue(con.is_db, "Connection is not a DB")

        # Check User
        self.assertEqual(con.user, "test1", "User is not test1")

        con.close()

        # Destroy
        self.tearDown()

        #
        # Multi User
        #

        con = htd.create(names[1],
                         users={"test1": htd.User("abc", "roadx"),
                                "test2": htd.User("def", "r")},
                         login_user="test2")

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertTrue(con.is_db, "Connection is not a DB")

        # Check User
        self.assertEqual(con.user, "test2", "User is not test2")

        con.close()

        # Destroy
        self.tearDown()

        #-------------------------------------------------- Create normal table

        #
        # Single user
        #

        con = htd.create(names[2],
                         identifier="id",
                         attributes={"id": htd.Attribute(htd.DataType.BYTES)},
                         users={"test1": htd.User("abc", "roadx")})

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertFalse(con.is_db, "Connection is not a table")

        # Check User
        self.assertEqual(con.user, "test1", "User is not test1")

        con.close()

        # Destroy
        self.tearDown()

        #
        # Multiuser
        #

        con = htd.create(names[3],
                         identifier="id",
                         attributes={"id": htd.Attribute(htd.DataType.BYTES)},
                         users={"test1": htd.User("abc", "roadx"),
                                "test2": htd.User("def", "r")},
                         login_user="test2")

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertFalse(con.is_db, "Connection is not a table")

        # Check User
        self.assertEqual(con.user, "test2", "User is not test2")

        con.close()

        # Destroy
        self.tearDown()

        #----------------------------------------------- Create composite table

        #
        # Single user
        #

        con = htd.create(names[0],
                         identifier=["id1", "id2"],
                         attributes={"id1": htd.Attribute(htd.DataType.BYTES),
                                     "id2": htd.Attribute(htd.DataType.BYTES)},
                         users={"test1": htd.User("abc", "roadx")})

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertFalse(con.is_db, "Connection is not a table")

        # Check User
        self.assertEqual(con.user, "test1", "User is not test1")

        con.close()

        # Destroy
        self.tearDown()

        #
        # Multi user
        #

        con = htd.create(names[1],
                         identifier=["id1", "id2"],
                         attributes={"id1": htd.Attribute(htd.DataType.BYTES),
                                     "id2": htd.Attribute(htd.DataType.BYTES)},
                         users={"test1": htd.User("abc", "roadx"),
                                "test2": htd.User("def", "r")},
                         login_user="test2")

        # Check is a Connection
        self.assertEqual(con.__class__.__name__, "Connection",
                         "Connection is not a connection instance.")

        # Check is a table
        self.assertFalse(con.is_db, "Connection is not a table")

        # Check User
        self.assertEqual(con.user, "test2", "User is not test2")

        con.close()

        # Destroy
        self.tearDown()

    def test_create_users_extreme(self):
        """
        Test create with extreme user values
        DB
        users is not dict
        user is not User instance
        no user will full access
        multiuser but no login_user
        login_user not in users
        username is not str
        """

        #------------------------------------------------- Invalid users format

        #
        # BUG 1036395
        #

        try:
            con = htd.create(names[0],
                             users=1)
            con.close()
        except ValueError:
            pass  # Expected result
        except Exception as e:
            raise ValueError("Incorrect exception, incorrect users type"
                             " when creating a table. (%s)" % e)
        else:
            raise ValueError("No errors using bad users type creating "
                             "a table")

        #------------------------------------------------ Invalid user in users
        try:
            con = htd.create(names[1],
                             users={"a": ("pass", "acc")})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, incorrect user in users"
                             " when creating a table. (%s)" % e)
        else:
            raise ValueError("No errors using bad user in users creating "
                             "a table")

        #--------------------------------------------- No user will full access
        try:
            con = htd.create(names[2],
                             users={"a": htd.User("pass", "ra")})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, no full perm user"
                             " when creating a table. (%s)" % e)
        else:
            raise ValueError("No errors using no full perm user creating "
                             "a table")

        #----------------------------------------- Multi user but no login_user
        try:
            con = htd.create(names[3],
                             users={"a": htd.User("pass"), "b": htd.User("b")})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, no login_user"
                             " when creating a table. (%s)" % e)
        else:
            raise ValueError("No errors using no login_user creating "
                             "a table")

        #---------------------------------------------- login_user not in users
        try:
            con = htd.create(names[0],
                             users={"a": htd.User("pass"), "b": htd.User("b")},
                             login_user="c")
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, login_user not in users"
                             " when creating a table. (%s)" % e)
        else:
            raise ValueError("No errors using login_user not in users "
                             "creating a table")

        #-------------------------------------------------- Username is not str
        try:
            con = htd.create(names[1],
                             users={1: htd.User("pass"), "b": htd.User("b")})
            con.close()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, bad username when creating "
                             "a table. (%s)" % e)
        else:
            raise ValueError("No errors using bad username creating a table")
