#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class RecordTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_dict_emu(self):
        """
        Test Dict emulation
        - table
        """

        #----------------------------------------------------------- Table test

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con = htd.create(names[3], identifier="a", attributes=atts)

        cur = con[None]

        cur[1] = {"b": 2, "c": 3}
        cur[4] = {"b": 5, "c": 6}

        self.assertEqual(cur[1], {"a": 1, "b": 2, "c": 3}, "Dict comparison")
        self.assertEqual(cur[4], {"a": 4, "b": 5, "c": 6}, "Dict comparison")

        con.close()

    def test_from_table(self):
        """
        Test from table
        - table, db
        - connection, get_item, table
        """

        #----------------------------------------------------------- Table test

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con = htd.create(names[3], identifier="a", attributes=atts)

        cur = con[None]

        cur[1] = {"b": 2, "c": 3}
        cur[4] = {"b": 5, "c": 6}

        rec = cur[1]

        # Connection
        self.assertEqual(rec.connection, con, "Connection is not correct")

        # Get item
        self.assertEqual(rec["a"], 1, "a (key) is not correct")
        self.assertEqual(rec["b"], 2, "b is not correct")
        self.assertEqual(rec["c"], 3, "c is not correct")

        # Table
        self.assertEqual(rec.table, None, "Table is incorrect")

        con.close()

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[4], identifier="a", attributes=atts)

        cur = con["abc"]

        cur[1] = {"b": 2, "c": 3}
        cur[4] = {"b": 5, "c": 6}

        rec = cur[1]

        # Connection
        self.assertEqual(rec.connection, con, "Connection is not correct")

        # Get item
        self.assertEqual(rec["a"], 1, "a (key) is not correct")
        self.assertEqual(rec["b"], 2, "b is not correct")
        self.assertEqual(rec["c"], 3, "c is not correct")

        # Table
        self.assertEqual(rec.table, "abc", "Table is incorrect")

        con.close()

    def test_relation_get(self):
        """
        Test relation get
        - table, db
        - many->one, one->many, no relation, no table
        - not enough atts (used select)
        """

        #-------------------------------------------------------------- DB Test
        con = htd.connect(names[1])

        # Prepare
        atts = {"a": htd.Attribute(htd.DataType.INTEGER),
                "b": htd.Attribute(htd.DataType.INTEGER),
                "c": htd.Attribute(htd.DataType.INTEGER)}
        con["abc"] = htd.create(names[3], identifier="a", attributes=atts)

        atts = {"d": htd.Attribute(htd.DataType.INTEGER),
                "e": htd.Attribute(htd.DataType.INTEGER),
                "a": htd.Attribute(htd.DataType.INTEGER)}
        con["dea"] = htd.create(names[4], identifier="d", attributes=atts)

        # Add data
        cur0 = con["abc"]
        cur1 = con["dea"]

        cur0[1] = {"b": 10, "c": 100}
        cur0[2] = {"b": 20, "c": 200}
        cur0[3] = {"b": 30, "c": 300}

        cur1[1] = {"e": 10, "a": 1}
        cur1[2] = {"e": 20, "a": 2}
        cur1[3] = {"e": 30, "a": 2}

        # No table
        rec = cur0[1]

        try:
            rec["att.notatable"]
        except KeyError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, no table (%s)" % e)
        else:
            raise ValueError("No errors using no table")

        # No relation
        try:
            rec["a.dea"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, no table (%s)" % e)
        else:
            raise ValueError("No errors using no table")

        con.link("abc.a", "dea.a")

        # Many->one
        self.assertEqual(cur1[1]["a.abc"]["c"], 100, "Many->one is incorrect")

        # One->Many
        l = [rec["e"] for rec in cur0[2]["a.dea"]]

        self.assertListEqual(l, [20, 30], "One->Many incorrect")

        # Not enough attributes Many->one
        rec = list(cur1.select(["d", "e"]))[0]

        try:
            rec["a.abc"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, m1 not enough att (%s)" % e)
        else:
            raise ValueError("No errors using m1 not enough att")

        # Not enough attributes one->many
        rec = list(cur0.select(["b", "c"]))[0]

        try:
            rec["a.dea"]
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, 1m not enough att (%s)" % e)
        else:
            raise ValueError("No errors using 1m not enough att")

        con.close()
