#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
import unittest


class RelationTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_attributes(self):
        """
        Test attributes module for relationships

        - delete attribute with relationship (m)
        - updating attribute with relationship (m&1)
        """

        con = htd.connect(names[1])

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING),
                      "count": htd.Attribute(htd.DataType.INTEGER)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        # key-att
        con.link("b.id", "a.parent")

        cur_a = con["a"]
        cur_b = con["b"]

        atts_a = cur_a.attributes
        atts_b = cur_b.attributes

        # Delete attribute with relationship m
        try:
            del atts_a["parent"]
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Delete attribute with relationship m")

        # Update attribute with relationship m
        try:
            atts_a["parent"] = htd.Attribute()
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Delete attribute with relationship m")

        # Update attribute with relationship 1
        try:
            atts_b["id"] = htd.Attribute()
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Delete attribute with relationship 1")

    def test_creation(self):
        """
        Test creation of relationships
        - db, table
        - att same/diff name exists/!exist
        - key-att, att-att, key-key, att-key
        - already exists
        - no tables
        - bad attribute name (no .)
        """

        #----------------------------------------------------------- Table test
        con = htd.connect(names[0])

        try:
            con.link("a.b", "c.d")
        except ValueError:
            pass
        else:
            raise ValueError("Should only work with database")

        try:
            con.unlink("a.b", "c.d")
        except ValueError:
            pass
        else:
            raise ValueError("Should only work with database")

        try:
            list(con.relation)
        except ValueError:
            pass
        else:
            raise ValueError("Should only work with database")

        con.close()

        #-------------------------------------------------------------- DB test
        con = htd.connect(names[1])

        # No tables
        try:
            con.link("a.b", "c.d")
        except ValueError:
            pass
        else:
            raise ValueError("Tables do not exist")

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING),
                      "count": htd.Attribute(htd.DataType.INTEGER)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        # key-att
        con.link("b.id", "a.parent")

        self.assertListEqual(list(con.relation), [(("a", "parent"),
                                                   ("b", "id"))],
                             "Relation is incorrect")

        con.unlink("b.id", "a.parent")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        # att-key
        con.link("a.parent", "b.id")

        self.assertListEqual(list(con.relation), [(("a", "parent"),
                                                   ("b", "id"))],
                             "Relation is incorrect")

        con.unlink("b.id", "a.parent")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        # att-att
        try:
            con.link("a.parent", "b.data")
        except ValueError:
            pass
        else:
            raise ValueError("att to att should fail")

        # key-key
        try:
            con.link("a.id", "b.id")
        except ValueError:
            pass
        else:
            raise ValueError("key to key should fail")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        # Already exists
        con.link("a.parent", "b.id")

        self.assertListEqual(list(con.relation), [(("a", "parent"),
                                                   ("b", "id"))],
                             "Relation is incorrect")

        con.link("a.parent", "b.id")

        self.assertListEqual(list(con.relation), [(("a", "parent"),
                                                   ("b", "id"))],
                             "Relation is incorrect")

        con.unlink("b.id", "a.parent")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        # Bad attribute name
        try:
            con.link("a", "b.id")
        except ValueError:
            pass
        else:
            raise ValueError("Bad attribute name")

        # Attribute !exist
        try:
            con.link("a.parent", "b.bad")
        except ValueError:
            pass
        else:
            raise ValueError("Attribute should not exist")

        # Attribute diff
        try:
            con.link("a.count", "b.id")
        except ValueError:
            pass
        else:
            raise ValueError("Attributes should be different")

        con.close()

    def test_creation_with_data(self):
        """
        Test creation with data in the tables
        - Many in 1, Many not in 1
        """

        con = htd.connect(names[1])

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING),
                      "count": htd.Attribute(htd.DataType.INTEGER)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        # Add data
        cur_a = con["a"]
        cur_b = con["b"]

        cur_a["1"] = {"parent": "1"}
        cur_a["2"] = {"parent": "2"}
        cur_a["3"] = {"parent": "3"}

        cur_b["1"] = {}
        cur_b["2"] = {}
        cur_b["3"] = {}

        # Many in 1
        con.link("b.id", "a.parent")

        self.assertListEqual(list(con.relation), [(("a", "parent"),
                                                   ("b", "id"))],
                             "Relation is incorrect")

        con.unlink("b.id", "a.parent")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        # Add data
        cur_a["4"] = {"parent": "4"}

        # Many not in 1
        try:
            con.link("b.id", "a.parent")
        except KeyError:
            pass
        else:
            raise KeyError("Many not in 1")

        self.assertListEqual(list(con.relation), [], "Relation not clear")

        con.close()

    def test_cursor_rules(self):
        """
        Test Cursor relationship rules

        Delete record from 1 end should fail
        Delete many from 1 end should fail

        Delete record from 1 end (none in m) should be fine

        Set record in many must be in 1
        Insert record(s) in many must be in 1
        Update record in many must be in 1
        Update record in many without relational fields

        Set record in many with NULL ref to 1
        """

        con = htd.connect(names[1])

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING),
                      "count": htd.Attribute(htd.DataType.INTEGER)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        cur_a = con["a"]
        cur_b = con["b"]

        con.link("b.id", "a.parent")

        # Set Record in many must be in 1
        try:
            cur_a["1"] = {"parent": "1"}
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Set Record in many must be in 1")

        # Insert record(s) in many must be in 1
        try:
            cur_a.insert([{"id": "1", "parent": "1"},
                          {"id": "2", "parent": "2"}])
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Insert Record in many must be in 1")

        # Prepare
        cur_b["1"] = {}
        cur_b["2"] = {}
        cur_b["3"] = {}

        cur_a["0"] = {"parent": "1"}

        # Delete record from 1 end should fail
        try:
            del cur_b["1"]
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Delete record from 1 end should fail")

        # Delete many record(s) from 1 end should fail
        try:
            cur_b.delete()
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Delete many records from 1 end should fail")

        # Prepare
        cur_a["1"] = {"parent": "2"}

        # Update record in many must be in 1
        try:
            updater = cur_a.update()

            for record in updater:
                record["parent"] = "5"
                updater.send(record)
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Update record in many must be in 1")

        # Update record in many without relational fields
        cur_a["1"] = {"count": 15}
        cur_a["1"] = {"count": 10}

        updater = cur_a.update()

        for record in updater:
            updater.send({"count": 10})

        # Set record in many with NULL ref to 1
        try:
            cur_a["4"] = {}
        except KeyError:
            pass
        else:
            raise ValueError("Set record in many with NULL ref to 1")

        # Delete record from 1 end (none in m) should be fine
        del cur_a["0"]
        del cur_b["1"]

        con.close()

    def test_identifier_rules(self):
        """
        Test rules for identifier

        - Change key when relationship exists
        - Change key when new key has relationship
        - Change composite key when relationship exists
        """
        con = htd.connect(names[1])

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING),
                      "count": htd.Attribute(htd.DataType.INTEGER)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        # key-att
        con.link("b.id", "a.parent")

        cur_a = con["a"]
        cur_b = con["b"]

        # Change key when relationship exists
        try:
            cur_b.identifier = "data"
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Change key when relationship exists")

        # Change key when new key has relationship
        try:
            cur_a.identifier = "parent"
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Change key when new key has relationship")

        # Prepare
        con.unlink("b.id", "a.parent")

        cur_b.identifier = ["id", "data"]

        con.link("b.id", "a.id")

        # Change composite key when relationship exists
        try:
            cur_b.identifier = "data"
        except htd.IntegrityError:
            pass
        else:
            raise ValueError("Change composite key when relationship exists")

        con.close()

    def test_record(self):
        """
        Test record relations
        - 1->m, m->1
        - No table
        - No relation
        - Attribute not in data (m1 and 1m)
        """

        con = htd.connect(names[1])

        # Prepare
        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "name": htd.Attribute(htd.DataType.STRING),
                      "parent": htd.Attribute(htd.DataType.STRING),
                      "count": htd.Attribute(htd.DataType.INTEGER)}

        con["a"] = htd.create(names[2], identifier="id", attributes=attributes)

        attributes = {"id": htd.Attribute(htd.DataType.STRING),
                      "data": htd.Attribute(htd.DataType.STRING),
                      "test": htd.Attribute(htd.DataType.STRING)}
        con["b"] = htd.create(names[3], identifier="id", attributes=attributes)

        cur_a = con["a"]
        cur_b = con["b"]

        con.link("b.id", "a.parent")

        cur_b["1"] = {"data": "one"}
        cur_b["2"] = {"data": "two"}
        cur_b["3"] = {"data": "three"}

        cur_a["1"] = {"parent": "2"}
        cur_a["2"] = {"parent": "2"}
        cur_a["3"] = {"parent": "1"}
        cur_a["4"] = {"parent": "3"}

        # 1->m
        l = [{"id": "1", "name": None, "parent": "2", "count": None},
             {"id": "2", "name": None, "parent": "2", "count": None}]

        self.assertListEqual(list(cur_b["2"]["id.a"]), l, "1->m list")

        # m->1
        self.assertEqual(cur_a["3"]["parent.b"]["data"], "one", "m->1 list")

        # No table
        try:
            cur_a["2"]["parent.no_table"]
        except KeyError:
            pass
        else:
            raise KeyError("No table")

        # No relation
        try:
            cur_a["2"]["count.b"]
        except ValueError:
            pass
        else:
            raise KeyError("No relation")

        # Attribute not in data m1
        for record in cur_a.select(["count", "name"]):
            try:
                record["parent.b"]
            except ValueError:
                pass
            else:
                raise ValueError("Attribute not in data m1")

        # Attribute not in data 1m
        for record in cur_b.select(["data", "test"]):
            try:
                record["id.a"]
            except ValueError:
                pass
            else:
                raise ValueError("Attribute not in data 1m")

        con.close()
