#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from os.path import exists
from os import unlink
from threading import Thread
from time import sleep
import unittest


class ThreadTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def delete(self, cur, start, end, jump=1, wait=0):
        for i in range(start, end, jump):
            del cur[i]
            sleep(wait)

    def get(self, cur, start, end, jump=1, wait=0):
        for i in range(start, end):
            x = cur[i]
            sleep(wait)

    def list(self, cur, repeat=100, wait=0):
        for i in range(repeat):
            for record in cur:
                sleep(wait)

    def set(self, cur, start, end, jump=1, wait=0):
        for i in range(start, end):
            cur[i] = {}
            sleep(wait)

    def test_set_list_same_cur(self):
        """
        Test setting different keys in the table
        at the *same* time list the records in the table
        """

        con = htd.create(names[1], identifier="id",
                         attributes={"id":
                                     htd.Attribute(htd.DataType.INTEGER)})

        cur1 = con[None]

        self.set(cur1, 0, 1000, 1)

        threads = []
        threads.append(Thread(target=self.set, args=(cur1, 0, 1000, 1)))
        threads.append(Thread(target=self.list, args=(cur1, 10, 0)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(len(cur1), 1000, "Should be 1000 recs (cur1)")

        con.close()

    def test_set_get_same_key_cur(self):
        """
        Test setting different keys in the table
        at the *same* time
        """

        con = htd.create(names[1], identifier="id",
                         attributes={"id":
                                     htd.Attribute(htd.DataType.INTEGER)})

        cur1 = con[None]

        self.set(cur1, 0, 1000, 1)

        threads = []
        threads.append(Thread(target=self.set, args=(cur1, 0, 1000, 1)))
        threads.append(Thread(target=self.get, args=(cur1, 0, 1000, 1)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(len(cur1), 1000, "Should be 1000 recs (cur1)")

        con.close()

    def test_set_diff_keys(self):
        """
        Test setting different keys in the table
        at the *same* time
        """

        con = htd.create(names[1], identifier="id",
                         attributes={"id":
                                     htd.Attribute(htd.DataType.INTEGER)})

        cur1 = con[None]
        cur2 = con[None]

        threads = []
        threads.append(Thread(target=self.set, args=(cur1, 0, 1000, 2)))
        threads.append(Thread(target=self.set, args=(cur2, 1, 1000, 2)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(len(cur1), len(cur2), "Cursors are different lengths")
        self.assertEqual(len(cur1), 1000, "Should be 1000 recs (cur1)")
        self.assertEqual(len(cur2), 1000, "Should be 1000 recs (cur2)")

        con.close()

    def test_set_same_keys(self):
        """
        Test setting the same keys in the table
        at the *same* time
        """

        con = htd.create(names[1], identifier="id",
                         attributes={"id":
                                     htd.Attribute(htd.DataType.INTEGER)})

        cur1 = con[None]
        cur2 = con[None]

        threads = []
        threads.append(Thread(target=self.set, args=(cur1, 0, 1000, 1)))
        threads.append(Thread(target=self.set, args=(cur2, 0, 1000, 1)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(len(cur1), len(cur2), "Cursors are different lengths")
        self.assertEqual(len(cur1), 1000, "Should be 1000 recs (cur1)")
        self.assertEqual(len(cur2), 1000, "Should be 1000 recs (cur2)")

        con.close()

    def test_del_diff_keys(self):
        """
        Test deleting different keys in the table
        at the *same* time
        """

        con = htd.create(names[1], identifier="id",
                         attributes={"id":
                                     htd.Attribute(htd.DataType.INTEGER)})

        cur1 = con[None]
        cur2 = con[None]

        self.set(cur1, 0, 100, 1)

        self.assertEqual(len(cur1), 100, "Should be 1000 recs (prepare)")

        threads = []
        threads.append(Thread(target=self.delete, args=(cur1, 0, 100, 2)))
        threads.append(Thread(target=self.delete, args=(cur2, 1, 101, 2)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(len(cur1), len(cur2), "Cursors are different lengths")
        self.assertEqual(len(cur1), 0, "Should be 0 recs (cur1)")
        self.assertEqual(len(cur2), 0, "Should be 0 recs (cur2)")

        con.close()
