#!/usr/bin/env python3
'''
This file is part of HTD (HayzenTech Database)

Copyright (C) 2011 - 2013 Andrew Hayzen

HTD is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HTD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with HTD. If not, see <http://www.gnu.org/licenses/>

This document conforms to the PEP 8 Style Guide
'''

from .base import htd_engine as htd, names
from hashlib import sha256
from os.path import exists
from os import unlink
import unittest


class AuthenticateTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test init
        - username, password, user_obj
        - all None, just user_obj (bad and good), just password, user+pass
        """

        # All None
        try:
            auth = htd.Authenticate()
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, All None (%s)" % e)
        else:
            raise ValueError("No errors using All None")

        # user_obj good
        user = htd.User("def")
        user.username = "abc"

        auth = htd.Authenticate(user_obj=user)

        self.assertEqual(auth.username, "abc", "Username incorrect")
        self.assertEqual(auth.password, user.password, "Password incorrect")

        # user_obj bad
        try:
            auth = htd.Authenticate(user_obj="abc")
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Bad user_obj (%s)" % e)
        else:
            raise ValueError("No errors using Bad user_obj")

        # Only password
        try:
            auth = htd.Authenticate(password="abc")
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Only password (%s)" % e)
        else:
            raise ValueError("No errors using Only password")

        # User + pass
        auth = htd.Authenticate(username="abc", password="def")

        self.assertEqual(auth.username, "abc", "User is not abc")

        password = sha256(b"def").hexdigest()
        self.assertEqual(auth.password, password, "Password is incorrect")

    def test_repr(self):
        """
        Test repr
        - username
        - None, exists
        """

        # None
        auth = htd.Authenticate(user_obj=htd.User())

        exp_repr = "<HTD%s Authenticated User (None)>" % htd.__version__
        self.assertEqual(repr(auth), exp_repr, "Username is None")

        # exists
        auth = htd.Authenticate(username="abc")

        exp_repr = "<HTD%s Authenticated User (abc)>" % htd.__version__
        self.assertEqual(repr(auth), exp_repr, "Username is abc")


class UserTest(unittest.TestCase):
    def setUp(self):
        for name in names:
            if exists(name):
                unlink(name)

        #-------------------------------------------- Table and composite table
        htd.create(names[0],
                   identifier="id",
                   attributes={"id": htd.Attribute(htd.DataType.STRING)}
                   ).close()

        #------------------------------------------------------------------- DB
        htd.create(names[1]).close()

    def tearDown(self):
        for name in names:
            if exists(name):
                unlink(name)

    def test_init(self):
        """
        Test init
        - password, access
        """

        # Standard
        user = htd.User()

        self.assertEqual(user.username, None, "Username is not None")
        self.assertEqual(user.password, None, "Password is not None")
        self.assertEqual(user.access, "roadx", "Access is not roadx")

        # Pass and access
        user = htd.User("abc", "rd")

        password = sha256(b"abc").hexdigest()
        self.assertEqual(user.username, None, "Username is not None")
        self.assertEqual(user.password, password, "Password is not correct")
        self.assertEqual(user.access, "rd", "Access is not rd")

    def test_repr(self):
        """
        Test repr
        - username, access
        """

        # Standard
        user = htd.User()

        repr_str = ("<HTD%s User (None with 'roadx' permissions)>" %
                    htd.__version__)

        self.assertEqual(repr(user), repr_str, "Standard user repr wrong")

        # Username and access
        user = htd.User(access="rd")
        user.username = "abc"

        repr_str = ("<HTD%s User (abc with 'rd' permissions)>" %
                    htd.__version__)

        self.assertEqual(repr(user), repr_str, "Username and pass repr wrong")

    def test_access(self):
        """
        Test access
        - None, roadx, ra, abc, 123 (wrong data type)
        """

        # Standard
        user = htd.User()

        self.assertEqual(user.access, "roadx", "Standard access incorrect")

        # None
        try:
            user.access = None
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Access is None (%s)" % e)
        else:
            raise ValueError("No errors using Access is None")

        # Ra
        user.access = "ra"
        self.assertEqual(user.access, "ra", "Access is not ra")

        # roadx
        user.access = "roadx"
        self.assertEqual(user.access, "roadx", "Access is not roadx")

        # abc
        try:
            user.access = "abc"
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Access is abc (%s)" % e)
        else:
            raise ValueError("No errors using Access is abc")

        # 123 bad dtype
        try:
            user.access = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Access is 123 int (%s)" % e)
        else:
            raise ValueError("No errors using Access is 123 int")

    def test_password(self):
        """
        Test access
        - None, qwerty, asdfgh, 123 (wrong data type)
        """

        # Standard
        user = htd.User()

        self.assertEqual(user.password, None, "Standard password incorrect")

        # qwerty
        user.password = "qwerty"
        password = sha256(b"qwerty").hexdigest()
        self.assertEqual(user.password, password, "Password is not qwerty")

        # asdfgh
        user.password = "asdfgh"
        password = sha256(b"asdfgh").hexdigest()
        self.assertEqual(user.password, password, "Password is not asdfgh")

        # 123 bad dtype
        try:
            user.password = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Password is 123 int (%s)" %
                             e)
        else:
            raise ValueError("No errors using Password is 123 int")

    def test_to_dict(self):
        """
        Test to_dict
        - password, access
        """

        # Standard
        user = htd.User()

        exp_dict = {"pass": None, "acc": "roadx"}
        self.assertDictEqual(user.to_dict(), exp_dict, "Standard to_dict bad")

        # pass=abc, access=ra
        user.password = "abc"
        user.access = "ra"

        exp_dict = {"pass": sha256(b"abc").hexdigest(), "acc": "ra"}
        self.assertDictEqual(user.to_dict(), exp_dict, "Pass abc, access ra "
                             "bad to_dict")

    def test_username(self):
        """
        Test username
        - None, qwerty, asdfgh, 123 (wrong data type)
        """

        # Standard
        user = htd.User()

        self.assertEqual(user.username, None, "Standard username incorrect")

        # qwerty
        user.username = "qwerty"
        self.assertEqual(user.username, "qwerty", "Username is not qwerty")

        # asdfgh
        user.username = "asdfgh"
        self.assertEqual(user.username, "asdfgh", "Username is not asdfgh")

        # 123 bad dtype
        try:
            user.username = 123
        except ValueError:
            pass
        except Exception as e:
            raise ValueError("Incorrect exception, Username is 123 int (%s)" %
                             e)
        else:
            raise ValueError("No errors using Username is 123 int")
